/*
 * Driver for NAND support, Rick Bronson
 * borrowed heavily from:
 * (c) 1999 Machine Vision Holdings, Inc.
 * (c) 1999, 2000 David Woodhouse <dwmw2@infradead.org>
 *
 * Added 16-bit nand support
 * (C) 2004 Texas Instruments
 */

#include <common.h>
#include <command.h>
#include <malloc.h>
#include <asm/io.h>
#include <watchdog.h>

#ifdef CONFIG_SHOW_BOOT_PROGRESS
# include <status_led.h>
# define SHOW_BOOT_PROGRESS(arg)	show_boot_progress(arg)
#else
# define SHOW_BOOT_PROGRESS(arg)
#endif

#if (CONFIG_COMMANDS & CFG_CMD_NAND)

#include <linux/mtd/nand.h>
#include <linux/mtd/nand_ids.h>
#include <jffs2/jffs2.h>
#include <asm/vixs_nand.h>
#include <asm/vixs_secu.h>
#include <asm/viper.h>

#define ROUND_DOWN(value,boundary)      ((value) & (~((boundary)-1)))

extern ulong load_addr;

/* the Nand Env offset was calculated by (4 + 8*2) * blocksize */
ulong CFG_ENV_OFFSET;

/*
 * Definition of the out of band configuration structure
 */
struct nand_oob_config {
	int ecc_pos[6];		/* position of ECC bytes inside oob */
	int badblock_pos;	/* position of bad block flag inside oob -1 = inactive */
	int eccvalid_pos;	/* position of ECC valid flag inside oob -1 = inactive */
} oob_config = { {0}, 0, 0};

#undef	NAND_DEBUG
#undef	PSYCHO_DEBUG

/* ****************** WARNING *********************
 * When ALLOW_ERASE_BAD_DEBUG is non-zero the erase command will
 * erase (or at least attempt to erase) blocks that are marked
 * bad. This can be very handy if you are _sure_ that the block
 * is OK, say because you marked a good block bad to test bad
 * block handling and you are done testing, or if you have
 * accidentally marked blocks bad.
 *
 * Erasing factory marked bad blocks is a _bad_ idea. If the
 * erase succeeds there is no reliable way to find them again,
 * and attempting to program or erase bad blocks can affect
 * the data in _other_ (good) blocks.
 */
#define	 ALLOW_ERASE_BAD_DEBUG 0

#define CONFIG_MTD_NAND_ECC  /* enable ECC */
#define CONFIG_MTD_NAND_ECC_JFFS2

#define readl(addr) (*(volatile unsigned int *) (addr))
#define writel(b,addr) (*(volatile unsigned int *) (addr) = (b))
#define ntohl(x) \
({ \
	unsigned int __x = (x); \
	((unsigned int)( \
		(((unsigned int)(__x) & (unsigned int)0x000000ffUL) << 24) | \
		(((unsigned int)(__x) & (unsigned int)0x0000ff00UL) <<  8) | \
		(((unsigned int)(__x) & (unsigned int)0x00ff0000UL) >>  8) | \
		(((unsigned int)(__x) & (unsigned int)0xff000000UL) >> 24) )); \
})


#define CFG_MAX_NAND_DEVICE	1	/* Max number of NAND Controller */

#define BOOT1_BLKCNT	4
#define BOOT2_BLKCNT	8
#define UBOOT_ENV_BLKCNT	4
#define KERNEL_SIZE 	0x800000

#define NAND_ChipID_UNKNOWN	0x00
#define NAND_MAX_FLOORS 1
#define NAND_MAX_CHIPS 4

#define XCODE3_NAND_REG_BASE		(XC_SOC_PROC_MMREG_BASE + 0xe000)
#define XC_NFC_MAX_BANK_NUM			0x8  //it is max chip in software too

#define XC_NFC_BANK_SEL_MASK		0xff000000
#define XC_NFC_BANK_SEL_SHIFT		24
#define XC_NFC_CMDPARAM_MASK		0xffff0000
#define XC_NFC_CMDPARAM_SHIFT		16
#define XC_NFC_CMDPARAM_SHIFT2		24

//command code
#define XC_NFC_CMDCODE_READ			0x00
#define XC_NFC_CMDCODE_READ_DWORD	0x02
#define XC_NFC_CMDCODE_PAGE_PROG	0x80
#define XC_NFC_CMDCODE_ERASE		0x60
#define XC_NFC_CMDCODE_RESET		0xFF
#define XC_NFC_CMDCODE_READ_STATUS	0x70
#define XC_NFC_CMDCODE_READ_ID		0x90


//list all the NFC registers, maybe move to a seperated header file later
#define XC_NFC_BANK_COL_ADDR	0x0000
#define XC_NFC_PAGE_ROW_ADDR	0x0004
#define XC_NFC_DATA_ADDR		0x0008
#define XC_NFC_SPARE_ADDR		0x000C
#define XC_NFC_COMMAND			0x0010
#define XC_NFC_NANDF_RDATA		0x0014
#define XC_NFC_DEV_STATUS		0x0018
#define XC_NFC_DEV_RB			0x001C
#define XC_NFC_CMD_STATUS		0x0020
#define XC_NFC_ID				0x0024
#define XC_NFC_EXTENDED_ID		0x0028
#define XC_NFC_ECC_STATUS0		0x002C
#define XC_NFC_ECC_STATUS1		0x0030
#define XC_NFC_ECC_STATUS2		0x0034 
#define XC_NFC_ECC_STATUS3		0x0038
#define XC_NFC_ECC_STATUS4		0x003C
#define XC_NFC_ECC_STATUS5		0x0040
#define XC_NFC_ECC_STATUS6		0x0044
#define XC_NFC_ECC_STATUS7		0x0048
#define XC_NFC_ECC_ERR_MAP0		0x004C
#define XC_NFC_ECC_ERR_MAP1		0x0050
#define XC_NFC_ECC_ERR_MAP2		0x0054
#define XC_NFC_ECC_ERR_MAP3		0x0058
#define XC_NFC_INT_STATUS		0x005C
#define XC_NFC_HOST_INT_MASK	0x0060
#define XC_NFC_MIPS_INT_MASK	0x0064
#define XC_NFC_CONTROL			0x0068
#define XC_NFC_ECC_CONTROL		0x006C
#define XC_NFC_BCH_CONTROL		0x0070
#define XC_NFC_BCH_GEN_POLY0	0x0074
#define XC_NFC_BCH_GEN_POLY1	0x0078
#define XC_NFC_BCH_GEN_POLY2	0x007C
#define XC_NFC_BCH_GEN_POLY3	0x0080
#define XC_BCH_SYNDRM_ADDR		0x0084
#define XC_NFC_WRITE_PROT		0x0088
#define XC_NFC_BOOT_CFG			0x008C
#define XC_NFC_BOOT_ADDR		0x0090
#define XC_NFC_BOOTSTRAP		0x0094
#define XC_NFC_BOOTSTRAP_OVRRIDE	0x0098
#define XC_NFC_IO_TIMING 		0x009C
#define XC_NFC_IO_TIMING2		0x00A0
#define XC_NFC_IO_TIMING3		0x00A4
#define XC_NFC_DUMMY_REG		0x00A8

/*
 * Function Prototypes
 */
static void nand_print(struct nand_chip *nand);
int nand_rw (struct nand_chip* nand, int cmd,
	    size_t start, size_t len,
	    size_t * retlen, u_char * buf);
int nand_erase(struct nand_chip* nand, size_t ofs, size_t len, int clean);
static int nand_read_ecc(struct nand_chip *nand, size_t start, size_t len,
		 size_t * retlen, u_char *buf, unsigned int pagesize);
static int nand_write_ecc (struct nand_chip* nand, size_t to, size_t len,
			   size_t * retlen, const u_char * buf, u_char * ecc_code);
static void nand_print_bad(struct nand_chip *nand);
static int nand_read_oob(struct nand_chip* nand, size_t ofs, size_t len,
		 size_t * retlen, u_char * buf);
static int nand_write_oob(struct nand_chip* nand, size_t ofs, size_t len,
		 size_t * retlen, const u_char * buf);
static int nand_write_yaffs2(struct nand_chip *, size_t, size_t, size_t *, const u_char *, u_char *);
static int nand_read_yaffs2(struct nand_chip *, size_t, size_t, size_t *, u_char *, u_char *);

int nand_rw_skip4 (struct nand_chip* nand, int cmd,
	    size_t start, size_t len,
	    size_t * retlen, u_char * buf);

struct nand_chip nand_dev_desc[CFG_MAX_NAND_DEVICE] = {{0}};

int nand_probed = 0;

/* Current NAND Device	*/
static int curr_device = -1;
static off_t KERNEL_0_OFFSET = 0;
static off_t KERNEL_1_OFFSET = 0;
/* ------------------------------------------------------------------------- */
static inline void get_bus(void)
{
	unsigned int temp, strap_bit;

	temp = readl(XC_SOC_PROC_MMREG_BASE + RBM_PADU_CTRL);
	strap_bit = readl(XC_SOC_PROC_MMREG_BASE + RBM_BOARD_ID) & 0x1;
	if(strap_bit)
		temp &= ~PADU_CTRL_FLASH_MASK;
	else
		temp |= PADU_CTRL_FLASH_MASK;
	writel(temp, XC_SOC_PROC_MMREG_BASE + RBM_PADU_CTRL);
	return;
}

static void disable_hw_wp(void)
{
	unsigned int bid, reg;
	bid = readl(MMREG_BASE + CG_DUMMY_REG);
	if ((bid == 0x1C00)||(bid == 0x1C01)||(bid == 0x1C02)||(bid == 0x1C03)) {
		reg = readl(MMREG_BASE + GPIO_J_OUT);
		writel(reg | (1 << 9), MMREG_BASE + GPIO_J_OUT);
	} 
	return;
}

static void enable_hw_wp(void)
{
	unsigned int bid, reg;
	bid = readl(MMREG_BASE + CG_DUMMY_REG);
	if ((bid == 0x1C00)||(bid == 0x1C01)||(bid == 0x1C02)||(bid == 0x1C03)){
		reg = readl(MMREG_BASE + GPIO_J_OUT);
		writel(reg & ~(1 << 9), MMREG_BASE + GPIO_J_OUT);
	} 
	return;
}

int do_nand (cmd_tbl_t *cmdtp, int flag, int argc, char *argv[])
{
	int ret;

	if(!nand_probed)
		nand_probe(0);

	switch (argc) {
	case 0:
	case 1:
		printf ("Usage:\n%s\n", cmdtp->usage);
		return 1;

	case 2:
		if (strcmp(argv[1],"info") == 0) {
			int i;

			putc ('\n');

			for (i=0; i<CFG_MAX_NAND_DEVICE; ++i) {
				if(nand_dev_desc[i].ChipID == NAND_ChipID_UNKNOWN)
					continue; /* list only known devices */
				printf ("Device %d: ", i);
				nand_print(&nand_dev_desc[i]);
			}
			return 0;

		} else if (strcmp(argv[1],"device") == 0) {
			if ((curr_device < 0) || (curr_device >= CFG_MAX_NAND_DEVICE)) {
				puts ("\nno devices available\n");
				return 1;
			}
			printf ("\nDevice %d: ", curr_device);
			nand_print(&nand_dev_desc[curr_device]);
			return 0;

		} else if (strcmp(argv[1],"bad") == 0) {
			if ((curr_device < 0) || (curr_device >= CFG_MAX_NAND_DEVICE)) {
				puts ("\nno devices available\n");
				return 1;
			}
			printf ("\nDevice %d bad blocks:\n", curr_device);
			nand_print_bad(&nand_dev_desc[curr_device]);
			return 0;

		}
#ifdef CFG_PARTITION_TABLE_IN_NAND
		else if (strcmp(argv[1],"partition_show") == 0) {
extern nand_print_partition_table();
            return nand_print_partition_table();
        }
#endif
		printf ("Usage:\n%s\n", cmdtp->usage);
		return 1;

	case 3:
		if (strcmp(argv[1],"device") == 0) {
			int dev = (int)simple_strtoul(argv[2], NULL, 10);

			printf ("\nDevice %d: ", dev);
			if (dev >= CFG_MAX_NAND_DEVICE) {
				puts ("unknown device\n");
				return 1;
			}
			nand_print(&nand_dev_desc[dev]);
			/*nand_print (dev);*/

			if (nand_dev_desc[dev].ChipID == NAND_ChipID_UNKNOWN) {
				return 1;
			}

			curr_device = dev;

			puts ("... is now current device\n");

			return 0;
		}
		else if (strcmp(argv[1],"erase") == 0 && strcmp(argv[2], "clean") == 0) {
			struct nand_chip* nand = &nand_dev_desc[curr_device];
			ulong off = 0;
			ulong size = nand->totlen;

			printf ("\nNAND erase: device %d offset %ld, size %ld ... ",
				curr_device, off, size);

			ret = nand_erase (nand, off, size, 1);

			printf("%s\n", ret ? "ERROR" : "OK");

			return ret;
		}

		printf ("Usage:\n%s\n", cmdtp->usage);
		return 1;

	default: /* at least 4 args */
		/* read, write */
		if (strncmp(argv[1], "read",4) == 0 ||
		    strncmp(argv[1], "write",5) == 0) {
			ulong addr = simple_strtoul(argv[2], NULL, 16);
			ulong off  = simple_strtoul(argv[3], NULL, 16);
			ulong size = simple_strtoul(argv[4], NULL, 16);
			int cmd    = (strncmp(argv[1], "read", 4) == 0) ?
					NANDRW_READ : NANDRW_WRITE;
			int total;
			char* cmdtail = strchr(argv[1], '.');

			if( cmdtail){
				if( !strncmp(cmdtail, ".oob", 2)) {
					/* read out-of-band data */
					if (cmd & NANDRW_READ) {
						ret = nand_read_oob(nand_dev_desc + curr_device,
							    off, size, &total,
							    (u_char*)addr);
					}
					else {
						ret = nand_write_oob(nand_dev_desc + curr_device,
						     off, size, &total,
						     (u_char*)addr);
					}
					return ret;
				}

				if( !strcmp(cmdtail, ".skip4")){
					ret = nand_rw_skip4(nand_dev_desc + curr_device, cmd, off, size,
								 &total, (u_char*)addr);
					return ret;
				}

				if( !strncmp(cmdtail, ".yaffs2", 2)){
					cmd |= NANDRW_YAFFS2;	/* skip bad blocks (on read too) */
				}
				else{
					printf ("Usage:\n%s\n", cmdtp->usage);
					return 1;
				}
			}
			printf ("\nNAND %s: device %d offset %ld, size %ld ... ",
				(cmd & NANDRW_READ) ? "read" : "write",
				curr_device, off, size);

			ret = nand_rw(nand_dev_desc + curr_device, cmd, off, size,
				     &total, (u_char*)addr);

			printf (" %d bytes %s: %s\n", total,
				(cmd & NANDRW_READ) ? "read" : "written",
				ret ? "ERROR" : "OK");

			return ret;
		}
#ifdef CFG_PARTITION_TABLE_IN_NAND
		else if (strcmp(argv[1],"partition_write") == 0) {
			ulong addr = simple_strtoul(argv[3], NULL, 16);
extern nand_read_partition();
            return nand_write_partition(argv[2], addr);
        }
		else if (strcmp(argv[1],"partition_read") == 0) {
			ulong addr = simple_strtoul(argv[3], NULL, 16);
extern nand_write_partition();
            return nand_read_partition(argv[2], addr);
        }
#endif
		/* erase */
		if (strcmp(argv[1],"erase") == 0 &&
			   (argc == 4 || strcmp("clean", argv[2]) == 0)) {
			int clean = argc == 5;
			ulong start = simple_strtoul(argv[2 + clean], NULL, 16);
			ulong size = simple_strtoul(argv[3 + clean], NULL, 16);
			printf ("\nNAND erase: device %d offset %ld, size %ld ... ",
				curr_device, start, size);

			ret = nand_erase (nand_dev_desc + curr_device, start, size, clean);

			printf("%s\n", ret ? "ERROR" : "OK");

			return ret;
		}

		/* unrecognized command */
		printf ("Usage:\n%s\n", cmdtp->usage);
		return 1;
    }
}

U_BOOT_CMD(
	nand,	5,	1,	do_nand,
	"nand    - NAND sub-system\n",
	"info  - show available NAND devices\n"
	"nand device [dev] - show or set current device\n"
	"nand read addr off size\n"
	"nand write addr off size - read/write 'size' bytes starting\n"
	"    at offset 'off' to/from memory address `addr'\n"
	"nand read.yaffs2 addr off size\n"
	"nand write.yaffs2 addr off size - read/write 'size' bytes starting\n"
	"    at offset 'off' to/from memory address 'addr'. size=data+oob\n"
	"nand erase [clean] [off size] - erase 'size' bytes from\n"
	"    offset 'off' (entire device if not specified)\n"
	"nand bad - show bad blocks\n"
	"nand read.oob addr off size - read out-of-band data\n"
	"nand write.oob addr off size - write out-of-band data\n"
#ifdef CFG_PARTITION_TABLE_IN_NAND
    "nand partition_show - show partition table info\n"
    "nand partition_read name addr - read partition of 'name' to memory 'addr'\n"
    "nand partition_write name addr - write partition of 'name' from memory 'addr'\n"
#endif
);

int do_nandboot (cmd_tbl_t *cmdtp, int flag, int argc, char *argv[])
{
	char *boot_device = NULL;
	char *tmp = NULL;
	char *ep;
	int dev;
	ulong cnt, encrypted = 0;
	ulong current = 2;
	ulong addr;
	ulong offset = 0;
	image_header_t *hdr;
	int rcode = 0;
	struct nand_chip *nand = NULL;

	if(!nand_probed)
		nand_probe(0);

	switch (argc) {
	case 1:
		addr = CFG_LOAD_ADDR;
		boot_device = getenv ("bootdevice");
		tmp = getenv ("current_active");
		break;
	case 2:
		addr = simple_strtoul(argv[1], NULL, 16);
		boot_device = getenv ("bootdevice");
		tmp = getenv ("current_active");
		break;
	case 3:
		addr = simple_strtoul(argv[1], NULL, 16);
		boot_device = argv[2];
		tmp = getenv ("current_active");
		break;
	case 4:
		addr = simple_strtoul(argv[1], NULL, 16);
		boot_device = argv[2];
		offset = simple_strtoul(argv[3], NULL, 16);
		break;
	default:
		printf ("Usage:\n%s\n", cmdtp->usage);
		SHOW_BOOT_PROGRESS (-1);
		return 1;
	}

	if (!boot_device) {
		puts ("\n** No boot device **\n");
		SHOW_BOOT_PROGRESS (-1);
		return 1;
	}
	dev = simple_strtoul(boot_device, &ep, 16);	
	nand = &nand_dev_desc[dev];

	if (tmp != NULL)
		current = simple_strtoul(tmp, NULL, 16);		
	else
		current = 0;	
	debug("Avtive image #%d\n", current);

	/* if no offset provided use current active setup the kernel offset */
	if(offset == 0) {
		if (current == 1)
			offset = KERNEL_1_OFFSET;	//MTDBLOCK 5 - mtd5
		else
			offset = KERNEL_0_OFFSET;	//MTDBLOCK 4 - mtd4
	}	

	if ((dev >= CFG_MAX_NAND_DEVICE) ||
	    (nand->ChipID == NAND_ChipID_UNKNOWN)) {
		printf ("\n** Device %d not available\n", dev);
		SHOW_BOOT_PROGRESS (-1);
		return 1;
	}

	printf ("\nLoading kernel[%d] image from device %d: %s offset: 0x%lx (IO 0x%lx) to 0x%lx\n",
		current, dev, nand->name, offset, nand->IO_ADDR, addr);

	if (nand_rw (nand, NANDRW_READ, offset,
		    nand->pagesize, NULL, (u_char *)addr)) {
		printf ("** Read error on %d\n", dev);
		SHOW_BOOT_PROGRESS (-1);
		return 1;
	}

	hdr = (image_header_t *)addr;

	if (ntohl(hdr->ih_magic) == IH_MAGIC) {

		print_image_hdr (hdr);

		cnt = (ntohl(hdr->ih_size) + sizeof(image_header_t));
		cnt -= nand->pagesize;
		encrypted = 0;
	} else {
		printf("Check for Encrypted kernel\n");
		cnt = sizeof(image_header_t);
		flush_dcache_range(addr, cnt);
		aes_ecb_fb((volatile void*)(addr&DMA_ADDR_MASK), (volatile void*)(CFG_DECRYPT_LOCATION&DMA_ADDR_MASK), cnt, 0);
		inv_dcache_range(CFG_DECRYPT_LOCATION, cnt);		
		hdr = (image_header_t *)CFG_DECRYPT_LOCATION;		
		if (ntohl(hdr->ih_magic) == IH_MAGIC) {			
			cnt = (ntohl(hdr->ih_size) + sizeof(image_header_t));
			if ((cnt%16) != 0)
				cnt += 16 - (cnt%16);
			//cnt -= SECTORSIZE;
			encrypted = 1;
		} else {
		printf ("\n** Bad Magic Number 0x%x **\n", hdr->ih_magic);
		SHOW_BOOT_PROGRESS (-1);
		return 1;
		}
	}

	if (nand_rw (nand, NANDRW_READ, offset + nand->pagesize, cnt,
		    NULL, (u_char *)(addr + nand->pagesize))) {
		printf ("** Read error on %d\n", dev);
		SHOW_BOOT_PROGRESS (-1);
		return 1;
	}

	/* Loading ok, update default load address */

	load_addr = addr;

	/* Check if we should attempt an auto-start */
	if (((ep = getenv("autostart")) != NULL) && (strcmp(ep,"yes") == 0)) {
		char *local_args[2];
		extern int do_bootm (cmd_tbl_t *, int, int, char *[]);

		local_args[0] = argv[0];
		local_args[1] = NULL;

		printf ("Automatic boot of image at addr 0x%08lx ...\n", addr);

		do_bootm (cmdtp, 0, 1, local_args);
		rcode = 1;
	}
	return rcode;
}

U_BOOT_CMD(
	nboot,	4,	1,	do_nandboot,
	"nboot   - boot from NAND device\n",
	"loadAddr dev [offset]\n"
);

/* returns 0 if block containing pos is OK:
 *		valid erase block and
 *		not marked bad, or no bad mark position is specified
 * returns 1 if marked bad or otherwise invalid
 * virt: On interleave mode, we have a virtual block, which combines
 *       2 or 4 chip blocks into one. If 'virt' is set to 1, we check
 *       blocks on each chip. This virtual block is considered a block
 *       if one of the 2 (or 4) blocks is bad.
 *	   2 chips interleave disabled by default.
 */
int check_block (struct nand_chip *nand, unsigned long pos, int virt)
{
	int retlen;
	uint8_t oob_data;
	uint16_t oob_data16[6];
	int page0 = pos & (-nand->erasesize);
	int page1 = page0 + nand->pagesize;
	int badpos = oob_config.badblock_pos;

	if (pos >= nand->totlen)
		return 1;

	if (badpos < 0)
		return 0;	/* no way to check, assume OK */

	if(virt == 0){
		page0 /= nand->para_num;
		page1 /= nand->para_num;
	}

	if (nand->bus16) {
		if (nand_read_oob(nand, (page0 + 0), 12, &retlen, (uint8_t *)oob_data16)
		    || (oob_data16[2] & 0xff00) != 0xff00)
			return 1;
		if (nand_read_oob(nand, (page1 + 0), 12, &retlen, (uint8_t *)oob_data16)
		    || (oob_data16[2] & 0xff00) != 0xff00)
			return 1;
	} else {
		/* Note - bad block marker can be on first or second page */
		if (nand_read_oob(nand, page0 + badpos, 1, &retlen, &oob_data)
		    || oob_data != 0xff
		    || nand_read_oob (nand, page1 + badpos, 1, &retlen, &oob_data)
		    || oob_data != 0xff)
			return 1;
		if(virt){
			int i;
			for(i=2; i<=nand->para_num; i++){
				page0 *= i;
				page1 *= i;
				if (nand_read_oob(nand, page0 + badpos, 1, &retlen, &oob_data)
					|| oob_data != 0xff
					|| nand_read_oob (nand, page1 + badpos, 1, &retlen, &oob_data)
					|| oob_data != 0xff)
					return 1;
			}
		}
	}

	return 0;
}

/* print bad blocks in NAND flash */
static void nand_print_bad(struct nand_chip* nand)
{
	unsigned long pos;

	for (pos = 0; pos < nand->totlen; pos += nand->erasesize) {
		if (check_block(nand, pos, 0))
			printf(" 0x%8.8lx\n", pos);
	}
	puts("\n");
}

int nand_rw (struct nand_chip* nand, int cmd,
	    size_t start, size_t len,
	    size_t * retlen, u_char * buf)
{
	int ret = 0, n, total = 0;
	char eccbuf[6];
	/* eblk (once set) is the start of the erase block containing the
	 * data being processed.
	 */
	/* there is case that env_flash call nand_rw and nand_erase directly*/
	if (curr_device < 0) {
		printf("Error: calling nand_rw while no nand device available\n");
		return 1;
	}
	
	unsigned long eblk = ~0;	/* force mismatch on first pass */
	unsigned long erasesize = nand->erasesize;
	unsigned int pagesize = nand->pagesize;
	if(cmd & NANDRW_ENV){
		/* No interleave on env read/write */
		erasesize /= nand->para_num;
		pagesize /= nand->para_num;
	}

	/* READ/WRITE is performed in 'page' basis. If this page is the first
	 * page of a block, a oob check is necessary. To skip this block or to
	 * perform the operation anyway is determined by the cmd option.
	 */
	while (len) {
		if ((start & (-erasesize)) != eblk) { /* this is a new block */
			eblk = start & (-erasesize); /* offset of this block */
			if(cmd & NANDRW_ENV)
				ret = check_block(nand, eblk, 0);
			else
				ret = check_block(nand, eblk, 1);
			if (ret) { /* is a bad block */
				start += erasesize;
				continue;
			}
		}
#if 0
		if(start != ROUND_DOWN(start, pagesize/nand->para_num)){
			printf("Warning: access %x should be at least 0x%x bytes and start on a 0x%x aligned address\n",
					start, pagesize, pagesize);
//			return 1;
		}
#endif
		if(cmd & NANDRW_YAFFS2){
			unsigned int data_len;
			data_len = (len / (pagesize + nand->oobsize)) * pagesize;
			if (cmd & NANDRW_READ){
				ret = nand_read_yaffs2(nand, start,
							min(data_len, eblk + erasesize - start),
							&n, (u_char*)buf, eccbuf);
			}
			else{
				ret = nand_write_yaffs2(nand, start,
							min(data_len, eblk + erasesize - start),
							&n, (u_char*)buf, eccbuf);
			}
			start += erasesize;
		}
		else{
			if (cmd & NANDRW_READ){
				ret = nand_read_ecc(nand, start,
							min(len, eblk + erasesize - start),
					 	  &n, (u_char*)buf, pagesize);
			}
			else{
				ret = nand_write_ecc(nand, start,
					    min(len, eblk + erasesize - start),
					    &n, (u_char*)buf, pagesize);
			}
			start += n;
		}

		if (ret)
			break;

		buf   += n;
		total += n;
		len   -= n;
	}
	if (retlen)
		*retlen = total;

	return ret;
}

static void nand_print(struct nand_chip *nand)
{
	if (nand->numchips > 1) {
		printf("%s at 0x%lx,\n"
		       "\t  %d chips %s, size %d MB, \n"
		       "\t  total size %ld MB, sector size %ld kB\n",
		       nand->name, nand->IO_ADDR, nand->numchips,
		       nand->chips_name, 1 << (nand->chipshift - 20),
		       nand->totlen >> 20, nand->erasesize >> 10);
	}
	else {
		printf("%s at 0x%lx (", nand->chips_name, nand->IO_ADDR);
		print_size64(nand->totlen, ", ");
		print_size(nand->erasesize, " sector)\n");
	}
}

/* ------------------------------------------------------------------------- */

static void nand_cmdfunc(struct nand_chip *nand, unsigned char command, size_t ofs)
{
	unsigned int current_chip, column, page, temp;
	int do_interleave = 0, bank_shift = 0;

	/* To save the space, we distribute env_copy0~3 on chip0~3,
	 * which can't use interleave read/write. Interleave I/O applies
	 * to kernel and yaffs2
	 */
	get_bus();
 
	if((ofs >= CFG_ENV_OFFSET+CFG_ENV_SIZE*4) && (nand->para_num > 1) && !(nand->para_num%2))		
		do_interleave = 1;

	temp = readl(XC_SOC_PROC_MMREG_BASE + NFC_CONTROL);
	if(do_interleave){
		bank_shift = nand->para_num/2;
		temp |= INTERLEAVE_MASK;
		temp |= bank_shift << BANK_NUM_SHIFT;
		writel(temp, XC_SOC_PROC_MMREG_BASE + NFC_CONTROL);
	}
	else{
		if(temp & INTERLEAVE_MASK){
			temp &= ~INTERLEAVE_MASK;
			temp &= ~BANK_NUM_MASK;
			writel(temp, XC_SOC_PROC_MMREG_BASE + NFC_CONTROL);
		}
		current_chip = (unsigned int)(ofs >> nand->chipshift);
	}

	page = (unsigned int)(ofs >> (nand->page_shift+bank_shift));
	page &= (1 << (nand->chipshift - (nand->page_shift+bank_shift))) - 1;

	column = (unsigned int)(ofs & ((1 << nand->page_shift) - 1));
	/* flush data in memory*/
	flush_dcache_all();

	temp = readl(XC_SOC_PROC_MMREG_BASE + NFC_CMD_STATUS);
	if(temp & NFC_CMD_STATUS_BUSY_MASK)
	{
		/* there is a un-finished command causing the busy bit stuck. clear it before
		    issuing the next command */
		temp = readl(XC_SOC_PROC_MMREG_BASE + NFC_CONTROL);
		temp &= ~NFC_CONTROL_RESETN_MASK;
		writel(temp, XC_SOC_PROC_MMREG_BASE + NFC_CONTROL);
		udelay(1);
		temp |= NFC_CONTROL_RESETN_MASK;
		writel(temp, XC_SOC_PROC_MMREG_BASE + NFC_CONTROL);
		udelay(1);
		writel(0xF, XC_SOC_PROC_MMREG_BASE + NFC_WRITE_PROT);
		udelay(1);
	}

	/*  the command we received is Xcode3 NFC command, not standard flash command*/
	switch (command) {
		case XC_NFC_CMDCODE_READ:
			if(do_interleave){
				/* We always start with bank 0, so start_bank is not assigned */
				writel(page<<bank_shift, XC_SOC_PROC_MMREG_BASE + NFC_PAGE_ROW_ADDR);

				writel(XC_NFC_CMDCODE_READ | (nand->para_num << XC_NFC_CMDPARAM_SHIFT),
						   XC_SOC_PROC_MMREG_BASE + NFC_COMMAND);
			}
			else{
				writel((1 << current_chip) << XC_NFC_BANK_SEL_SHIFT, 
					XC_SOC_PROC_MMREG_BASE + NFC_BANK_COL_ADDR);
				writel(page, XC_SOC_PROC_MMREG_BASE + NFC_PAGE_ROW_ADDR);			
				writel(XC_NFC_CMDCODE_READ + (1 << XC_NFC_CMDPARAM_SHIFT), XC_SOC_PROC_MMREG_BASE + NFC_COMMAND);
			}
			break;
			
		case XC_NFC_CMDCODE_PAGE_PROG:
			disable_hw_wp();
			if(do_interleave){
				/* We always start with bank 0, so start_bank is not assigned */
				writel(page<<bank_shift, XC_SOC_PROC_MMREG_BASE + NFC_PAGE_ROW_ADDR);
				writel(XC_NFC_CMDCODE_PAGE_PROG + (nand->para_num << XC_NFC_CMDPARAM_SHIFT),
					XC_SOC_PROC_MMREG_BASE + NFC_COMMAND);
			}
			else{
				writel((1 << current_chip) << XC_NFC_BANK_SEL_SHIFT, 
					XC_SOC_PROC_MMREG_BASE + NFC_BANK_COL_ADDR);
				writel(page, XC_SOC_PROC_MMREG_BASE + NFC_PAGE_ROW_ADDR);
				writel(XC_NFC_CMDCODE_PAGE_PROG + (1 << XC_NFC_CMDPARAM_SHIFT), XC_SOC_PROC_MMREG_BASE + NFC_COMMAND);
			}
			break;
			
		case XC_NFC_CMDCODE_ERASE:
			disable_hw_wp();
			if(do_interleave){
				/* switch to non-interleave mode to erase banks parallelly */
				temp = readl(XC_SOC_PROC_MMREG_BASE + NFC_CONTROL);
				temp &= ~INTERLEAVE_MASK;
				writel(temp, XC_SOC_PROC_MMREG_BASE + NFC_CONTROL);
				temp = readl(XC_SOC_PROC_MMREG_BASE + NFC_BANK_COL_ADDR);
				temp &= ~NF_BANK_ADDR_MASK;
				if(nand->para_num == 4)
					temp |= (0xf<<NF_BANK_ADDR_SHIFT);
				else // nand->para_num = 2
					temp |= (0x3<<NF_BANK_ADDR_SHIFT);
				writel(temp, XC_SOC_PROC_MMREG_BASE + NFC_BANK_COL_ADDR);
			}
			else{
				temp = readl(XC_SOC_PROC_MMREG_BASE + NFC_CONTROL);
				temp |= (1 << current_chip) << BANK_NUM_SHIFT;
				writel(temp, XC_SOC_PROC_MMREG_BASE + NFC_CONTROL);
				writel((1 << current_chip) << XC_NFC_BANK_SEL_SHIFT, 
				XC_SOC_PROC_MMREG_BASE + NFC_BANK_COL_ADDR);
			}
			writel(page, XC_SOC_PROC_MMREG_BASE + NFC_PAGE_ROW_ADDR);
			writel(XC_NFC_CMDCODE_ERASE , XC_SOC_PROC_MMREG_BASE + NFC_COMMAND);
			break;
			
		case XC_NFC_CMDCODE_READ_STATUS:
			writel(XC_NFC_CMDCODE_READ_STATUS + ((1 << current_chip) << XC_NFC_CMDPARAM_SHIFT), 
				XC_SOC_PROC_MMREG_BASE + NFC_COMMAND);
			break;
			
		case XC_NFC_CMDCODE_READ_ID:
			writel(XC_NFC_CMDCODE_READ_ID + ((1 << current_chip) << XC_NFC_CMDPARAM_SHIFT)
				+ (5 << XC_NFC_CMDPARAM_SHIFT2), 
				XC_SOC_PROC_MMREG_BASE + NFC_COMMAND);
			break;
				
		case XC_NFC_CMDCODE_RESET:
			writel(XC_NFC_CMDCODE_RESET + ((1 << current_chip) << XC_NFC_CMDPARAM_SHIFT), 
				XC_SOC_PROC_MMREG_BASE + NFC_COMMAND);
			break;
			
		default:
			break;
	}

	/* wait until the command is done */
	/* we need to setup a time out in case of no NAND device at all. here just set 10ms, believe it will cover all possibility */
	column = 0;
	do {
		udelay(10);
		column++;
	} while ((readl(XC_SOC_PROC_MMREG_BASE + NFC_INT_STATUS) & 0x10) == 0 && column < 1000);
	writel(0xffffffff, XC_SOC_PROC_MMREG_BASE + NFC_INT_STATUS);
	/* If board support hardware write protection, enable it after command done. Distinguished by  board id */
	if((command == XC_NFC_CMDCODE_ERASE)||(command == XC_NFC_CMDCODE_PAGE_PROG))
		enable_hw_wp();

	/* re-enable cache*/
	inv_dcache_all();

	return;
}


static int nand_read_oob(struct nand_chip *nand, size_t ofs, size_t len,
			 size_t *retlen, u_char *buf)
{
	unsigned int column;
	unsigned int pagesize = nand->pagesize;
	unsigned int oobsize = nand->oobsize;

	if(ofs < CFG_ENV_OFFSET + CFG_ENV_SIZE*4){
		pagesize /= nand->para_num;
		oobsize /= nand->para_num;
	}
	column = (unsigned int)(ofs & (pagesize - 1));
	
	/* Do not allow reads cross page */
	if ((column + len) > (pagesize + oobsize)){		
		printf ("%s: Attempt read oob cross a page %x %x %x %x %x\n",
			__FUNCTION__, (uint) ofs, column, (uint) len, (uint)pagesize, (uint) oobsize);		
		*retlen = 0;
		return -1;
	}

	nand_cmdfunc(nand, XC_NFC_CMDCODE_READ, ofs);
	memcpy(buf, nand->IO_ADDR + nand->pagesize + column, len);

	*retlen = len;
	return 0;
}

static int nand_write_oob(struct nand_chip* nand, size_t ofs, size_t len,
		  size_t * retlen, const u_char * buf)
{
	unsigned int column;
	
	column = (unsigned int)(ofs & (nand->pagesize - 1));
	/* Do not allow reads cross page */
	if ((column + len) > (nand->pagesize + nand->oobsize)) {
		printf ("%s: Attempt write oob cross a page %x %x %x %x %x\n",
			__FUNCTION__, (uint) ofs, column, (uint) len, (uint)nand->pagesize, (uint) nand->oobsize);
		*retlen = 0;
		return -1;
	}

	/* in fact, we are doing read-modify-write back*/
	nand_cmdfunc(nand, XC_NFC_CMDCODE_READ, ofs);
	memcpy(nand->IO_ADDR + nand->pagesize + column, buf, len);
	nand_cmdfunc(nand, XC_NFC_CMDCODE_PAGE_PROG, ofs);

	*retlen = len;
	return 0;
}

static int nand_read_yaffs2(struct nand_chip *nand, size_t start, size_t data_len,
                           size_t *retlen, u_char *buf, u_char *ecc_code)
{
	unsigned int bytes;

	if ((start + data_len) > nand->totlen) {
		printf ("%s: Attempt to read behind the range\n", __FUNCTION__);
		return -1;
	}

	*retlen = 0; /* return the size of data+oob written */
	bytes = 0; /* count the data programmed (no oob) */
	while(bytes < data_len){
		nand_cmdfunc(nand, XC_NFC_CMDCODE_READ, start + bytes);
		memcpy(buf + *retlen, nand->IO_ADDR, nand->pagesize);
		*retlen += nand->pagesize;
		bytes += nand->pagesize;
		memcpy(buf + *retlen, nand->IO_ADDR + nand->pagesize, nand->oobsize);
		*retlen += nand->oobsize;
	}

	return 0;
}

static int nand_read_ecc(struct nand_chip *nand, size_t start, size_t len,
		 size_t * retlen, u_char *buf, unsigned int pagesize)
{
	unsigned int reg, col, stat; 
	/* Do not allow reads past end of device */
	if ((start + len) > nand->totlen) {
		printf ("%s: Attempt read beyond end of device %x %x %x\n",
			__FUNCTION__, (uint) start, (uint) len, (uint) nand->totlen);
		*retlen = 0;
		return -1;
	}

	/* Initialize return value */
	*retlen = 0;
	col = start & (pagesize - 1);

	/* Loop until all data read */
	while (*retlen < len) {
		nand_cmdfunc(nand, XC_NFC_CMDCODE_READ, start + *retlen);

		reg = readl(XC_SOC_PROC_MMREG_BASE + NFC_INT_STATUS);
		if(reg & ECC_ERR_INT_MASK){
			printf("%s: ecc read failed at page 0x%08x\n", __FUNCTION__, (start + *retlen) >> nand->page_shift);
			reg |= ECC_ERR_INT_MASK;
			writel(reg, XC_SOC_PROC_MMREG_BASE + NFC_INT_STATUS);
			return -1;
		}
		/* the old ecc check won't work when the data is all 0xff, such as newly erased */
#if 0
		for ( i = 0 ; i < nand->pagesize / 512; i++ ) {
			stat = readl(XCODE3_NAND_REG_BASE + XC_NFC_ECC_STATUS0 + i*4);
			if ((stat & 0x2)) {
				for (j=0; j < 4; j++) {
					if ( *((unsigned char *)nand->IO_ADDR + nand->pagesize + oob_config.ecc_pos[0] + j + i * 4) != 0xff ) {
						printf ("%s: Failed ECC read, page 0x%08x\n", __FUNCTION__, (start + *retlen) >> nand->page_shift);
						return -1;
						}
				}
				
				//check first 16 bytes of data.
				for (j=0; j < 16; j++) {
					if ( *((unsigned char *)nand->IO_ADDR + j + i * 512) != 0xff ) {
						printf ("%s: Failed ECC read, page 0x%08x\n", __FUNCTION__, (start + *retlen) >> nand->page_shift);
						return -1;
					}
				}
				
			}
		}
#endif		
		if ((*retlen + (pagesize - col)) >= len) {
			memcpy (buf + *retlen, nand->IO_ADDR + col, len - *retlen);
			*retlen = len;
		} else {
			memcpy (buf + *retlen, nand->IO_ADDR + col,  pagesize - col);
			*retlen += pagesize - col;
		}

		col = 0;
	}

	return 0;

}

/* nand_write_yaffs2: program a yaffs2 image to the flash. this image is in a format of
 *                              2048byte data + 64byte oob
 *                              there is no partial write on yaffs2. it's always written in pages
 * nand - the nand chip
 * offset - the nand addr to write to
 * data_len - the data length, not including oob
 * retlen - written length. this is the length of data+oob
 * buf - buffer to read data+oob for programming
 * ecc_code - not used
 *
 * return -1 for any error; otherwise, return 0
 */
static int nand_write_yaffs2(struct nand_chip *nand, size_t offset, size_t data_len,
                           size_t *retlen, const u_char *buf, u_char *ecc_code)
{
	unsigned int pos, bytes;

	/* Do not allow write past end of device */
	if ((offset + data_len) > nand->totlen) {
		printf ("%s: Attempt to write behind the range\n", __FUNCTION__);
		return -1;
	}

	*retlen = 0; /* return the size of data+oob written */
	bytes = 0; /* count the data programmed (no oob) */
	while(bytes < data_len){
		pos = offset + bytes;
		/* data */
		memcpy(nand->IO_ADDR, buf + *retlen, nand->pagesize);
		*retlen += nand->pagesize;
		bytes += nand->pagesize;
		/* oob */
		memcpy(nand->IO_ADDR + nand->pagesize, buf + *retlen, nand->oobsize);
		memset(nand->IO_ADDR + nand->pagesize, 0xff, 2); /* never  modify the bad block mark */
		*retlen += nand->oobsize;
		nand_cmdfunc(nand, XC_NFC_CMDCODE_PAGE_PROG, pos);
	}
		
	return 0;
}

static int nand_write_ecc (struct nand_chip* nand, size_t to, size_t len,
			   size_t * retlen, const u_char * buf, u_char * ecc_code)
{
	unsigned int pos, col;

	/* Do not allow write past end of device */
	if ((to + len) > nand->totlen) {
		printf ("%s: Attempt to write past end of page\n", __FUNCTION__);
		return -1;
	}
	
	/* Initialize return value */
	*retlen = 0;
	col = to & (nand->pagesize - 1);

	/* Loop until all data read */
	while (*retlen < len) {

		pos = to + *retlen;
		/* read data out af first if user write only part of a page*/
		if (col || ((*retlen + (nand->pagesize - col)) > len))
			nand_cmdfunc(nand, XC_NFC_CMDCODE_READ, pos);

		/* Write data into buffer */
		if ((*retlen + (nand->pagesize - col)) > len) {
			memcpy (nand->IO_ADDR + col, buf + *retlen, len - *retlen);
			*retlen = len;
		} else {
			memcpy (nand->IO_ADDR + col, buf + *retlen, nand->pagesize - col);
			*retlen += nand->pagesize - col;
		}

		/* reset oob part */
		memset (nand->IO_ADDR + nand->pagesize, 0xff, nand->oobsize);
		
		nand_cmdfunc(nand, XC_NFC_CMDCODE_PAGE_PROG, pos);		

		col = 0;
	}

	return 0;

}

int nand_rw_skip4 (struct nand_chip* nand, int cmd,
	    size_t start, size_t len,
	    size_t * retlen, u_char * buf)
{
	int ret = 0, n, total = 0;
	char eccbuf[6];

	
	/* there is case that env_flash call nand_rw and nand_erase directly*/
	if (curr_device < 0) {
		printf("Error: calling nand_rw while no nand device available\n");
		return 1;
	}
	
	if (start & (nand->erasesize-1)) {
		printf ("for skip4 operation, start address must be block aligned, , start = %x, blocksize = %x\n",
			 start, (int) nand->erasesize);
		return -1;
	}
	
	printf ("\nNAND %s: device %d offset %ld, size %ld with 4 block skip ... ",
		(cmd & NANDRW_READ) ? "read" : "write",
		curr_device, start, len);
	
	/* eblk (once set) is the start of the erase block containing the
	 * data being processed.
	 */
	
	unsigned long eblk = ~0;	/* force mismatch on first pass */
	unsigned long erasesize = nand->erasesize;

	while (len) {
		if ((start & (-erasesize)) != eblk) {
			/* have crossed into new erase block, deal with
			 * it if it is sure marked bad.
			 */
			eblk = start & (-erasesize); /* start of block */
			/* we don't need to skip bad block here, just comment out to speed up
			if (check_block(nand, eblk)) {
				if (cmd == (NANDRW_READ | NANDRW_JFFS2)) {
					while (len > 0 &&
						   start - eblk < erasesize) {
						*(buf++) = 0xff;
						++start;
						++total;
						--len;
					}
					continue;
				} else if (cmd == (NANDRW_READ | NANDRW_JFFS2 | NANDRW_JFFS2_SKIP)) {
					start += erasesize;
					continue;
				} else if (cmd == (NANDRW_WRITE | NANDRW_JFFS2)) {
					start += erasesize;
					continue;
				} else {
					ret = 1;
					break;
				}
			}*/
		}
		/* The ECC will not be calculated correctly if
		   less than 512 is written or read */
		/* Is request at least 512 bytes AND it starts on a proper boundry */
        /* we are using large device, to be easier, do nothing if start and len isn't on page boundry*/
		if((start != ROUND_DOWN(start, nand->pagesize)) || (len < nand->pagesize))
			printf("Warning block writes should be at least 512 bytes and start on a 512 byte boundry\n");

		if (cmd & NANDRW_READ) {
			ret = nand_read_ecc(nand, start,
					   min(len, eblk + erasesize - start),
					   &n, (u_char*)buf, eccbuf);
		} else {
			ret = nand_write_ecc(nand, start,
						min(len, eblk + erasesize - start),
						&n, (u_char*)buf, eccbuf);
		}

		if (ret)
			break;

		start  += n + 3 * erasesize; /* next block is 4 blocks away*/
		buf   += n;
		total += n;
		len   -= n;
	}
	if (retlen)
		*retlen = total;
	
	
	printf (" %d bytes %s: %s with 4 block skip \n", total,
		(cmd & NANDRW_READ) ? "read" : "written",
		ret ? "ERROR" : "OK");

	return ret;
}


int nand_erase(struct nand_chip* nand, size_t ofs, size_t len, int clean)
{
	/* there is case that env_flash call nand_rw and nand_erase directly*/
	if (curr_device < 0) {
		printf("Error: calling nand_erase while no nand device available\n");
		return 1;
	}


	if (ofs & (nand->erasesize-1) || len & (nand->erasesize-1)) {
		printf ("Offset and size must be sector aligned, , ofs = %x, erasesize = %x\n",
			 ofs, (int) nand->erasesize);
		return -1;
	}

	while(len) {
		/* always check for bad block first, genuine bad blocks
		 * should _never_  be erased.
		 */
		if (ALLOW_ERASE_BAD_DEBUG || !check_block(nand, ofs, 1)) {
			nand_cmdfunc(nand, XC_NFC_CMDCODE_ERASE, ofs);
		}
		else {
			printf("cannot erase 0x%x\n", ofs);
		}
		ofs += nand->erasesize;
		len -= nand->erasesize;
	}

	return 0;
}

static int nand_get_flash_type(struct nand_chip *nand, int chip)
{
	int mfr, id, i;
	int cellinfo;
	unsigned int extid;
	unsigned long temp;

	//Micro MT29F4G08AAC need a reset before issue command. Do so for all chips since we don't know
	//which chip it is before reading ID
	nand_cmdfunc(nand, XC_NFC_CMDCODE_READ_ID, chip * (1 << nand->chipshift) );
	i = readl(XC_SOC_PROC_MMREG_BASE + NFC_ID);
	if (i == 0 || i ==0xffffffff ) {
		return 0;
	}
	nand_cmdfunc(nand, XC_NFC_CMDCODE_RESET, chip * (1 << nand->chipshift) );
	//end of trick for MT29F4G08AAC
		
	nand_cmdfunc(nand, XC_NFC_CMDCODE_READ_ID, chip * (1 << nand->chipshift) );

	i = readl(XC_SOC_PROC_MMREG_BASE + NFC_ID);
	mfr = 0xff & i;
	id = 0xff & (i >> 8);
	cellinfo = (i >> ID_REG3_SHIFT) & 0xff;
	extid = readl(XC_SOC_PROC_MMREG_BASE + NFC_EXTENDED_ID);

	debug("Chip id = %x %x\n", i, extid);
	if (mfr == 0xff || mfr == 0) {
		/* No response - return failure */
		return 0;
	}
	
	/* Check it's the same as the first chip we identified.
	 * M-Systems say that any given nand_chip device should only
	 * contain _one_ type of flash part, although that's not a
	 * hardware restriction. */
	if (nand->mfr) {
		if (nand->mfr == mfr && nand->id == id) {
			return 1;	/* This is another the same the first */
		} else {
			printf("Flash chip at chip %d is different:\n", chip);
			return 0;
		}
	}

	/* Print and store the manufacturer and ID codes. */
	for (i = 0; nand_flash_ids[i].name != NULL; i++) {
		if (mfr == nand_flash_ids[i].manufacture_id &&
			id == nand_flash_ids[i].model_id) {
			if (!nand->mfr) {
				nand->mfr = mfr;
				nand->id = id;
				nand->chipshift =
					nand_flash_ids[i].chipshift;
				nand->erasesize  = nand_flash_ids[i].erasesize;
				nand->chips_name = nand_flash_ids[i].name;
				nand->bus16  = nand_flash_ids[i].bus16;
				nand->page_shift = nand_flash_ids[i].pageshift;
				nand->pagesize = 1 << nand->page_shift; /*page size */
				nand->oobsize = nand_flash_ids[i].oobsize;

				if((cellinfo & 0xc) == 0){ // SLC with 1 bit ECC
					/* SLC uses different IO TIMING */
					writel(0x28000052, XC_SOC_PROC_MMREG_BASE + NFC_IO_TIMING);
					writel(0x20158012, XC_SOC_PROC_MMREG_BASE + NFC_IO_TIMING2);
					writel(0x00000101, XC_SOC_PROC_MMREG_BASE + NFC_IO_TIMING3);
					if((extid & 0x3) == 0) /* internal ECC level */						
						return 1;						
				}

				/* MLC or SLC but internal ECC level requires 4-bit correction: 
				  * use 4 bits error detection/correction
				  * STRAP SLC_MLC = 1 BCH_9BIT = 1
				  */
				writel(0x1, XC_SOC_PROC_MMREG_BASE + NFC_ECC_CONTROL);						
				writel(0x00073447, XC_SOC_PROC_MMREG_BASE + NFC_BCH_CONTROL);
				writel(0xb8418945, XC_SOC_PROC_MMREG_BASE + NFC_BCH_GEN_POLY0);
				writel(0x001aac3a, XC_SOC_PROC_MMREG_BASE + NFC_BCH_GEN_POLY1);
				writel(0x0, XC_SOC_PROC_MMREG_BASE + NFC_BCH_GEN_POLY2);
				writel(0x0, XC_SOC_PROC_MMREG_BASE + NFC_BCH_GEN_POLY3);
				return 1;
			}
			return 0;
		}
	}

	return 0;
}

static void nand_scan_ident(struct nand_chip *nand)
{
	int i = 0;
	unsigned int temp;

	nand->numchips = 0;
	nand->mfr = 0;
	nand->id = 0;

	if (nand_get_flash_type(nand, 0)) {
		for (i=1; i<NAND_MAX_CHIPS; i++) {
			if (nand_get_flash_type(nand, i) == 0) break;
		}
	}

	nand->numchips = i;
	nand->totlen = (unsigned long long)nand->numchips * (1 << nand->chipshift);
}

static void xcode_nand_init_hw(void)
{
	unsigned int temp;
	unsigned int strap_bit;

	/* hard reset */
	get_bus();

    // for low power mode, enable NFC clk
    temp = readl(XC_SOC_PROC_MMREG_BASE + CG1_BLK_CLK_STOP0);
    temp &= ~NFC_MCLK_STOP_MASK;
    temp &= ~NFC_SCLK_STOP_MASK;
    writel(temp, XC_SOC_PROC_MMREG_BASE + CG1_BLK_CLK_STOP0);
    
	temp = readl(XC_SOC_PROC_MMREG_BASE + CG_RESET_REG1);
	temp |= NFC0_RESET_MASK;
	writel(temp, XC_SOC_PROC_MMREG_BASE + CG_RESET_REG1);

	temp = readl(XC_SOC_PROC_MMREG_BASE + CG_RESET_REG1);
	temp &= ~NFC0_RESET_MASK;
	writel(temp, XC_SOC_PROC_MMREG_BASE + CG_RESET_REG1);

	/* soft reset */
	writel(0x00010000, XC_SOC_PROC_MMREG_BASE + NFC_CONTROL);
	udelay(10);
	writel(0x00010001, XC_SOC_PROC_MMREG_BASE + NFC_CONTROL);
	udelay(10);

	/* set IO timing */
	// MLC
	writel(0x38000051, XC_SOC_PROC_MMREG_BASE + NFC_IO_TIMING);
	writel(0x20158392, XC_SOC_PROC_MMREG_BASE + NFC_IO_TIMING2);
	writel(0x00000101, XC_SOC_PROC_MMREG_BASE + NFC_IO_TIMING3);
	// SLC
//	writel(0x28000052, XC_SOC_PROC_MMREG_BASE + NFC_IO_TIMING);
//	writel(0x20158012, XC_SOC_PROC_MMREG_BASE + NFC_IO_TIMING2);
//	writel(0x00000101, XC_SOC_PROC_MMREG_BASE + NFC_IO_TIMING3);
	// 8 chips
//	writel(0x280000F2, XC_SOC_PROC_MMREG_BASE + NFC_IO_TIMING);
//	writel(0x2015B012, XC_SOC_PROC_MMREG_BASE + NFC_IO_TIMING2);
//	writel(0x00000101, XC_SOC_PROC_MMREG_BASE + NFC_IO_TIMING3);

	/* unlock NAND programming */
	writel(0xF, XC_SOC_PROC_MMREG_BASE + NFC_WRITE_PROT);

	/* GPIO may be different on different designs. This may not apply to all */
	temp = readl(XC_SOC_PROC_MMREG_BASE + GPIO_J_OE);
	temp |= 0x3000; //MODAT2, MODAT3
	writel(temp, XC_SOC_PROC_MMREG_BASE + GPIO_J_OE);
	temp = readl(XC_SOC_PROC_MMREG_BASE + GPIO_J_OUT);
	temp |= 0x3000; //MODAT2, MODAT3 high
	writel(temp, XC_SOC_PROC_MMREG_BASE + GPIO_J_OUT);

	/* set bank info: xc4 uses page_size=2048 flash */
	temp = readl(XC_SOC_PROC_MMREG_BASE + NFC_BANK_COL_ADDR);
	temp &= ~NF_BANK_ADDR_MASK;
	temp |= 0x01000000;
	writel(temp, XC_SOC_PROC_MMREG_BASE + NFC_BANK_COL_ADDR);
	temp = readl(XC_SOC_PROC_MMREG_BASE + NFC_BANK_COL_ADDR);
	temp &= ~NF_COL_ADDR_MASK;
	writel(temp, XC_SOC_PROC_MMREG_BASE + NFC_BANK_COL_ADDR);
	
	writel(0x0, XC_SOC_PROC_MMREG_BASE + NFC_BOOTSTRAP_OVRRIDE);

	temp = readl(XC_SOC_PROC_MMREG_BASE + NFC_CONTROL);
	temp &= ~BANK_NUM_MASK;
	writel(temp, XC_SOC_PROC_MMREG_BASE + NFC_CONTROL);

	//disable interleave mode, disable STOP_ON_ERR
	writel(1, XC_SOC_PROC_MMREG_BASE + NFC_CONTROL);

	//enable command done interrupt
	writel(0x10, XC_SOC_PROC_MMREG_BASE + NFC_HOST_INT_MASK);
	return;	
}

static inline void xcode_nand_setup_strap_ovrride(struct nand_chip *nand)
{
	u32 temp, reg;

	debug("chips [%d] pagesize [%d] blocksize [%d]\n", 
				nand->numchips, nand->pagesize, nand->erasesize);

	temp = reg = readl(XC_SOC_PROC_MMREG_BASE + NFC_BOOTSTRAP);
	
	if(reg & BOOT_CTRL_EN_MASK){
		temp |= BOOT_CTRL_EN_OVRRIDE_MASK;
		if(reg & RADDR_CYC_MASK)
			temp |= RADDR_CYC_OVRRIDE_MASK;
		else
			temp &= ~RADDR_CYC_OVRRIDE_MASK;
	} else {
		temp &= ~BOOT_CTRL_EN_OVRRIDE_MASK;
		temp |= RADDR_CYC_OVRRIDE_MASK;
	}

	if(nand->bus16){
		printf("xcode NFC do not support 16 bit bus width\n");
		temp &= ~BUS_WIDTH_OVRRIDE_MASK;
	} else {
		temp &= ~BUS_WIDTH_OVRRIDE_MASK;
	}
	if(nand->pagesize == 0x1000)
		temp |= PAGE_2_4K_OVRRIDE_MASK;
	else
		temp &= ~PAGE_2_4K_OVRRIDE_MASK;

	if(nand->erasesize >= 0x20000)
		temp |= BLOCK_SIZE_OVRRIDE_MASK;
	else
		temp &= ~BLOCK_SIZE_OVRRIDE_MASK;

	if(nand->erasesize /nand->pagesize == 128)
		temp |= PAGE_BLOCK_OVRRIDE_MASK;
	else
		temp &= ~PAGE_BLOCK_OVRRIDE_MASK;

	temp |= BOOTSTRAP_OVRRIDE_EN_MASK;
	debug("NFC over ride reg = %x\n", temp);
	writel(temp, XC_SOC_PROC_MMREG_BASE + NFC_BOOTSTRAP_OVRRIDE);	
	return;
}

unsigned long nand_probe(unsigned long physadr)
{
	struct nand_chip *nand = NULL;
	int i = 0, ChipID = 1;

	nand_probed = 1;

	for (i=0; i<CFG_MAX_NAND_DEVICE; i++) {
		if (nand_dev_desc[i].ChipID == NAND_ChipID_UNKNOWN) {
			nand = &nand_dev_desc[i];
			break;
		}
	}
	if (!nand) 
		return -1;

	memset((char *)nand, 0, sizeof(struct nand_chip));

	xcode_nand_init_hw();

	nand_scan_ident(nand);
	xcode_nand_setup_strap_ovrride(nand);
	
	/* only support 4 chips interleave for SDK board now, 2 chips was disabled */
	if((nand->numchips % 4) == 0){
		nand->para_num = nand->numchips;
		nand->pagesize *= nand->para_num;
		nand->oobsize *= nand->para_num;
		nand->erasesize *= nand->para_num;		
	} else {
		nand->para_num = 1;
	}

	debug("%s: chip# %d paral chips %d pagesize %x erasesize %x oobsize %x\n", __func__, 
		nand->numchips, nand->para_num, nand->pagesize, nand->erasesize, nand->oobsize);

	if (nand->totlen == 0) {
		/* no chips found, clean up and quit */
		puts ("Cannot find NAND chip on board.\n");
		memset((char *)nand, 0, sizeof(struct nand_chip));
		nand->ChipID = NAND_ChipID_UNKNOWN;
		return -1;
	}

	nand->data_buf = malloc(nand->pagesize + nand->oobsize + 16 + 256); //256: max syndrome for 8192 byte page
	if (!nand->data_buf) {
		puts ("Cannot allocate memory for data structures nand->data_buf.\n");
		/* can't allocate data buffer, clean up and quit */
		memset((char *)nand, 0, sizeof(struct nand_chip));
		nand->ChipID = NAND_ChipID_UNKNOWN;
		return -1;
	}

	/* use IO_ADDR for our DMA buffer address, make sure it 16 aligned */
	if ((unsigned int)nand->data_buf & 0xf) 
		nand->IO_ADDR = ((unsigned int)nand->data_buf + 16 ) & (~0xf);
	else
		nand->IO_ADDR = nand->data_buf;
		
    /* set DMA buffer in hardware */
    writel((unsigned long)nand->IO_ADDR, XC_SOC_PROC_MMREG_BASE + NFC_DATA_ADDR);
    writel((unsigned long) (nand->IO_ADDR + nand->pagesize), XC_SOC_PROC_MMREG_BASE + NFC_SPARE_ADDR);
	writel((unsigned long) (nand->IO_ADDR + nand->pagesize + nand->oobsize), XC_SOC_PROC_MMREG_BASE + BCH_SYNDRM_ADDR);
	
	nand->ChipID = ChipID;
	if (curr_device == -1)
		curr_device = i;

	printf ("NAND Device found: Chip type: %s, Total Size: %4lu MB\n", nand->chips_name, nand->totlen >> 20);

	CFG_ENV_OFFSET = (BOOT1_BLKCNT+ BOOT2_BLKCNT*2) * nand->erasesize / nand->para_num;	
	KERNEL_0_OFFSET = (BOOT1_BLKCNT+ BOOT2_BLKCNT*2 + UBOOT_ENV_BLKCNT) * nand->erasesize;
	KERNEL_1_OFFSET = KERNEL_0_OFFSET + KERNEL_SIZE;
	
	debug("KERN 0 offset: 0x%x; KERN 1 offset: 0x%x UBOOT_ENV offset: 0x%x\n", 
			KERNEL_0_OFFSET	, KERNEL_1_OFFSET, CFG_ENV_OFFSET);

	return 0;

}

#endif /* (CONFIG_COMMANDS & CFG_CMD_NAND) */
