/*
 * Driver for the ViXS System Inc USB SOC Device.
 *
 * ViXS System Inc. supported the development of this driver.
 * 
 * CODE STATUS HIGHLIGHTS
 *
 * This driver should work well with the Ethernet/RNDIS gadget drivers
 *
 */

/*drivers/
 * Copyright (C) 2007 Alan Tong
 * Copyright (C) 2007-ViXS System Inc.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#undef	DEBUG		/* messages on error and most fault paths */
#undef	VERBOSE		/* extra debug messages (success too) */

#include <common.h>
#include <asm/errno.h>
#include <linux/list.h>

#include <linux/usb/ch9.h>
#include <linux/usb/gadget.h>

#include <asm/byteorder.h>
#include <asm/io.h>

#include <asm/viper.h>

#include <asm/vixstypes.h>
/* XXX: we are not jaguar anymore. update this header? */
#include <asm/viper_usb.h>

#include <asm/vixs_usbd.h>
#include "ic_debug.h"
#define	DRIVER_DESC         "ViXS System Inc"
#define	DRIVER_AUTHOR       "Alan Tong"
#define	DRIVER_VERSION      "2008 April 21"

#define	DMA_ADDR_INVALID	(~(dma_addr_t)0)

#define XC4_DRV_NAME "xc4-usb"

/* vixs_usbd.h defined DEBUG with other meanings! */
#undef DEBUG

static const char driver_name [] = XC4_DRV_NAME;
static const char driver_desc [] = DRIVER_DESC;

static const char ep0name [] = "ep0";
static const char *const ep_name [] = 
{
	ep0name, "ep1in", "ep2in", "ep3out",
};

static struct vixs	*the_controller = NULL;

extern unsigned long vixs_usbd_skip_enum;

#define writel(b,addr) (*(volatile unsigned int *) (addr) = (b))
#define readl(addr) (*(volatile unsigned int *) (addr))


#define MMR_READ(reg)               readl( (int)((the_controller)->pmmr + (reg)) )
#define MMR_WRITE(reg, data)        writel( (data), ((the_controller)->pmmr + (reg)) )
#define MMFB_READ(off)              readl( ((the_controller)->pmmfb + (off)) )
#define MMFB_WRITE(off, data)       writel( (data), ((the_controller)->pmmfb + (off)) )

#ifdef DEBUG
/* DEBUG the PHY CLK */
#define PRINT_PHY_CLK() \	
do { \
	u32 temp = 0; \
	temp = MMR_READ(XC_USBD_PHY_CLK_CTRL); \
	printf("%s@%d: USBD_PHY_CLK_CTRL 0x%x\n", __func__, __LINE__, temp); \
} while(0);
#else
#define PRINT_PHY_CLK()
#endif



static struct usb_endpoint_descriptor vixs_ep0_desc = 
{
    .bLength =		USB_DT_ENDPOINT_SIZE,
    .bDescriptorType =	USB_DT_ENDPOINT,
    .bEndpointAddress =	0,
    .bmAttributes =		USB_ENDPOINT_XFER_CONTROL,
    .wMaxPacketSize =	USB_MAX_CTRL_PAYLOAD,
};

#define	DIR_STRING(bAddress) (((bAddress) & USB_DIR_IN) ? "in" : "out")
#define ALL_MSG "XC3, "
/* JASON XXX: no kernel header. add it back */
//#define DEBUG_MSG_DISPLAY_LEVEL KERN_INFO
//#define DEBUG_MSG_DISPLAY_LEVEL KERN_ERR
#define DEBUG_MSG_DISPLAY_LEVEL 
//#define DEBUG
#ifdef DEBUG
//#define PDEBUG(_fmt, _args...)		\
//	printf(_fmt, ## _args)

#define PDEBUG printf

#else
#define PDEBUG(_fmt, _args...)
#endif


#if 0
void PDEBUG(char *fmt, ...)
{
    va_list ap;
    char s[256];

    va_start(ap, fmt);

//    vsnprintf(s, sizeof(s), fmt, ap);
//    printk(DEBUG_MSG_DISPLAY_LEVEL ALL_MSG "%s", s);
    printf("%s", s);

    va_end(ap);
}
#endif

static void vixs_fifo_flush (struct usb_ep *_ep);
static void done (struct vixs_ep *ep, struct vixs_request *req, int status);
static void EP0_Send(struct vixs *dev, u16 length, unsigned int* data);
static void Interrupt_Send(struct vixs *dev, u16 length, unsigned int * data);
static int vixs_remove (struct platform_device *pdev);


//-------------------------------------------------------------------------//

//////////////////////////////////////
//
//  ViXS -- EP Reset
//
//////////////////////////////////////
static int vixs_ep_enable (struct usb_ep *_ep, const struct usb_endpoint_descriptor *desc)
{
    struct vixs         *dev;
    struct vixs_ep      *ep;
    u32                 max, tmp;
    unsigned long       flags;
    int                 retval = 0;

    ep = container_of (_ep, struct vixs_ep, ep);

    if (!_ep || !desc || ep->desc || _ep->name == ep0name || desc->bDescriptorType != USB_DT_ENDPOINT)
    {
        PDEBUG("vixs_ep_enable -- ERROR01: EP = %s\n", _ep->name);        
        DBF("_ep:%x desc:%x ep->desc:%x ep0name%s desc->bDescriptorType:%x\n",
		_ep,desc,ep->desc,ep0name, desc->bDescriptorType);
	return -EINVAL;
    }

    dev = ep->dev;

    if (!dev->driver || (dev->gadget.speed == USB_SPEED_UNKNOWN))
    {
        PDEBUG("vixs_ep_enable -- ERROR02: EP = %s\n", _ep->name);
        DBF("dev->driver:%x dev->gadget.speed:%x\n",dev->driver,dev->gadget.speed);
	return -ESHUTDOWN;
    }

    max = le16_to_cpu (desc->wMaxPacketSize);

//    spin_lock_irqsave (&dev->lock, flags);
    ep->ep.maxpacket = max;
    ep->desc = desc;
    ep->stopped = 0;

    // Assume the max packet size from gadget is always correct
    switch (desc->bmAttributes & USB_ENDPOINT_XFERTYPE_MASK)
    {
        case USB_ENDPOINT_XFER_INT:
            if (desc->bEndpointAddress == VIXS_INTERRUPT_IN_ADDR)
            {
                // Enable Interrupt IN Soft Reset
                MMR_WRITE(XC_USBD_SOFT_RST, MMR_READ(XC_USBD_SOFT_RST)&~XC_USBD_SOFT_RST_INTIN0_SOFT_RST_MASK);
            
                // Enable Interrupt In
                tmp = MMR_READ(XC_USBD_INTIN0_CTRL);
                tmp |= XC_USBD_INTIN0_CTRL_ENABLE_MASK;
                tmp |= (0x0F&VIXS_INTERRUPT_IN_ADDR) << XC_USBD_INTIN0_CTRL_EP_NUM_SHIFT;
                tmp |= VIXS_STATUS_ALTER << XC_USBD_INTIN0_CTRL_ALTERNATE_SHIFT;
                tmp |= VIXS_STATUS_INTERFACE << XC_USBD_INTIN0_CTRL_INTERFACE_SHIFT;
                tmp |= VIXS_CONFIG << XC_USBD_INTIN0_CTRL_CONFIG_SHIFT;
                MMR_WRITE(XC_USBD_INTIN0_CTRL, tmp);
                PDEBUG("Vixs_ep_enable:(Interrupt In 0x%lx) %s - max size 0x%lx\n", MMR_READ(XC_USBD_INTIN0_CTRL), _ep->name, max);
            }
            break;
        case USB_ENDPOINT_XFER_BULK:
            if (desc->bEndpointAddress == VIXS_BULK_IN_ADDR)
            {
                // Enable Bulk In Soft Reset
                MMR_WRITE(XC_USBD_SOFT_RST, MMR_READ(XC_USBD_SOFT_RST)&~XC_USBD_SOFT_RST_BULKIN0_SOFT_RST_MASK);

                // Enable Bulk In
                tmp = MMR_READ(XC_USBD_BULKIN0_CTRL);
                tmp |= XC_USBD_BULKIN0_CTRL_ENABLE_MASK;
                tmp |= (0x0F & VIXS_BULK_IN_ADDR) << XC_USBD_BULKIN0_CTRL_EP_NUM_SHIFT;
                tmp |= VIXS_DATA_ALTER << XC_USBD_BULKIN0_CTRL_ALTERNATE_SHIFT;
                tmp |= VIXS_DATA_INTERFACE << XC_USBD_BULKIN0_CTRL_INTERFACE_SHIFT;
                tmp |= VIXS_CONFIG << XC_USBD_BULKIN0_CTRL_CONFIG_SHIFT;
                MMR_WRITE(XC_USBD_BULKIN0_CTRL, tmp);
                PDEBUG("Vixs_ep_enable:(Bulk In) %s - max size 0x%lx\n", _ep->name, max);
            }
            else if (desc->bEndpointAddress == VIXS_BULK_OUT_ADDR)
            {
                // Enable Bulk Out Soft Reset
                MMR_WRITE(XC_USBD_SOFT_RST, MMR_READ(XC_USBD_SOFT_RST)&~XC_USBD_SOFT_RST_BULKOUT0_SOFT_RST_MASK);
            
                // Enable Bulk Out
                tmp = MMR_READ(XC_USBD_BULKOUT0_CTRL);
                tmp |= XC_USBD_BULKOUT0_CTRL_ENABLE_MASK;
                tmp |= (0x0F & VIXS_BULK_OUT_ADDR) << XC_USBD_BULKOUT0_CTRL_EP_NUM_SHIFT;
                tmp |= VIXS_DATA_ALTER << XC_USBD_BULKOUT0_CTRL_ALTERNATE_SHIFT;
                tmp |= VIXS_DATA_INTERFACE << XC_USBD_BULKOUT0_CTRL_INTERFACE_SHIFT;
                tmp |= VIXS_CONFIG << XC_USBD_BULKOUT0_CTRL_CONFIG_SHIFT;
                MMR_WRITE(XC_USBD_BULKOUT0_CTRL, tmp);
                PDEBUG("Vixs_ep_enable:(Bulk Out) %s - max size 0x%lx\n", _ep->name, max);
            }
            break;
        case USB_ENDPOINT_XFER_CONTROL:
            break;
        case USB_ENDPOINT_XFER_ISOC:
            break;
        default:
            goto en_done;
    }

    PDEBUG ("Vixs_ep_enable: EP Name = %s : EP Addr = 0x%lx : Dir = %s : max packet %d\n",
            ep->ep.name, ep->desc->bEndpointAddress,
            (desc->bEndpointAddress & USB_DIR_IN)?"in":"out",
            max);

//    spin_unlock_irqrestore (&dev->lock, flags);

en_done:    
    return retval;
}


static const struct usb_ep_ops vixs_ep_ops;


//////////////////////////////////////
//
//  ViXS -- Reset Endpoint
//
//////////////////////////////////////
static void ep_reset (struct vixs_ep *ep)
{
    if (ep->desc->bEndpointAddress == VIXS_BULK_IN_ADDR)
    {
        // Reset Bulk IN 
        MMR_WRITE(XC_USBD_SOFT_RST, MMR_READ(XC_USBD_SOFT_RST)|XC_USBD_SOFT_RST_BULKIN0_SOFT_RST_MASK);
        MMR_WRITE(XC_USBD_SOFT_RST, MMR_READ(XC_USBD_SOFT_RST)&~XC_USBD_SOFT_RST_BULKIN0_SOFT_RST_MASK);
    }
    else if (ep->desc->bEndpointAddress == VIXS_BULK_OUT_ADDR)
    {
        // Reset Bulk OUT 
        MMR_WRITE(XC_USBD_SOFT_RST, MMR_READ(XC_USBD_SOFT_RST)|XC_USBD_SOFT_RST_BULKOUT0_SOFT_RST_MASK);
        MMR_WRITE(XC_USBD_SOFT_RST, MMR_READ(XC_USBD_SOFT_RST)&~XC_USBD_SOFT_RST_BULKOUT0_SOFT_RST_MASK);
    }
    else if (ep->desc->bEndpointAddress == VIXS_INTERRUPT_IN_ADDR)
    {
        // Reset Interrupt IN 
        MMR_WRITE(XC_USBD_SOFT_RST, MMR_READ(XC_USBD_SOFT_RST)|XC_USBD_SOFT_RST_INTIN0_SOFT_RST_MASK);
        MMR_WRITE(XC_USBD_SOFT_RST, MMR_READ(XC_USBD_SOFT_RST)&~XC_USBD_SOFT_RST_INTIN0_SOFT_RST_MASK);
    }

    ep->desc = NULL;
    INIT_LIST_HEAD (&ep->queue);

    ep->ep.maxpacket = ~0;
    ep->ep.ops = &vixs_ep_ops;
    udelay(2000);
}


//////////////////////////////////////
//
//  ViXS -- Nuke
//
//////////////////////////////////////
static void nuke (struct vixs_ep *ep)
{
    struct vixs_request	*req;

    // called with spinlock held
    ep->stopped = 1;

    // Flush FIFO
    vixs_fifo_flush(&ep->ep);

    // Whether this ep has request linked
    while (!list_empty (&ep->queue)) 
    {
        req = list_entry (ep->queue.next, struct vixs_request, queue);
        done (ep, req, -ESHUTDOWN);
    }
}


//////////////////////////////////////
//
//  ViXS -- Disable
//
//////////////////////////////////////
static int vixs_ep_disable (struct usb_ep *_ep)
{
    struct vixs_ep	*ep;
    unsigned long		flags;

    ep = container_of (_ep, struct vixs_ep, ep);
    if (!_ep || !ep->desc || _ep->name == ep0name)
        return -EINVAL;

    if (ep->desc->bEndpointAddress== VIXS_BULK_IN_ADDR)
    {
        // Disable Bulk IN 
        //MMR_WRITE(XC_USBD_BULKIN0_CTRL, MMR_READ(XC_USBD_BULKIN0_CTRL)&~XC_USBD_BULKIN0_CTRL_ENABLE_MASK);
        //PDEBUG("Vixs_ep_disable:(Bulk In 0x%lx) %s\n", MMR_READ(XC_USBD_INTIN0_CTRL), _ep->name);
    }
    else if (ep->desc->bEndpointAddress == VIXS_BULK_OUT_ADDR)
    {
        // Disable Bulk OUT 
        //MMR_WRITE(XC_USBD_BULKOUT0_CTRL, MMR_READ(XC_USBD_BULKOUT0_CTRL)&~XC_USBD_BULKOUT0_CTRL_ENABLE_MASK);
        //PDEBUG("Vixs_ep_disable:(Bulk Out 0x%lx) %s\n", MMR_READ(XC_USBD_INTIN0_CTRL), _ep->name);
    }
    else if (ep->desc->bEndpointAddress == VIXS_INTERRUPT_IN_ADDR)
    {
        // Disable Interrupt IN 
        //MMR_WRITE(XC_USBD_INTIN0_CTRL, MMR_READ(XC_USBD_INTIN0_CTRL)&~XC_USBD_INTIN0_CTRL_ENABLE_MASK);
        //PDEBUG("Vixs_ep_disable:(Interrupt In 0x%lx) %s\n", MMR_READ(XC_USBD_INTIN0_CTRL), _ep->name);
    }

//    spin_lock_irqsave (&ep->dev->lock, flags);
    nuke (ep);
    ep_reset (ep);

    ep->desc = 0;
    ep->stopped = 1;

//    spin_unlock_irqrestore (&ep->dev->lock, flags);
    return 0;
}

//-------------------------------------------------------------------------//

//////////////////////////////////////
//
//  ViXS -- Allocate Request
//
//////////////////////////////////////
static struct usb_request *vixs_alloc_request (struct usb_ep *_ep)
{
    struct vixs_ep	*ep;
    struct vixs_request	*req = NULL;

    ep = container_of (_ep, struct vixs_ep, ep);

    req = malloc(sizeof(*req));
    if (!req)
    {
        PDEBUG("Vixs_alloc_request: ERROR -- EP %s\n", _ep->name);
        return NULL;
    }

    if (ep)
    {
        PDEBUG("Vixs_alloc_request $$$: %s, Req=0x%lx, Req size=0x%lx\n", _ep->name, req, sizeof *req);
    }
    else
    {
        PDEBUG("Vixs_alloc_request: _ep = NULL\n");
    }

    req->req.dma = DMA_ADDR_INVALID;
    req->insert = 0;
    INIT_LIST_HEAD (&req->queue);

    return &req->req;
}

//////////////////////////////////////
//
//  ViXS -- Free Request
//
//////////////////////////////////////
static void vixs_free_request (struct usb_ep *_ep, struct usb_request *_req)
{
    struct vixs_ep	*ep;
    struct vixs_request	*req;

    ep = container_of (_ep, struct vixs_ep, ep);
    if (!_ep || !_req)
        return;

    req = container_of (_req, struct vixs_request, req);

//    WARN_ON (!list_empty (&req->queue));

    if (ep)
    {
        PDEBUG("Vixs_free_request: EP Name = %s : Req = 0x%lx\n", ep->ep.name, req);
    }
    else
    {
        PDEBUG("Vixs_free_request: _ep = NULL\n");
    }
    
    if (_req)
    {   
	    free (req);
    }
}

//-------------------------------------------------------------------------//

//////////////////////////////////////
//
//  ViXS -- Allocate Buffer
//
//////////////////////////////////////
static void * vixs_alloc_buffer (struct usb_ep	*_ep, unsigned bytes, dma_addr_t *dma, gfp_t gfp_flags)
{
    void			*retval;
    struct vixs_ep	*ep;

    ep = container_of (_ep, struct vixs_ep, ep);
    if ((!_ep)||(!bytes))
    {
        return NULL;
    }

    PDEBUG("Vixs_alloc_buffer: -- Enter\n");
    retval = kmalloc(bytes, gfp_flags);
    if (retval)
    {
        *dma = virt_to_phys(retval);
    }
    PDEBUG("Vixs_alloc_buffer: -- Leave\n");
    return retval;
}

static LIST_HEAD(buffers);


//////////////////////////////////////
//
//  ViXS -- Free Buffer
//
//////////////////////////////////////
static void vixs_free_buffer (struct usb_ep *_ep, void *address, dma_addr_t dma, unsigned bytes) 
{
    // free memory into the right allocator //
    if (address)
    {
        PDEBUG("Vixs_free_buffer: -- Addr = 0x%x\n", address);
        kfree (address);
    }
}

//-------------------------------------------------------------------------//
//////////////////////////////////////////////////////
//
//      ViXS - Write FIFO
//      *** load a packet into the FIFO we use for usb IN transfers.
//
//////////////////////////////////////////////////////
static void write_fifo (struct vixs_ep *ep, struct usb_request *req)
{
    struct vixs *dev;
    u8          *buf;
    //u32         i=0;
    u32         tmp;
    unsigned    count, total, buf_pos=0;
    void        *phy_addr_buf;
    //u32         tmp_buf_idx=0;
    u32         num_of_entry=0, temp=0;
    struct vixs_request *vixs_req;

    // usually actual is equal to zero
    if (req)
    {
        buf = req->buf + req->actual;
        total = req->length - req->actual;
    } 
    else 
    {
        total = 0;
        buf = NULL;
    }
//    printf("%s: total %d \n", __func__, total);

    dev = ep->dev;
                 
    if (ep->desc->bEndpointAddress == VIXS_BULK_IN_ADDR)
    {
        temp = MMR_READ(XC_USBD_BULKIN0_STATUS2);
        num_of_entry = (temp & XC_USBD_BULKIN0_STATUS2_FINISHED_ENTRIES_MASK)>>(XC_USBD_BULKIN0_STATUS2_FINISHED_ENTRIES_SHIFT);
//        PDEBUG("%s: %s - Finished Entries=0x%x, pending req=%d\n", __func__,
	       //ep->ep.name, num_of_entry, dev->pending_in_request);

        // process all the finished entry
        while (num_of_entry)
        {               
            vixs_req = list_entry (ep->queue.next, struct vixs_request, queue);
            if (vixs_req->req.actual == vixs_req->req.length)
            {
                //vixs_req->req.actual = bytes_received;
                done (ep, vixs_req, 0);
                dev->num_of_in_desc --;
                MMR_WRITE(XC_USBD_BULKIN0_STATUS2, 0);
//                PDEBUG("Vixs_irq: %s - Num entry queue up =%d, Finished Entries=0x%x, pending req=%d\n", ep->ep.name, dev->num_of_in_desc, num_of_entry, dev->pending_in_request);
            }
            else
            {
                while(1)                    
                    PDEBUG("Vixs_irq: - BULK IN ERROR : actual 0x%x != length 0x%x\n", vixs_req->req.actual, vixs_req->req.length);
            }
            num_of_entry -= 1;
        }

        count = total;
//        PDEBUG("Write Fifo BULKIN: %s : Data size=%d, Length=%d, Actual=%d, count = %d\n", ep->name, count, req->length, req->actual, count);
        //flush_and_inv_dcache_all();
        flush_dcache_range((unsigned long)req->buf, (unsigned long)req->buf+req->length);
        inv_dcache_range((unsigned long)req->buf, (unsigned long)req->buf+req->length);
	
        // Bulk IN 
        while (count)
        {
// XC4 FPGA
//
//            // Only fill in descriptor when the list is not full
//            tmp = MMR_READ(XC_USBD_BULKIN0_CUR_STATUS);         
//            while ((tmp & XC_USBD_BULKIN0_CUR_STATUS_DESC_FULL_MASK))
//            {
//                tmp = MMR_READ(XC_USBD_BULKIN0_CUR_STATUS);
//                //cpu_relax();
//            }

            // Only fill in descriptor when the list is not full
            tmp = MMR_READ(XC_USBD_BULKIN0_STATUS1);         
            while ((tmp & XC_USBD_BULKIN0_STATUS1_DESC_FULL_MASK))
            {
                tmp = MMR_READ(XC_USBD_BULKIN0_STATUS1);
                //cpu_relax();
            }
 
            phy_addr_buf = (void *)virt_to_phys(req->buf);
            //tmp_buf_idx = dev->in_buffer_index;
            //dev->in_buffer_index = (dev->in_buffer_index+1)%MAX_TRANS_DESC;
            //PDEBUG("Write Fifo: %s : Phy addr = 0x%lx\n", ep->name, phy_addr_buf);
            
            MMR_WRITE(XC_USBD_BULKIN0_DESC_FIFO_ADR_WR, (unsigned int)((unsigned char *)phy_addr_buf+buf_pos));
            
            if (count < MAX_TRANSFER_BYTES)
            {
                tmp = count;
                count -= tmp;
                buf_pos += tmp;                
                req->actual += tmp;
            }
            else
            {
                tmp = MAX_TRANSFER_BYTES;
                count -= tmp;
                buf_pos += tmp; 
                req->actual += tmp;
            }
//            PDEBUG("Write Fifo: %s : Receive %d bytes, bytes left = %d\n", ep->name, tmp, count);            
		

            // Trigger interrupt to the host for the last transaction of this request
	    if (!count)
            {
            //    PDEBUG("Write Fifo: %s : Trigger Interrupt: bytes %d\n", ep->name, tmp);
                tmp |= XC_USBD_BULKIN0_DESC_FIFO_LOAD_INT_EN_MASK;
            }
            
            //PDEBUG("Write Fifo: %s : Req=0x%x, Data size=%d, Length=%d, Actual=%d, buf idx=%d\n", ep->name, req, count, req->length, req->actual, tmp_buf_idx);
            MMR_WRITE(XC_USBD_BULKIN0_DESC_FIFO_LOAD, tmp);
            dev->num_of_in_desc+=1;
        }
    }
    else if (ep->desc->bEndpointAddress == VIXS_BULK_OUT_ADDR)
    {
        count = total;
//        PDEBUG("Write Fifo BULKOUT: %s : Data size=%d, Length=%d, Actual=%d\n", ep->name, count, req->length, req->actual);

        // Bulk OUT
        while (count)
        {
// XC4 FPGA
            // Only fill in descriptor when the list is not full
            //tmp = MMR_READ(XC_USBD_BULKOUT0_CUR_STATUS);         
            //while ((tmp & XC_USBD_BULKOUT0_CUR_STATUS_DESC_FULL_MASK))
            //{
            //    tmp = MMR_READ(XC_USBD_BULKOUT0_CUR_STATUS);
            //    cpu_relax();
            //}

            // Only fill in descriptor when the list is not full
            tmp = MMR_READ(XC_USBD_BULKOUT0_STATUS1);         
            while ((tmp & XC_USBD_BULKOUT0_STATUS1_DESC_FULL_MASK))
            {
                tmp = MMR_READ(XC_USBD_BULKOUT0_STATUS1);
                //cpu_relax();
            }

            tmp = phy_addr_buf = (void *)virt_to_phys(req->buf);
            //PDEBUG("Write Fifo: %s : Phy addr = 0x%lx\n", ep->name, phy_addr_buf);
            //tmp_buf_idx = dev->out_buffer_index;
            //dev->out_buffer_index = (dev->out_buffer_index+1)%MAX_TRANS_DESC;
            MMR_WRITE(XC_USBD_BULKOUT0_DESC_FIFO_ADR_WR, (u32)phy_addr_buf);
            //PDEBUG("Write Fifo: %s : USBD_BULKOUT0_DESC_FIFO_ADR_WR = 0x%lx\n", ep->name, MMR_READ(XC_USBD_BULKOUT0_DESC_FIFO_ADR_WR));

            if (count < MAX_TRANSFER_BYTES)
            {
                tmp = count;
                count -= tmp;
                buf_pos += tmp;                
                req->actual += tmp;
            }
            else
            {
                tmp = MAX_TRANSFER_BYTES;
                count -= tmp;
                buf_pos += tmp; 
                req->actual += tmp;
            }
            //PDEBUG("Write Fifo: %s : Receive %d bytes, bytes left = %d\n", ep->name, tmp, count);            

            // Trigger interrupt to the host for the last transaction of this request
            if (!count)
            {
//		    PDEBUG("Write Fifo: %s : Trigger Interrupt: bytes %d\n", ep->name, tmp);
		    tmp |= XC_USBD_BULKOUT0_DESC_FIFO_LOAD_INT_EN_MASK;
            }
            //flush_and_inv_dcache_all();
            flush_dcache_range((unsigned long)req->buf, (unsigned long)req->buf+req->length);
            inv_dcache_range((unsigned long)req->buf, (unsigned long)req->buf+req->length);

//            PDEBUG("Write Fifo: %s : Req=0x%x, Data size=%d, Length=%d, Actual=%d, tmp 0x%x\n", ep->name, req, count, req->length, req->actual, tmp);            
            MMR_WRITE(XC_USBD_BULKOUT0_DESC_FIFO_LOAD, tmp);
            dev->num_of_out_desc+=1;
        }
    }
    else if (ep->desc->bEndpointAddress == VIXS_CONTROL_ADDR)
    {
        // Control IN 
        EP0_Send(dev, total, req->buf);
        req->actual = total;
    }
    else if (ep->desc->bEndpointAddress == VIXS_INTERRUPT_IN_ADDR)
    {
        // Interrupt IN
        Interrupt_Send(dev, total, req->buf);
        req->actual = total;
    }
}


//////////////////////////////////////////////////////////////////////////
//
//  ViXS -- Done
//  done() - retire a request; caller blocked irqs
//  @status : request status to be set, only works when request is still in progress.
//
//////////////////////////////////////////////////////////////////////////
static void done (struct vixs_ep *ep, struct vixs_request *req, int status)
{
    struct vixs     *dev = NULL;
    unsigned char   stopped = ep->stopped;
    struct list_head *n;
    int i = 0;

    dev = (struct vixs*)ep->dev;

    
    // Removed the request from vixs_ep->queue
    list_del_init (&req->queue);

    // req.status should be set as -EINPROGRESS in ep_queue() 
    if (req->req.status == -EINPROGRESS)
    {  
        req->req.status = status;
    }
    else
    {  
        status = req->req.status;
    }

    if (req->mapped) 
    {
	    /* JASON XXX: disable temporary, dunno we have PCI dma or not */
#if 0
        dma_unmap_single (ep->dev->gadget.dev.parent, req->req.dma, req->req.length, ep->is_in ? PCI_DMA_TODEVICE : PCI_DMA_FROMDEVICE);
        req->req.dma = DMA_ADDR_INVALID;
#endif
	/* unmap the DMA */
	flush_and_inv_dcache_all();

        req->mapped = 0;
    }

    if (status && status != -ESHUTDOWN)
    {   
        //PDEBUG ("Done(1): -- complete %s, req 0x%p, stat %d, len %u/%u\n",
        //        ep->ep.name, &req->req, status,
        //        req->req.actual, req->req.length);
    }        

    // don't modify queue heads during completion callback //
    ep->stopped = 1;
//    spin_unlock (&dev->lock);

    if (req->req.complete)
    {              
        //if ((ep->desc->bEndpointAddress == VIXS_BULK_IN_ADDR)||(ep->desc->bEndpointAddress == VIXS_BULK_OUT_ADDR))        
        //{
        //   PDEBUG ("Done(2): -- complete %s, req 0x%p, buf 0x%p, stat %d, len %u/%u\n",
        //            ep->ep.name, &req->req, req->req.buf, status,
        //            req->req.actual, req->req.length);
        //}  
        req->req.complete (&ep->ep, &req->req);
    }

//    spin_lock (&dev->lock);
    ep->stopped = stopped;
}

//-------------------------------------------------------------------------//

//////////////////////////////////////
//
//  ViXS -- Get_Control_Out_Packet
//
//////////////////////////////////////
void Get_Control_Out_Packet(struct vixs_ep *ep, struct vixs_request *req, struct usb_request *_req)
{
    unsigned int         temp=0, i=0;
    unsigned int         setup_out_data=0;
    unsigned int         bytes_obtain=0;
    unsigned int         bytes_left=0;


retry:
    // EP0 Receive SETUP Data
    temp = MMR_READ (XC_USBD_INT_STATUS);
    if (temp & XC_USBD_INT_STATUS_EP0_INT_EP0_OUT_RCVD_MASK)
    {
        bytes_left = setup_out_data = _req->length;
        //PDEBUG("Get_Control_Out_Packet - EP %s Receive %d bytes Setup Data\n", ep->ep.name, _req->length);
        while (bytes_left)
        {
            if (bytes_left > USB_MAX_EP_FIFO)
            {
                setup_out_data = USB_MAX_EP_FIFO;
            }
            else
            {
                setup_out_data = bytes_left;
            }

            bytes_left -= setup_out_data;
                
            for (i=0; i<setup_out_data/4; i++)
            {
                temp = MMR_READ(XC_USBD_EP0_STATUS); 
                while (temp & XC_USBD_EP0_STATUS_OP_BUSY_MASK)
                {
                    temp = MMR_READ(XC_USBD_EP0_STATUS);             
                    //cpu_relax();
                }
            
                temp = (i%(USB_MAX_CTRL_PAYLOAD/4))*4;
                temp |= 0 << XC_USBD_EP0_CTRL_REG_WOP_SHIFT; //read operation
                temp |= 1 << XC_USBD_EP0_CTRL_REG_CSR_SEL_SHIFT; // EP0
                temp |= 3 << XC_USBD_EP0_CTRL_OP_MODE_SHIFT;       
                MMR_WRITE(XC_USBD_EP0_CTRL, temp);       

                temp = MMR_READ(XC_USBD_EP0_STATUS); 
                while (temp & XC_USBD_EP0_STATUS_OP_BUSY_MASK)
                {
                    temp = MMR_READ(XC_USBD_EP0_STATUS);             
                    //cpu_relax();
                }

                *((unsigned int *)_req->buf+i+(bytes_obtain/4)) = (unsigned int *)MMR_READ(XC_USBD_EP0_RD_DATA); 
                //PDEBUG("Get_Control_Out_Packet - Data = 0x%lx, buf = 0x%lx\n",  MMR_READ(XC_USBD_EP0_RD_DATA), *((unsigned int *)_req->buf+i+(bytes_obtain/4)));
            }
            bytes_obtain += setup_out_data;
            MMR_WRITE(XC_USBD_ACK, XC_USBD_ACK_OUT_DATA_FIFO_EMPTY_ACK_MASK);
            MMR_WRITE (XC_USBD_INT_STATUS, XC_USBD_INT_STATUS_EP0_INT_EP0_OUT_RCVD_MASK);
            //PDEBUG("DONE DONE DONE !!! : (After Reset) XC_USBD_INT_STATUS = 0x%lx\n", MMR_READ (XC_USBD_INT_STATUS));
            if (bytes_left)
            {
                temp = MMR_READ (XC_USBD_INT_STATUS);
                while (!(temp & XC_USBD_INT_STATUS_EP0_INT_EP0_OUT_RCVD_MASK))
                {
                    temp = MMR_READ(XC_USBD_INT_STATUS);             
                    //cpu_relax();
                }
            }
        }
        done (ep, req, 0);            
    }
    else
    {
        //PDEBUG("(No Data) - EP %s Receive 0x%lx bytes Setup Data\n", ep->ep.name, _req->length);
        goto retry;
    }
}

static int vixs_wait_bulk_done(int ep_addr,
			       struct vixs *dev,
			       struct vixs_ep *ep,
			       struct vixs_request *req)
{
	volatile u32 irqstatus;
	volatile u32 temp;
	volatile u32 bulk_status1;
	volatile u32 bulk_status2;
	u32 bytes_recv = 0;

//	printf("%s: start\n", __func__);
	/* XXX: make a timeout */
	while (1) {
		irqstatus = MMR_READ(XC_USBD_INT_STATUS);

//		printf("%s: irqstatus 0x%08x\n", __func__, irqstatus);

		if (ep_addr == VIXS_BULK_IN_ADDR) {
			bulk_status1 = MMR_READ(XC_USBD_BULKIN0_STATUS1);

			if ((irqstatus & 
			     XC_USBD_INT_STATUS_BULKIN0_INT_TRANS_DONE_MASK) ||
			    (bulk_status1 &
			     XC_USBD_BULKIN0_STATUS1_ENTRY_FINISH_MASK)) {
				
				MMR_WRITE(XC_USBD_INT_STATUS,
					  XC_USBD_INT_STATUS_BULKIN0_INT_TRANS_DONE_MASK);
				MMR_WRITE(XC_USBD_BULKIN0_STATUS1,
					  bulk_status1 & ~XC_USBD_BULKIN0_STATUS1_ENTRY_FINISH_MASK);

				bulk_status2 = MMR_READ(XC_USBD_BULKIN0_STATUS2);
				MMR_WRITE(XC_USBD_BULKIN0_STATUS2, 0);
				temp = bulk_status2 >> XC_USBD_BULKIN0_STATUS2_FINISHED_ENTRIES_SHIFT;
				if (temp != 1) {
					PDEBUG("BULKIN0 finished entry is %d\n",
					       temp);
					return (-EIO);
				}

				done(ep, req, 0);
				dev->num_of_in_desc --;
//				PDEBUG("%s: done BULKIN\n", __func__);

				return (0);
			}
	    
		} else if (ep_addr == VIXS_BULK_OUT_ADDR) {
			bulk_status1 = MMR_READ(XC_USBD_BULKOUT0_STATUS1);
			
			if ((irqstatus & 
			     XC_USBD_INT_STATUS_BULKOUT0_INT_TRANS_DONE_MASK) ||
			    (bulk_status1 & 
			     XC_USBD_BULKOUT0_STATUS1_ENTRY_FINISH_MASK)) {

//				MMR_WRITE(XC_USBD_BULKOUT0_STATUS1,
//					  bulk_status1 & ~XC_USBD_BULKOUT0_STATUS1_ENTRY_FINISH_MASK);
				
				bulk_status2 = MMR_READ(XC_USBD_BULKOUT0_STATUS2);

				bytes_recv = bulk_status2 & XC_USBD_BULKOUT0_STATUS2_RCVD_BYTES_MASK;
				temp = (bulk_status2 & XC_USBD_BULKOUT0_STATUS2_FINISHED_ENTRIES_MASK) >> 
					(XC_USBD_BULKOUT0_STATUS2_FINISHED_ENTRIES_SHIFT);
				MMR_WRITE(XC_USBD_BULKOUT0_STATUS2, 0);

				req->req.actual = bytes_recv;

//				PDEBUG("%s: done BULKOUT\n", __func__);

				return (0);
			}
		}

		udelay(100);
	}

        temp = MMR_READ (XC_MIPS3_INTERRUPT);
        temp &= MMR_READ (XC_MIPS3_INTERRUPT_MASK);


}

//////////////////////////////////////
//
//  ViXS -- Queue requests
//
//////////////////////////////////////
static int vixs_queue (struct usb_ep *_ep, struct usb_request *_req, gfp_t gfp_flags)
{
    struct vixs_request *req;
    struct vixs_ep      *ep;
    struct vixs         *dev;
    unsigned long       flags=0;
    unsigned int        descriptor_num=0;
    int r = 0;

    req = container_of (_req, struct vixs_request, req);
    ep = container_of (_ep, struct vixs_ep, ep);

    if (strcmp(ep->ep.name, "ep0") == 0) {
//	    ep->desc->bEndpointAddress = 0;
//	    printf("HACK ep0 %p\n", ep);
	    memset(&(ep->desc->bEndpointAddress), 0, sizeof(u8));
    }

//    if ((ep->desc->bEndpointAddress == VIXS_BULK_IN_ADDR)||(ep->desc->bEndpointAddress == VIXS_BULK_OUT_ADDR)) 

//	    PDEBUG ("vixs_queue(Enter): %s(0x%lx), dir=%d, Req=0x%lx, buf=0x%lx, size=%d\n", ep->ep.name, ep->desc->bEndpointAddress, ep->is_in, req, _req->buf, _req->length);
    

    // catch wrong parameters
#ifdef CONFIG_USB_ETH_RNDIS    
    if (!_req || !_req->complete || !_req->buf || !list_empty (&req->queue))
#else
    if (!_req || !_req->complete || !_req->buf)
#endif        
    {   
        PDEBUG ("vixs_queue: -- (%s) ERROR01: _req = 0x%x\n", ep->ep.name, _req);
        PDEBUG ("vixs_queue: -- (%s) ERROR01: _req->complete = 0x%x\n", ep->ep.name, _req->complete);
        PDEBUG ("vixs_queue: -- (%s) ERROR01: _req->buf = 0x%x\n", ep->ep.name, _req->buf);
        PDEBUG ("vixs_queue: -- (%s) ERROR01: list_empty (&req->queue) = 0x%x\n", ep->ep.name, list_empty (&req->queue));
        return -EINVAL;
    }

    if (!_ep || (!ep->desc && ep->num != 0))
    {   
        PDEBUG ("vixs_queue: -- ERROR02\n");
        return -EINVAL;
    }
    
    dev = ep->dev;
    if (!dev->driver || dev->gadget.speed == USB_SPEED_UNKNOWN)
    {   
        printf("%x %x \n", dev->driver, dev->gadget.speed);
        PDEBUG ("vixs_queue: -- ERROR03\n");
        return -ESHUTDOWN;
    }

    req->ep = ep;
    
//    spin_lock_irqsave (&dev->lock, flags);

    _req->status = -EINPROGRESS;
    _req->actual = 0;

    /* JASON: we always relied on the DMA, let the done() do the unmap work */
    if (_req->dma == DMA_ADDR_INVALID) {
	    req->mapped = 1;
    }

    if (ep->desc->bEndpointAddress == VIXS_BULK_IN_ADDR)
    {
        descriptor_num = dev->num_of_in_desc;
//        PDEBUG ("vixs_queue: -- IN: descriptor_num=%d\n", descriptor_num);
    }
    else if (ep->desc->bEndpointAddress == VIXS_BULK_OUT_ADDR)
    {
        descriptor_num = dev->num_of_out_desc;
//	PDEBUG ("vixs_queue: -- OUT: descriptor_num=%d\n", descriptor_num);
    }
    
    // Start process this i/o queue?
    //if (list_empty (&ep->queue) && !ep->stopped) 
    if (descriptor_num < MAX_TRANS_DESC && !ep->stopped)
    {
        // maybe there's no control data, just status ack //
        if (ep->num == 0 && _req->length == 0) 
        {
            done (ep, req, 0);
            PDEBUG ("Done: EP Name = %s : Status ack\n", ep->ep.name);
            goto done;
        }
        else if (ep->num == 0 && _req->length != 0 && ep->is_in == 0)
        {
            Get_Control_Out_Packet(ep, req, _req);
            goto done;            
        }

        req->insert = 1;
        write_fifo (ep, _req);
    }
    else
    {        
        req->insert = 0;
        
        if (ep->desc->bEndpointAddress == VIXS_BULK_IN_ADDR)
        {
            dev->pending_in_request ++;
            PDEBUG ("vixs_queue: -- Bulk In: Pending request = %d\n", dev->pending_in_request);
        }
        else if (ep->desc->bEndpointAddress == VIXS_BULK_OUT_ADDR)
        {
            dev->pending_out_request ++;
        }
    }

    if (req)
    {
        list_add_tail (&req->queue, &ep->queue);
    }
done:
//    spin_unlock_irqrestore (&dev->lock, flags);



    if (ep->desc->bEndpointAddress == VIXS_BULK_IN_ADDR)
//	||
//	ep->desc->bEndpointAddress == VIXS_BULK_OUT_ADDR)
	    /* Wait the transfer complete */
	    r = vixs_wait_bulk_done(ep->desc->bEndpointAddress,
				    dev, ep, req);


    //if (ep->desc->bEndpointAddress == VIXS_BULK_OUT_ADDR)
	//    PDEBUG ("vixs_queue(Leave): %s(0x%lx), dir=%d, Req=0x%lx, buf=0x%lx, size=%d\n", ep->ep.name, ep->desc->bEndpointAddress, ep->is_in, req, _req->buf, _req->length);

    return 0;
}


//////////////////////////////////////
//
//  ViXS -- Dequeue
//
//////////////////////////////////////
static int vixs_dequeue (struct usb_ep *_ep, struct usb_request *_req)
{
    struct vixs_ep	*ep;
    struct vixs_request	*req;
    unsigned long		flags;
    int			stopped;

    PDEBUG("vixs_dequeue: -- Enter\n");
    ep = container_of (_ep, struct vixs_ep, ep);
    if (!_ep || (!ep->desc && ep->num != 0) || !_req)
        return -EINVAL;

//    spin_lock_irqsave (&ep->dev->lock, flags);
    stopped = ep->stopped;

    ep->stopped = 1;

    // make sure it's still queued on this endpoint 
    list_for_each_entry (req, &ep->queue, queue) 
    {
        if (&req->req == _req)
            break;
    }

    if (&req->req != _req) 
    {
//        spin_unlock_irqrestore (&ep->dev->lock, flags);
        return -EINVAL;
    }

    // queue head may be partially complete. 
    if (ep->queue.next == &req->queue) 
    {
        PDEBUG ("unlink (%s) pio\n", _ep->name);
        done (ep, req, -ECONNRESET);
        req = NULL;
    }
    // The request hasn't been processed //
    else
    {
    }

    if (req)
    {
        done (ep, req, -ECONNRESET);
    }
    ep->stopped = stopped;

//    spin_unlock_irqrestore (&ep->dev->lock, flags);
    PDEBUG("vixs_dequeue: -- Leave\n");    
    return 0;
}

//-------------------------------------------------------------------------//

static int vixs_fifo_status (struct usb_ep *_ep);


//////////////////////////////////////
//
//  ViXS -- Set Halt
//
//////////////////////////////////////
static int set_halt(struct vixs_ep *target_ep)
{
    unsigned int temp=0;
    if (target_ep->desc->bEndpointAddress == VIXS_BULK_IN_ADDR)
    {
        // Bulk IN 
        // Set Halt = 1 / Clear Halt = 0
        temp = MMR_READ(XC_USBD_BULKIN0_CTRL);
        temp &= ~(XC_USBD_BULKIN0_CTRL_ENABLE_MASK);
        MMR_WRITE(XC_USBD_BULKIN0_CTRL, temp);
    }
    else if (target_ep->desc->bEndpointAddress == VIXS_BULK_OUT_ADDR)
    {
        // Bulk OUT 
        // Set Halt = 1 / Clear Halt = 0
        temp = MMR_READ(XC_USBD_BULKOUT0_CTRL);
        temp &= ~(XC_USBD_BULKOUT0_CTRL_ENABLE_MASK);
        MMR_WRITE(XC_USBD_BULKOUT0_CTRL, temp);
    }
    else if (target_ep->desc->bEndpointAddress == VIXS_INTERRUPT_IN_ADDR)    
    {
        // Interrupt IN 
        temp = MMR_READ(XC_USBD_INTIN0_CTRL);
        temp &= ~(XC_USBD_INTIN0_CTRL_ENABLE_MASK);
        MMR_WRITE(XC_USBD_INTIN0_CTRL, temp);
    }
    
    return 0;
}


//////////////////////////////////////
//
//  ViXS -- Clear Halt
//
//////////////////////////////////////
static int clear_halt(struct vixs_ep *target_ep)
{
    unsigned int temp=0;
    if (target_ep->desc->bEndpointAddress == VIXS_BULK_IN_ADDR)
    {
        // Bulk IN 
        temp = MMR_READ(XC_USBD_BULKIN0_CTRL);
        temp |= (XC_USBD_BULKIN0_CTRL_ENABLE_MASK);
        MMR_WRITE(XC_USBD_BULKIN0_CTRL, temp);
    }
    else if (target_ep->desc->bEndpointAddress == VIXS_BULK_OUT_ADDR)
    {
        // Bulk OUT 
        temp = MMR_READ(XC_USBD_BULKOUT0_CTRL);
        temp |= (XC_USBD_BULKOUT0_CTRL_ENABLE_MASK);
        MMR_WRITE(XC_USBD_BULKOUT0_CTRL, temp);
    }
    else if (target_ep->desc->bEndpointAddress == VIXS_INTERRUPT_IN_ADDR)    
    {
        // Interrupt IN 
        temp = MMR_READ(XC_USBD_INTIN0_CTRL);
        temp |= (XC_USBD_INTIN0_CTRL_ENABLE_MASK);
        MMR_WRITE(XC_USBD_INTIN0_CTRL, temp);
    }
    return 0;
}


//////////////////////////////////////
//
//  ViXS -- Set Halt
//
//////////////////////////////////////
static int vixs_set_halt (struct usb_ep *_ep, int value)
{
    struct vixs_ep	*ep;
    unsigned long		flags;
    int			retval = 0;

    PDEBUG("Vixs_set_halt: -- Enter\n");
    ep = container_of (_ep, struct vixs_ep, ep);
    if (!_ep || (!ep->desc && ep->num != 0))
    {
        PDEBUG("Vixs_set_halt: -- Error1\n");
        return -EINVAL;
    }
    
    if (!ep->dev->driver || ep->dev->gadget.speed == USB_SPEED_UNKNOWN)
    {
        PDEBUG("Vixs_set_halt: -- Error2\n");
        return -ESHUTDOWN;
    }

//    spin_lock_irqsave (&ep->dev->lock, flags);
    if (!list_empty (&ep->queue))
    {
        PDEBUG("Vixs_set_halt: -- Error3\n");
        retval = -EAGAIN;
    }
    else if (ep->is_in && value && vixs_fifo_status (_ep) != 0)
    {
        PDEBUG("Vixs_set_halt: -- Error4\n");
        retval = -EAGAIN;
    }
    else 
    {
        // set/clear, then synch memory views with the device //
        if (value) 
        {
            if (ep->num == 0)
            {
                ep->dev->protocol_stall = 1;
            }
            else
            {
                PDEBUG("Vixs_set_halt: -- Set Halt\n");
                set_halt (ep);
            }
        } 
        else
        {
            PDEBUG("Vixs_set_halt: -- Clear Halt\n");
            clear_halt (ep);
        }
    }
//    spin_unlock_irqrestore (&ep->dev->lock, flags);
    PDEBUG("Vixs_set_halt: -- Leave\n");
    return retval;
}


//////////////////////////////////////
//
//  ViXS -- FIFO Flush
//
//////////////////////////////////////
static int vixs_fifo_status (struct usb_ep *_ep)
{
    struct vixs_ep	*ep;
    u32			avail=0;

    PDEBUG("Vixs_fifo_status: -- Enter\n");

    ep = container_of (_ep, struct vixs_ep, ep);
    if (!_ep || (!ep->desc && ep->num != 0))
    {   
        PDEBUG("Vixs_fifo_status: -- Leave: ERROR 1\n");
        return -ENODEV;
    }
    if (!ep->dev->driver || ep->dev->gadget.speed == USB_SPEED_UNKNOWN)
    {   
        PDEBUG("Vixs_fifo_status: -- Leave: ERROR 2\n");
        return -ESHUTDOWN;
    }

    PDEBUG("Vixs_fifo_status: -- Leave\n");

    return avail;
}


//////////////////////////////////////
//
//  ViXS -- FIFO Flush
//
//////////////////////////////////////
static void vixs_fifo_flush (struct usb_ep *_ep)
{
    struct vixs_ep	*ep;

    ep = container_of (_ep, struct vixs_ep, ep);
    if (!_ep || (!ep->desc && ep->num != 0))
        return;
    if (!ep->dev->driver || ep->dev->gadget.speed == USB_SPEED_UNKNOWN)
        return;

    if (ep->desc->bEndpointAddress == VIXS_BULK_IN_ADDR)
    {
        // Reset Bulk IN 
        MMR_WRITE(XC_USBD_SOFT_RST, MMR_READ(XC_USBD_SOFT_RST)|XC_USBD_SOFT_RST_BULKIN0_SOFT_RST_MASK);
        MMR_WRITE(XC_USBD_SOFT_RST, MMR_READ(XC_USBD_SOFT_RST)&~XC_USBD_SOFT_RST_BULKIN0_SOFT_RST_MASK);
    }
    else if (ep->desc->bEndpointAddress == VIXS_BULK_OUT_ADDR)
    {
        // Reset Bulk OUT 
        MMR_WRITE(XC_USBD_SOFT_RST, MMR_READ(XC_USBD_SOFT_RST)|XC_USBD_SOFT_RST_BULKOUT0_SOFT_RST_MASK);
        MMR_WRITE(XC_USBD_SOFT_RST, MMR_READ(XC_USBD_SOFT_RST)&~XC_USBD_SOFT_RST_BULKOUT0_SOFT_RST_MASK);
    }
    else if (ep->desc->bEndpointAddress == VIXS_INTERRUPT_IN_ADDR)
    {
        // Reset Interrupt IN 
        MMR_WRITE(XC_USBD_SOFT_RST, MMR_READ(XC_USBD_SOFT_RST)|XC_USBD_SOFT_RST_INTIN0_SOFT_RST_MASK);
        MMR_WRITE(XC_USBD_SOFT_RST, MMR_READ(XC_USBD_SOFT_RST)&~XC_USBD_SOFT_RST_INTIN0_SOFT_RST_MASK);
    }
}

static const struct usb_ep_ops vixs_ep_ops = {
    .enable         = vixs_ep_enable,
    .disable        = vixs_ep_disable,

    .alloc_request  = vixs_alloc_request,
    .free_request   = vixs_free_request,

//    .alloc_buffer   = vixs_alloc_buffer,
//    .free_buffer    = vixs_free_buffer,

    .queue          = vixs_queue,
    .dequeue        = vixs_dequeue,

    .set_halt       = vixs_set_halt,
    .fifo_status    = vixs_fifo_status,
    .fifo_flush     = vixs_fifo_flush,
};

//-------------------------------------------------------------------------//

//////////////////////////////////////
//
//  ViXS -- Gat Frame
//
//////////////////////////////////////
static int vixs_get_frame (struct usb_gadget *_gadget)
{
    struct vixs		*dev;
    u16			retval=0;

    if (!_gadget)
        return -ENODEV;
    dev = container_of (_gadget, struct vixs, gadget);
    return retval;
}


//////////////////////////////////////
//
//  ViXS -- Wake up
//
//////////////////////////////////////
static int vixs_wakeup (struct usb_gadget *_gadget)
{
    struct vixs		*dev;

    if (!_gadget)
        return 0;
    dev = container_of (_gadget, struct vixs, gadget);

    return 0;
}


//////////////////////////////////////
//
//  ViXS -- Self Power
//
//////////////////////////////////////
static int vixs_set_selfpowered (struct usb_gadget *_gadget, int value)
{
    struct vixs		*dev;

    if (!_gadget)
        return 0;
    dev = container_of (_gadget, struct vixs, gadget);

    return 0;
}

//////////////////////////////////////
//
//  ViXS -- Pullup
//
//////////////////////////////////////
static int vixs_pullup(struct usb_gadget *_gadget, int is_on)
{
    struct vixs  *dev;

    if (!_gadget)
        return -ENODEV;
    dev = container_of (_gadget, struct vixs, gadget);

    return 0;
}

static const struct usb_gadget_ops vixs_ops = {
	.get_frame          = vixs_get_frame,
	.wakeup             = vixs_wakeup,
	.set_selfpowered    = vixs_set_selfpowered,
	.pullup             = vixs_pullup,
};


//////////////////////////////////////
//
//  ViXS -- USB Reinit
//
//////////////////////////////////////
static void usb_reinit (struct vixs *dev)
{
    u32	tmp;
    PDEBUG("Usb_reinit: -- Enter\n");

    dev->num_of_in_desc = 0;
    dev->num_of_out_desc = 0;
    dev->recv_buffer_index = 0;
    dev->out_buffer_index = 0;
    dev->in_buffer_index = 0;
    dev->pending_in_request = 0;    
    dev->pending_out_request = 0;
    
    // basic endpoint init
    for (tmp = 0; tmp < NUM_OF_EP; tmp++) 
    {
        struct vixs_ep	*ep = &dev->ep[tmp];

        //strcpy(ep->name, ep_name[tmp]);
        ep->ep.name = ep_name [tmp];
        ep->dev = dev;
        ep->num = tmp;
        ep->stopped = 0;
        ep->ep.ops = &vixs_ep_ops;
        ep->gadget = &dev->gadget;
        ep->is_in = 1;
        
        if (tmp==0 || tmp == 1) 
        {
            // Control / Interrupt Endpoint
            ep->fifo_size = USB_MAX_CTRL_PAYLOAD;
            dev->ep [tmp].ep.maxpacket = USB_MAX_CTRL_PAYLOAD;
        } 
        else
        {
            // Bulk In and Bulk Out Endpoint
            ep->fifo_size = USB_MAX_BULK_PAYLOAD;
            dev->ep [tmp].ep.maxpacket = USB_MAX_BULK_PAYLOAD;
            if (tmp == 3)
            {
                ep->is_in=0;
            }

        }

        // the queue lists any req for this ep
        INIT_LIST_HEAD (&ep->queue);

        PDEBUG("Usb_reinit: EP Name = %s, EP Loc = %d, EP Max Size = 0x%lx\n", ep->ep.name, tmp, dev->ep [tmp].ep.maxpacket);
        PDEBUG("Usb_reinit: EP = 0x%lx, EP = 0x%lx, ops = 0x%lx\n", ep, ep->ep, ep->ep.ops);

        // gagdet.ep_list used for ep_autoconfig so no ep0
        if (tmp)
            list_add_tail(&ep->ep.ep_list, &dev->gadget.ep_list);
    }

    PDEBUG("Usb_reinit: -- Leave\n");
}


//////////////////////////////////////
//
//  ViXS -- Register USB Gadget Driver
//
//////////////////////////////////////
int usb_gadget_register_driver (struct usb_gadget_driver *driver)
{
    struct vixs		*dev = the_controller;
    int			retval;
    unsigned int       temp=0;

    // insist on high speed support from the driver, since
    // (dev->usb->xcvrdiag & FORCE_FULL_SPEED_MODE)
    // "must not be used in normal operation"    
    
    PDEBUG("@@@@usb_gadget_register_driver -- Driver=0x%x : Enter \n", driver);
    if (!driver
        || driver->speed != USB_SPEED_HIGH
        || !driver->bind
        || !driver->unbind
        || !driver->setup)
    {
        PDEBUG("@@@@usb_gadget_register_driver -- ERROR01: driver = 0x%lx\n", driver);
        PDEBUG("@@@@usb_gadget_register_driver -- ERROR01: driver->speed = 0x%lx, USB_SPEED_HIGH = 0x%lx\n", driver->speed, USB_SPEED_HIGH);
        PDEBUG("@@@@usb_gadget_register_driver -- ERROR01: driver->bind = 0x%lx\n", driver->bind);
        PDEBUG("@@@@usb_gadget_register_driver -- ERROR01: driver->unbind = 0x%lx\n", driver->unbind);
        PDEBUG("@@@@usb_gadget_register_driver -- ERROR01: driver->setup = 0x%lx\n", driver->setup);
        return -EINVAL;
    }

    if (!dev)
    {
        PDEBUG("@@@@usb_gadget_register_driver -- ERROR02: dev (vixs) = 0x%lx\n", dev);
        return -ENODEV;
    }

    if (dev->driver)
    {
        PDEBUG("@@@@usb_gadget_register_driver -- ERROR03: dev->driver = 0x%lx\n", dev->driver);
        return -EBUSY;
    }

    // hook up the driver ... //
    dev->softconnect = 1;
//    driver->driver.bus = NULL;
    dev->driver = driver;
//    dev->gadget.dev.driver = &driver->driver;
    retval = driver->bind (&dev->gadget);
    if (retval) 
    {
	    PDEBUG ("@@@@usb_gadget_register_driver: ERROR bind to driver  %d\n", retval);
//        dev->driver = NULL;
//        dev->gadget.dev.driver = NULL;       
        PDEBUG("@@@@usb_gadget_register_driver -- Leave \n");
        return retval;
    }
    else
    {
//        PDEBUG ("@@@@usb_gadget_register_driver: SUCCESS bind to driver %s --> %d\n", driver->driver.name, retval);
    }

    // Enable the interrupt mask for USBD from MIP4
    //temp = MMR_READ (XC_MIPS3_INTERRUPT_MASK);
    //PDEBUG("@@@@usb_gadget_register_driver : read = XC_MIPS3_INTERRUPT_MASK = 0x%lx\n", temp);
    
    //temp |= XC_MIPS3_INTERRUPT_MASK_USBD_INT_MASK;
   //MMR_WRITE(XC_MIPS3_INTERRUPT_MASK, temp);
    //PDEBUG("@@@@usb_gadget_register_driver : write = XC_MIPS3_INTERRUPT_MASK = 0x%lx\n", temp);

    // Enable the interrupt mask for USBD block
    temp = MMR_READ(XC_USBD_MIPS_INT_MASK);
    //PDEBUG("@@@@usb_gadget_register_driver : read = XC_USBD_MIPS_INT_MASK = 0x%lx\n", temp);
    temp |= (XC_USBD_MIPS_INT_MASK_MIPS_BULKIN0_INT_EN_TRANS_DONE_MASK|
            XC_USBD_MIPS_INT_MASK_MIPS_BULKOUT0_INT_EN_TRANS_DONE_MASK|
            XC_USBD_MIPS_INT_MASK_MIPS_EP0_INT_EN_SETUP_RCVD_MASK|
            XC_USBD_MIPS_INT_MASK_MIPS_INT_EN_SETCONFIGURATION_MASK|
            XC_USBD_MIPS_INT_MASK_MIPS_INT_EN_SETINTERFACE_MASK|
            XC_USBD_MIPS_INT_MASK_MIPS_INT_EN_RESET_MASK|
            XC_USBD_MIPS_INT_MASK_MIPS_INT_EN_SUSPEND_MASK|
            XC_USBD_MIPS_INT_MASK_MIPS_INT_EN_WAKEUP_MASK|
            XC_USBD_MIPS_INT_MASK_MIPS_INT_EN_SPEED_MASK);
    MMR_WRITE(XC_USBD_MIPS_INT_MASK, temp);
    PDEBUG("@@@@usb_gadget_register_driver : write = XC_USBD_MIPS_INT_MASK = 0x%lx\n", temp);

    // Put VBUS out of Reset
    //temp = readl (XC_USBD_PHY_CTRL + VIXS_MMR_BASE);
    //temp |= (XC_USBD_PHY_CTRL_DETECT_VBUS_0_ENABLE_MASK|XC_USBD_PHY_CTRL_DETECT_VBUS_1_ENABLE_MASK);
    //writel(temp, (XC_USBD_PHY_CTRL+VIXS_MMR_BASE));
    //PDEBUG("@@@@usb_gadget_register_driver : XC_USBD_PHY_CTRL(VBUS)=(0x%lx)\n", temp);
    //udelay(1000);

    // Take the Phy out of Suspend state
    //temp = readl (XC_USBD_UDC20_CTRL + VIXS_MMR_BASE);
    //temp |= (XC_USBD_UDC20_CTRL_PHY_SUSPEND_ENABLE_MASK);
    //writel(temp, (XC_USBD_UDC20_CTRL+VIXS_MMR_BASE));
    //PDEBUG("@@@@usb_gadget_register_driver : XC_USBD_UDC20_CTRL(Phys)=(0x%lx)\n", temp);
    //udelay(1000);

    // Enable Interrupt IN Soft Reset
    MMR_WRITE(XC_USBD_SOFT_RST, MMR_READ(XC_USBD_SOFT_RST)&~XC_USBD_SOFT_RST_INTIN0_SOFT_RST_MASK);
            
    // Enable Interrupt In
    temp = MMR_READ(XC_USBD_INTIN0_CTRL);
    temp |= XC_USBD_INTIN0_CTRL_ENABLE_MASK;
    temp |= (0x0F&VIXS_INTERRUPT_IN_ADDR) << XC_USBD_INTIN0_CTRL_EP_NUM_SHIFT;
    temp |= VIXS_STATUS_ALTER << XC_USBD_INTIN0_CTRL_ALTERNATE_SHIFT;
    temp |= VIXS_STATUS_INTERFACE << XC_USBD_INTIN0_CTRL_INTERFACE_SHIFT;
    temp |= VIXS_CONFIG << XC_USBD_INTIN0_CTRL_CONFIG_SHIFT;
    MMR_WRITE(XC_USBD_INTIN0_CTRL, temp);
    udelay(2000);
    PDEBUG("@@@@usb_gadget_register_driver -- Leave\n");
    return 0;
}

EXPORT_SYMBOL (usb_gadget_register_driver);


//////////////////////////////////////
//
//  ViXS -- Reset all EP queues
//
//////////////////////////////////////
static inline int vixs_reset_ep_queue(struct vixs *dev, u8 pipe)
{
    struct vixs_ep *ep = &dev->ep[pipe];

    PDEBUG("vixs_reset_ep_queue: EP = %s -- Enter\n", ep->ep.name);

    if (!ep->name)
    {   
        PDEBUG("vixs_reset_ep_queue: Error1 => EP = %s", ep->ep.name);
        return 0;
    }

    nuke(ep);

    PDEBUG("vixs_reset_ep_queue: -- Leave\n");
    return 0;
}


//////////////////////////////////////
//
//  ViXS -- Reset all EP queues
//
//////////////////////////////////////
static int reset_queues(struct vixs *dev)
{
    u8 pipe;

    PDEBUG("reset_queues: -- Enter\n");

    for (pipe = 0; pipe < NUM_OF_EP; pipe++)
    {
        vixs_reset_ep_queue(dev, pipe);
    }

    // report disconnect; the driver is already quiesced
    if (dev->driver) 
    {
//        spin_unlock (&dev->lock);
        dev->driver->disconnect (&dev->gadget);
//        spin_lock (&dev->lock);
    }

    PDEBUG("reset_queues: -- Leave\n");
    return 0;
}


//////////////////////////////////////
//
//  ViXS -- Stop all activity
//
//////////////////////////////////////
static void stop_activity (struct vixs *dev, struct usb_gadget_driver *driver)
{
    PDEBUG("stop_activity: -- Enter\n");
    // don't disconnect if it's not connected
    if (dev->gadget.speed == USB_SPEED_UNKNOWN)
    {
        PDEBUG("stop_activity: -- Error1\n");
        driver = NULL;
    }
    reset_queues(dev);
    PDEBUG("stop_activity: -- Leave\n");
}


//////////////////////////////////////
//
//  ViXS -- Unregister USB Gadget Driver
//
//////////////////////////////////////
int usb_gadget_unregister_driver (struct usb_gadget_driver *driver)
{
    struct vixs	*dev = the_controller;
    unsigned long	flags;

    PDEBUG ("usb_gadget_unregister_driver: -- Enter\n");

    if (!dev)
    {   
        PDEBUG ("usb_gadget_unregister_driver: -- Leaving : ERROR 1\n");
        return -ENODEV;
    }
    if (!driver || driver != dev->driver || !driver->unbind)
    {   
        PDEBUG ("usb_gadget_unregister_driver: -- Leaving : ERROR 2\n");
        return -EINVAL;
    }

    // in fact, no needed //
    dev->usb_state = USB_STATE_ATTACHED;
    dev->ep0_state = WAIT_FOR_SETUP;
    dev->ep0_dir = 0;

//    spin_lock_irqsave (&dev->lock, flags);
    stop_activity (dev, driver);
//    spin_unlock_irqrestore (&dev->lock, flags);

    vixs_pullup (&dev->gadget, 0);

    driver->unbind (&dev->gadget);
//    dev->gadget.dev.driver = NULL;
    dev->driver = NULL;

//    PDEBUG ("usb_gadget_unregister_driver(%s): -- Leaving\n", driver->driver.name);
    return 0;
}

EXPORT_SYMBOL (usb_gadget_unregister_driver);


//////////////////////////////////////
//
//  ViXS -- Get Endpoint by Address
//
//////////////////////////////////////
static struct vixs_ep *get_ep_by_addr (struct vixs *dev, u16 wIndex)
{
    struct vixs_ep	*ep;

    // Endpoint 0 - Control Endpoint
    if ((wIndex & USB_ENDPOINT_NUMBER_MASK) == 0)
    {
        return &dev->ep [0];
    }
    
    list_for_each_entry (ep, &dev->gadget.ep_list, ep.ep_list) 
    {
        u8	bEndpointAddress;

        if (!ep->desc)
        {
            continue;
        }
        bEndpointAddress = ep->desc->bEndpointAddress;
        if ((wIndex ^ bEndpointAddress) & USB_DIR_IN)
        {
            continue;
        }
        if ((wIndex & 0x0f) == (bEndpointAddress & 0x0f))
        {
            return ep;
        }
    }
    return NULL;
}


//////////////////////////////////////
//
//  ViXS -- Speed IRQ
//
//////////////////////////////////////
static void USBD_speed_irq (struct vixs *dev)
{
    unsigned int    usb_speed=0, tmp=0;;

    // We can determine the device speed now
    usb_speed = MMR_READ(XC_USBD_UDC20_STATUS0);
    usb_speed &= XC_USBD_UDC20_STATUS0_UDC20_ENUM_SPEED_MASK;
    usb_speed = usb_speed >> XC_USBD_UDC20_STATUS0_UDC20_ENUM_SPEED_SHIFT;

    if (usb_speed == VIXS_HIGH_SPEED) 
    {
        // Change Bulk In Max Packet Size to 64bytes
        tmp = MMR_READ(XC_USBD_BULKIN0_CTRL);
        tmp &= ~XC_USBD_BULKIN0_CTRL_PKT_SIZE_MASK;
        tmp |= USB_MAX_BULK_PAYLOAD << XC_USBD_BULKIN0_CTRL_PKT_SIZE_SHIFT;
        MMR_WRITE(XC_USBD_BULKIN0_CTRL, tmp);
    
        // Change Bulk Out Max Packet Size to 64bytes
        tmp = MMR_READ(XC_USBD_BULKOUT0_CTRL);
        tmp &= ~XC_USBD_BULKIN0_CTRL_PKT_SIZE_MASK;
        tmp |= USB_MAX_BULK_PAYLOAD << XC_USBD_BULKOUT0_CTRL_PKT_SIZE_SHIFT;
        MMR_WRITE(XC_USBD_BULKOUT0_CTRL, tmp);

        dev->gadget.speed = USB_SPEED_HIGH;
        PDEBUG(": USB_SPEED_HIGH\n");
    }
    else if (usb_speed == VIXS_FULL_SPEED) 
    {
        // Change Bulk In Max Packet Size to 64bytes
        tmp = MMR_READ(XC_USBD_BULKIN0_CTRL);
        tmp &= ~XC_USBD_BULKIN0_CTRL_PKT_SIZE_MASK;
        tmp |= USB_MAX_BULK_PAYLOAD_FULL << XC_USBD_BULKIN0_CTRL_PKT_SIZE_SHIFT;
        MMR_WRITE(XC_USBD_BULKIN0_CTRL, tmp);
    
        // Change Bulk Out Max Packet Size to 64bytes
        tmp = MMR_READ(XC_USBD_BULKOUT0_CTRL);
        tmp &= ~XC_USBD_BULKIN0_CTRL_PKT_SIZE_MASK;
        tmp |= USB_MAX_BULK_PAYLOAD_FULL << XC_USBD_BULKOUT0_CTRL_PKT_SIZE_SHIFT;
        MMR_WRITE(XC_USBD_BULKOUT0_CTRL, tmp);
    
        dev->gadget.speed = USB_SPEED_FULL;
        PDEBUG(": USB_SPEED_FULL\n");
    }
    else if (usb_speed == VIXS_LOW_SPEED) 
    {
        dev->gadget.speed = USB_SPEED_LOW;
        PDEBUG(": USB_SPEED_LOW\n");
    }
    else
    {
        PDEBUG(": USB_SPEED_UNKNOW\n");
    }
}


//////////////////////////////////////
//
//  ViXS -- Suspend IRQ
//
//////////////////////////////////////
static void USBD_suspend_irq (struct vixs *dev)
{
    //unsigned int temp=0;
    PDEBUG("USBD_suspend_irq -- Enter\n");
    dev->resume_state = dev->usb_state;
    dev->usb_state = USB_STATE_SUSPENDED;

    // report suspend to the driver, serial.c does not support this //
    if (dev->driver->suspend)
	    dev->driver->suspend(&dev->gadget);
}

//////////////////////////////////////
//
//  ViXS -- Wakeup IRQ
//
//////////////////////////////////////
static void USBD_wakeup_irq (struct vixs *dev)
{
    unsigned int pipe=0;
    PDEBUG("USBD_wakeup_irq -- Enter\n");

    usb_reinit (dev);

    dev->usb_state = dev->resume_state;
    dev->resume_state = 0;

    for (pipe = 0; pipe < NUM_OF_EP; pipe++)
    {
        dev->ep[pipe].stopped=0;
    }

    // report resume to the driver, serial.c does not support this //
    if (dev->driver->resume)
        dev->driver->resume(&dev->gadget);
}


//////////////////////////////////////
//
//  ViXS -- Reset IRQ
//
//////////////////////////////////////
static void USBD_reset_irq (struct vixs *dev)
{
    //unsigned int temp=0;
    
	//PDEBUG("USBD_reset_irq -- Enter\n");
    dev->device_address = 0;

    // Clear usb state //
    dev->resume_state = 0;
    dev->ep0_dir = 0;
    dev->ep0_state = WAIT_FOR_SETUP;
    dev->remote_wakeup = 0;	// default to 0 on reset //
    dev->gadget.b_hnp_enable = 0;
    dev->gadget.a_hnp_support = 0;
    dev->gadget.a_alt_hnp_support = 0;
    dev->usb_state = USB_STATE_ATTACHED;
}


//////////////////////////////////////
//
//  ViXS -- Send Data Through EP0
//
//////////////////////////////////////
static void EP0_Send(struct vixs *dev, u16 length, unsigned int* data)
{
    unsigned int temp=0, temp1=0;
    unsigned int i=0;
    unsigned int trigger_send=0;
    unsigned int bytes_left=length;
    unsigned int bytes_sent=0;
    //unsigned int send_bytes=length;
    
    // Submit a zero length packet
    if (!length)
    {
        //PDEBUG("EP0_Send:(Size 0)  Length=%d, Data=0x%lx\n", length, *data);
        temp = (length<<XC_USBD_EP0_CTRL1_EP0_IN_LEN_SHIFT);
        temp |= (USB_MAX_EP_FIFO<<XC_USBD_EP0_CTRL1_EP0_PKT_SIZE_SHIFT);
        temp |= XC_USBD_EP0_CTRL1_EP0_IN_FIFO_EMPTY_INT_EN_MASK;       
        MMR_WRITE(XC_USBD_EP0_CTRL1, temp);

        // Enable the interrupt mask
        temp1 = MMR_READ(XC_USBD_MIPS_INT_MASK);
        temp1 |= XC_USBD_MIPS_INT_MASK_MIPS_EP0_INT_EN_EP0_IN_EMPTY_MASK;
        MMR_WRITE(XC_USBD_MIPS_INT_MASK, temp1);
    
        // Trigger the send
        MMR_WRITE(XC_USBD_ACK, XC_USBD_ACK_IN_DATA_FIFO_FULL_ACK_MASK);
        return;
    }

    //PDEBUG("EP0_Send:  Length=%d, Data=0x%lx, USBD_INT_STATUS=0x%lx\n", length, *data, MMR_READ(XC_USBD_INT_STATUS));

    while (bytes_left)
    {
        if (bytes_left > USB_MAX_EP_FIFO)
        {
            length = USB_MAX_EP_FIFO;
        }
        else
        {
            length= bytes_left;
        }
        bytes_left -= length;

        // Check if FIFO is empty
        temp = MMR_READ(XC_USBD_INT_STATUS);
        while (!(temp & XC_USBD_INT_STATUS_EP0_INT_EP0_IN_EMPTY_MASK))
        {
            temp = MMR_READ(XC_USBD_INT_STATUS);
	    //PDEBUG("EP0_Send: Wait for XC_USBD_INT_STATUS_EP0_INT_EP0_IN_EMPTY_MASK = 0x%lx\r", temp);
            //cpu_relax();
        }
     
        temp = (length<<XC_USBD_EP0_CTRL1_EP0_IN_LEN_SHIFT);
        temp |= (USB_MAX_EP_FIFO<<XC_USBD_EP0_CTRL1_EP0_PKT_SIZE_SHIFT);
        temp |= XC_USBD_EP0_CTRL1_EP0_IN_FIFO_EMPTY_INT_EN_MASK;       
        MMR_WRITE(XC_USBD_EP0_CTRL1, temp);
        //PDEBUG("Transfer Length=%d, Left=%d, XC_USBD_EP0_CTRL1=0x%lx\n", length, bytes_left, MMR_READ(XC_USBD_EP0_CTRL1));  
        // Start put data into the FIFO
        for (i=0; i<(length+3)/4; i++)
        {
            trigger_send = 0;
            temp = MMR_READ(XC_USBD_EP0_STATUS); 
            while (temp & XC_USBD_EP0_STATUS_OP_BUSY_MASK)
            {
                temp = MMR_READ(XC_USBD_EP0_STATUS);             
                //cpu_relax();
            }

            MMR_WRITE(XC_USBD_EP0_WR_DATA, *(data+i+(bytes_sent/4)));
            //PDEBUG("EP0_Send(%d): Length=0x%lx, Data=0x%lx\n", i, length, *(data+i+(bytes_sent/4)));

            temp = (i%(USB_MAX_CTRL_PAYLOAD/4))*4;
            temp |= 1 << XC_USBD_EP0_CTRL_REG_WOP_SHIFT; //write operation
            temp |= 1 << XC_USBD_EP0_CTRL_REG_CSR_SEL_SHIFT;
            temp |= 3 << XC_USBD_EP0_CTRL_OP_MODE_SHIFT;       
            MMR_WRITE(XC_USBD_EP0_CTRL, temp);
            
            // 64-bytes : FIFO Full
            if (i==((USB_MAX_CTRL_PAYLOAD/4)-1))
            {
                trigger_send = 1;               
                // Trigger the send
                MMR_WRITE(XC_USBD_ACK, XC_USBD_ACK_IN_DATA_FIFO_FULL_ACK_MASK);
                //PDEBUG("EP0_Send: Length=%d (Trigger transfer-01), USBD_INT_STATUS=0x%lx\n", length, MMR_READ(XC_USBD_INT_STATUS));        

                if (!bytes_left)
                {
                    //PDEBUG("EP0_Send: Length=%d (Interrupt Enable-01): bytes_left=0x%lx\n\n", length, bytes_left);        
                    // Trigger interrupt on the last chunk of transfer
                    temp1 = MMR_READ(XC_USBD_MIPS_INT_MASK);
                    temp1 |= XC_USBD_MIPS_INT_MASK_MIPS_EP0_INT_EN_EP0_IN_EMPTY_MASK;
                    MMR_WRITE(XC_USBD_MIPS_INT_MASK, temp1);            
                }
                else
                {       
                    // Disable the interrupt if it is not the last chunk of transfer
                    //temp1 = MMR_READ(XC_USBD_MIPS_INT_MASK);
                    //temp1 &= ~XC_USBD_MIPS_INT_MASK_MIPS_EP0_INT_EN_EP0_IN_EMPTY_MASK;
                    //MMR_WRITE(XC_USBD_MIPS_INT_MASK, temp1);            
                    //PDEBUG("EP0_Send: Length=%d (Interrupt Disable-01): bytes_left=%d\n\n", length, bytes_left);
                    MMR_WRITE(XC_USBD_INT_STATUS, XC_USBD_INT_STATUS_EP0_INT_EP0_IN_EMPTY_MASK);
                }
                //udelay(1000);
            }
        }

        bytes_sent += length;

        // In this case, the data size is less than 64 bytes
        if (!trigger_send)
        {
            // Trigger the send
            MMR_WRITE(XC_USBD_ACK, XC_USBD_ACK_IN_DATA_FIFO_FULL_ACK_MASK);
            //PDEBUG("EP0_Send: Length=%d (Trigger transfer-02), USBD_INT_STATUS=0x%lx\n", length, MMR_READ(XC_USBD_INT_STATUS));        

            if (!bytes_left)
            {
                //PDEBUG("EP0_Send: Length=%d (Interrupt Enable-02)\n\n", length);        
                // Trigger interrupt on the last chunk of transfer
                temp1 = MMR_READ(XC_USBD_MIPS_INT_MASK);
                temp1 |= XC_USBD_MIPS_INT_MASK_MIPS_EP0_INT_EN_EP0_IN_EMPTY_MASK;
                MMR_WRITE(XC_USBD_MIPS_INT_MASK, temp1);            
            }
            else
            {
                // Disable the interrupt if it is not the last chunk of transfer
                //temp1 = MMR_READ(XC_USBD_MIPS_INT_MASK);
                //temp1 &= ~XC_USBD_MIPS_INT_MASK_MIPS_EP0_INT_EN_EP0_IN_EMPTY_MASK;
                //MMR_WRITE(XC_USBD_MIPS_INT_MASK, temp1);            
                //PDEBUG("EP0_Send: Length=%d (Interrupt Disable-02)\n\n", length);        
                MMR_WRITE(XC_USBD_INT_STATUS, XC_USBD_INT_STATUS_EP0_INT_EP0_IN_EMPTY_MASK);
            }
        }
    }
}


///////////////////////////////////////////
//
//  ViXS -- Send Data Through Interrupt Endpoint
//
///////////////////////////////////////////
static void Interrupt_Send(struct vixs *dev, u16 length, unsigned int * data)
{
    unsigned int temp=0, temp1=0;
    unsigned int i=0;
    unsigned int trigger_send=0;
    unsigned int bytes_left=length;

    // Submit a zero length packet
    if (!length)
    {
        //PDEBUG("Interrupt_Send:(Size 0)  Length=0x%lx, Data=0x%lx\n", length, *data);
        temp = ((VIXS_INTERRUPT_IN_ADDR&0x0F) << XC_USBD_INTIN0_CTRL_EP_NUM_SHIFT);
        temp |= (length << XC_USBD_INTIN0_CTRL_INTIN0_PKT_LEN_SHIFT);
        temp |= (1 << XC_USBD_INTIN0_CTRL_ENABLE_SHIFT);
        temp |= (1 << XC_USBD_INTIN0_CTRL_INTIN0_EMPTY_INT_EN_SHIFT);
        temp |= VIXS_STATUS_ALTER << XC_USBD_INTIN0_CTRL_ALTERNATE_SHIFT;
        temp |= VIXS_STATUS_INTERFACE << XC_USBD_INTIN0_CTRL_INTERFACE_SHIFT;
        temp |= VIXS_CONFIG << XC_USBD_INTIN0_CTRL_CONFIG_SHIFT;
        MMR_WRITE(XC_USBD_INTIN0_CTRL, temp);

        // Enable the interrupt mask
        temp1 = MMR_READ(XC_USBD_MIPS_INT_MASK);
        temp1 |= XC_USBD_MIPS_INT_MASK_MIPS_INTIN0_INT_EN_EMPTY_MASK;
        MMR_WRITE(XC_USBD_MIPS_INT_MASK, temp1);
    
        // Trigger the send
        MMR_WRITE(XC_USBD_ACK, XC_USBD_ACK_INT_IN_FIFO_FULL_ACK_MASK);
        return;
    }

    //PDEBUG("Interrupt_Send: Length=0x%lx, Data=0x%lx, USBD_INTIN0_CTRL=0x%lx, USBD_INT_STATUS=0x%lx\n", length, *data, MMR_READ(XC_USBD_INTIN0_CTRL), MMR_READ(XC_USBD_INT_STATUS));

    while (bytes_left)
    {
        if (bytes_left > USB_MAX_EP_FIFO)
        {
            length = USB_MAX_EP_FIFO;
        }
        else
        {
            bytes_left = length;
        }
        bytes_left -= length;

        // Check if FIFO is empty
        temp = MMR_READ(XC_USBD_INT_STATUS);
        while (!(temp & XC_USBD_INT_STATUS_INTIN0_INT_EMPTY_MASK))
        {
            temp = MMR_READ(XC_USBD_INT_STATUS);
            //cpu_relax();
        }
        
        temp = (length<<XC_USBD_INTIN0_CTRL_INTIN0_PKT_LEN_SHIFT);
        temp |= (1<<XC_USBD_INTIN0_CTRL_ENABLE_SHIFT);
        temp |= ((0x0F&VIXS_INTERRUPT_IN_ADDR)<<XC_USBD_INTIN0_CTRL_EP_NUM_SHIFT);
        temp |= (1<<XC_USBD_INTIN0_CTRL_INTIN0_EMPTY_INT_EN_SHIFT);
        temp |= VIXS_STATUS_ALTER << XC_USBD_INTIN0_CTRL_ALTERNATE_SHIFT;
        temp |= VIXS_STATUS_INTERFACE << XC_USBD_INTIN0_CTRL_INTERFACE_SHIFT;
        temp |= VIXS_CONFIG << XC_USBD_INTIN0_CTRL_CONFIG_SHIFT;
        
        MMR_WRITE(XC_USBD_INTIN0_CTRL, temp);

        // Start put data into the FIFO
        for (i=0; i<(length+3)/4; i++)
        {
            trigger_send = 0;
            temp = MMR_READ(XC_USBD_EP0_STATUS); 
            while (temp & XC_USBD_EP0_STATUS_OP_BUSY_MASK)
            {
                temp = MMR_READ(XC_USBD_EP0_STATUS);             
                //cpu_relax();
            }

            MMR_WRITE(XC_USBD_EP0_WR_DATA, *(data+i));
            //PDEBUG("Interrupt_Send: Length=0x%lx, Data=0x%lx\n", length, *(data+i));

            temp = (i%(USB_MAX_CTRL_PAYLOAD/4))*4;
            temp |= 1 << XC_USBD_EP0_CTRL_REG_WOP_SHIFT; //write operation
            temp |= 2 << XC_USBD_EP0_CTRL_REG_CSR_SEL_SHIFT;
            temp |= 3 << XC_USBD_EP0_CTRL_OP_MODE_SHIFT;       
            MMR_WRITE(XC_USBD_EP0_CTRL, temp);

            // 64-bytes : FIFO Full
            if (i==(USB_MAX_CTRL_PAYLOAD/4))
            {
                trigger_send = 1;
                // Trigger the send
                MMR_WRITE(XC_USBD_ACK, XC_USBD_ACK_INT_IN_FIFO_FULL_ACK_MASK);
                //PDEBUG("Interrupt_Send: Length=0x%lx (Trigger the transfer), USBD_INT_STATUS=0x%lx\n", length, MMR_READ(XC_USBD_INT_STATUS));        
                
                if (!bytes_left)
                {
                    // Trigger interrupt on the last chunk of transfer
                    //PDEBUG("Interrupt_Send: Length=0x%lx (Interrupt Enable)\n", length);        
                    temp1 = MMR_READ(XC_USBD_MIPS_INT_MASK);
                    temp1 |= XC_USBD_MIPS_INT_MASK_MIPS_INTIN0_INT_EN_EMPTY_MASK;
                    MMR_WRITE(XC_USBD_MIPS_INT_MASK, temp1);            
                }
                else
                {
                    // Disable the interrupt if it is not the last chunk of transfer
                    //PDEBUG("Interrupt_Send: Length=0x%lx (Interrupt Disable)\n", length);        
                    temp1 = MMR_READ(XC_USBD_MIPS_INT_MASK);
                    temp1 &= ~XC_USBD_MIPS_INT_MASK_MIPS_INTIN0_INT_EN_EMPTY_MASK;
                    MMR_WRITE(XC_USBD_MIPS_INT_MASK, temp1);            
                    MMR_WRITE(XC_USBD_INT_STATUS, XC_USBD_INT_STATUS_INTIN0_INT_EMPTY_MASK);
                }                
            }
        }

        // In this case, the data size is less than 64 bytes
        if (!trigger_send)
        {
            // Trigger the send
            MMR_WRITE(XC_USBD_ACK, XC_USBD_ACK_INT_IN_FIFO_FULL_ACK_MASK);
            //PDEBUG("Interrupt_Send: Length=0x%lx (Trigger the transfer), USBD_INT_STATUS=0x%lx\n", length, MMR_READ(XC_USBD_INT_STATUS));        
            
            if (!bytes_left)
            {
                // Trigger interrupt on the last chunk of transfer
                //PDEBUG("Interrupt_Send: Length=0x%lx (Interrupt Enable)\n", length);        
                temp1 = MMR_READ(XC_USBD_MIPS_INT_MASK);
                temp1 |= XC_USBD_MIPS_INT_MASK_MIPS_INTIN0_INT_EN_EMPTY_MASK;
                MMR_WRITE(XC_USBD_MIPS_INT_MASK, temp1);            
            }
            else
            {
                // Disable the interrupt if it is not the last chunk of transfer
                //PDEBUG("Interrupt_Send: Length=0x%lx (Interrupt Disable)\n", length);        
                temp1 = MMR_READ(XC_USBD_MIPS_INT_MASK);
                temp1 &= ~XC_USBD_MIPS_INT_MASK_MIPS_INTIN0_INT_EN_EMPTY_MASK;
                MMR_WRITE(XC_USBD_MIPS_INT_MASK, temp1);            
                MMR_WRITE(XC_USBD_INT_STATUS, XC_USBD_INT_STATUS_INTIN0_INT_EMPTY_MASK);
            }            
        }
    }
}


//////////////////////////////////////
//
//  ViXS -- Set Address
//
//////////////////////////////////////
static void ch9setaddress(struct vixs *dev, u16 value, u16 index, u16 length)
{
    struct vixs_ep *ep;
    struct vixs_request *req = dev->status_req;

    PDEBUG("ch9setaddress: -- Enter\n");

    // Save the new address to device struct //
    dev->device_address = (u8) value;

    // Update usb state //
    dev->usb_state = USB_STATE_ADDRESS;

    dev->ep0_dir = USB_DIR_IN;
    ep = &dev->ep[0];
    dev->ep0_state = WAIT_FOR_OUT_STATUS;

    req->ep = ep;
    req->req.length = 0;
    req->req.status = -EINPROGRESS;
    req->req.actual = 0;
    req->req.complete = NULL;
    EP0_Send(dev, 0, NULL);
    list_add_tail(&req->queue, &ep->queue);

    PDEBUG("ch9setaddress: -- Leave\n");    
}

 
//////////////////////////////////////
//
//  ViXS -- Get Status
//
//////////////////////////////////////
static void ch9getstatus(struct vixs *dev, u8 request_type, u16 value, u16 index, u16 length)
{
    u16 tmp = 0;		

    struct vixs_request *req;
    struct vixs_ep *ep;
    int ep_status = 0;

    PDEBUG("ch9getstatus: -- Enter\n");

    ep = &dev->ep[0];

    if ((request_type & USB_RECIP_MASK) == USB_RECIP_DEVICE)
    {
        // Get device status //
        tmp = 1 << USB_DEVICE_SELF_POWERED;
        tmp |= dev->remote_wakeup << USB_DEVICE_REMOTE_WAKEUP;
    } 
    else if ((request_type & USB_RECIP_MASK) == USB_RECIP_INTERFACE) 
    {
        // Get interface status //
        tmp = 0;
    } 
    else if ((request_type & USB_RECIP_MASK) == USB_RECIP_ENDPOINT) 
    {
        // Get endpoint status //
        struct vixs_ep *target_ep;

        target_ep = get_ep_by_addr(dev, index);

        //stall if endpoint doesn't exist
        if (!target_ep->desc)
        {
            PDEBUG("ch9getstatus: -- ERROR 1\n");
            if ((ep_index(target_ep) == 0x01)&&(ep_is_in(target_ep)==1))
            {
                // Reset Bulk IN 
                if (!(MMR_READ(XC_USBD_BULKIN0_CTRL)&XC_USBD_BULKIN0_CTRL_ENABLE_MASK))
                {
                    ep_status = 1;
                }
            }
            else if ((ep_index(target_ep) == 0x03)&&(ep_is_in(target_ep)==0))
            {
                // Reset Bulk OUT 
                if (!(MMR_READ(XC_USBD_BULKOUT0_CTRL)&XC_USBD_BULKOUT0_CTRL_ENABLE_MASK))
                {
                    ep_status = 1;
                }
            }
            else if ((ep_index(target_ep) == 0x02)&&(ep_is_in(target_ep)==1))
            {
                // Reset Interrupt IN 
                if (!(MMR_READ(XC_USBD_INTIN0_CTRL)&XC_USBD_INTIN0_CTRL_ENABLE_MASK))
                {
                    ep_status = 1;
                }
            }
        }
        // 0: not stalled; 1:stalled 
        tmp = ep_status << USB_ENDPOINT_HALT;
    }

    dev->ep0_dir = USB_DIR_IN;
    
    req = dev->status_req;
    *((u16 *) req->req.buf) = cpu_to_le16(tmp);
    req->ep = ep;
    req->req.length = 2;
    req->req.status = -EINPROGRESS;
    //req->req.actual = 0;
    req->req.actual = 2;
    req->req.complete = NULL;
    EP0_Send(dev, 2, req->req.buf);
    list_add_tail(&req->queue, &ep->queue);

    dev->ep0_state = DATA_STATE_XMIT;

    PDEBUG("ch9getstatus: -- Leave\n");
    return;
}


//////////////////////////////////////
//
//  ViXS -- SETUP IRQ
//
//////////////////////////////////////
static void USBD_setup_irq (struct vixs *dev, struct usb_ctrlrequest *setup)
{
    struct vixs_ep *ep;
    u16 wValue = le16_to_cpu(setup->wValue);
    u16 wIndex = le16_to_cpu(setup->wIndex);
    u16 wLength = le16_to_cpu(setup->wLength);

    // Set the defaul EP0 direction is IN
    dev->ep[0].is_in=1;

    switch (setup->bRequest) 
    {
        // Request that need Data+Status phase from udc //
        case USB_REQ_GET_STATUS:            
		PDEBUG("bRequestType = 0x%x\n", setup->bRequestType);
		PDEBUG("bReqest = USB_REQ_GET_STATUS\n");
		PDEBUG("wValue = 0x%x : wIndex = 0x%x : wLength = 0x%x\n", wValue, wIndex, wLength);
            
            if ((setup->bRequestType & (USB_DIR_IN | USB_TYPE_STANDARD)) != (USB_DIR_IN | USB_TYPE_STANDARD))
            {
                dev->ep[0].is_in=0;
                goto pass_up;
            }
            ch9getstatus(dev, setup->bRequestType, wValue, wIndex, wLength);
            break;

        // Requests that need Status phase from udc //
        case USB_REQ_SET_ADDRESS:
    		PDEBUG(": bRequestType = 0x%x\n", setup->bRequestType);
    		PDEBUG(": bReqest = USB_REQ_SET_ADDRESS\n");
    		PDEBUG(": wValue = 0x%x : wIndex = 0x%x : wLength = 0x%x\n", wValue, wIndex, wLength);
        
            if (setup->bRequestType != (USB_DIR_OUT | USB_TYPE_STANDARD | USB_RECIP_DEVICE))
            {
                break;
            }
            ch9setaddress(dev, wValue, wIndex, wLength);
        break;
        // Handled by udc, no data, status by udc //
        case USB_REQ_CLEAR_FEATURE:
        case USB_REQ_SET_FEATURE:
        {	// status transaction //
            int rc = -EOPNOTSUPP;

            if ((setup->bRequestType & USB_RECIP_MASK) == USB_RECIP_ENDPOINT) 
            {                
		    PDEBUG(": bRequestType = 0x%x (USB_RECIP_ENDPOINT)\n", setup->bRequestType);
		    if (setup->bRequest & USB_REQ_CLEAR_FEATURE)
			    PDEBUG(": bReqest = USB_REQ_CLEAR_FEATURE\n");
		    else 
			    PDEBUG(": bReqest = USB_REQ_SET_FEATURE\n");
		    PDEBUG(": wValue = 0x%x : wIndex = 0x%x : wLength = 0x%x\n", wValue, wIndex, wLength);
                
                ep = get_ep_by_addr(dev, wIndex);

                if (wValue != 0 || wLength != 0)
                {
                    break;
                }

                rc = vixs_set_halt(&ep->ep, (setup->bRequest == USB_REQ_SET_FEATURE) ? 1 : 0);
            } 
            else if ((setup->bRequestType & USB_RECIP_MASK) == USB_RECIP_DEVICE) 
            {                
		    PDEBUG(": bRequestType = 0x%x (USB_RECIP_DEVICE)\n", setup->bRequestType);
		    if (setup->bRequest & USB_REQ_CLEAR_FEATURE)
			    PDEBUG(": bReqest = USB_REQ_CLEAR_FEATURE\n");
		    else 
			    PDEBUG(": bReqest = USB_REQ_SET_FEATURE\n");
		    PDEBUG(": wValue = 0x%x : wIndex = 0x%x : wLength = 0x%x\n", wValue, wIndex, wLength);
                
                if (!dev->gadget.is_otg)
                {
                    break;
                }
                else if (setup->bRequest == USB_DEVICE_B_HNP_ENABLE)
                {
                    dev->gadget.b_hnp_enable = 1;
                }
                else if (setup->bRequest == USB_DEVICE_A_HNP_SUPPORT)
                {
                    dev->gadget.a_hnp_support = 1;
                }
                else if (setup->bRequest == USB_DEVICE_A_ALT_HNP_SUPPORT)
                {
                    dev->gadget.a_alt_hnp_support = 1;
                }
                rc = 0;
            }
            else
            {                
		    PDEBUG(": bRequestType = 0x%x (Unknow)\n", setup->bRequestType);
		    if (setup->bRequest & USB_REQ_CLEAR_FEATURE)
			    PDEBUG(": bReqest = USB_REQ_CLEAR_FEATURE\n");
		    else 
			    PDEBUG(": bReqest = USB_REQ_SET_FEATURE\n");
		    PDEBUG(": wValue = 0x%x : wIndex = 0x%x : wLength = 0x%x\n", wValue, wIndex, wLength);
                
                goto pass_up;
            }

            break;
        }
        // Requests handled by gadget //
        default:
pass_up:
            if (wLength) 
            {
                if ((setup->bRequest != USB_REQ_GET_STATUS) && 
                    (setup->bRequest != USB_REQ_CLEAR_FEATURE) && 
                    (setup->bRequest != USB_REQ_SET_ADDRESS) &&
                    (setup->bRequest != USB_REQ_SET_FEATURE))
                {
                    PDEBUG(": bRequestType = 0x%x\n", setup->bRequestType);
                    PDEBUG(": bReqest = 0x%x\n", setup->bRequest);
                    PDEBUG(": wValue = 0x%x : wIndex = 0x%x : wLength = 0x%x\n", wValue, wIndex, wLength);
                }
                // Data phase from gadget, status phase from udc //
                dev->ep0_dir = (setup->bRequestType & USB_DIR_IN) ? USB_DIR_IN : USB_DIR_OUT;
//                spin_unlock(&dev->lock);

                if (dev->driver->setup(&dev->gadget, setup) < 0)
                {         
                    //ep0stall(dev);
                    PDEBUG("===>>> EP0 - Stall\n");
                }                
//                spin_lock(&dev->lock);

                dev->ep0_state = (setup->bRequestType & USB_DIR_IN) ? DATA_STATE_XMIT : DATA_STATE_RECV;
            } 
            else 
            {
                if ((setup->bRequest != USB_REQ_GET_STATUS) && 
                    (setup->bRequest != USB_REQ_CLEAR_FEATURE) && 
                    (setup->bRequest != USB_REQ_SET_ADDRESS) &&
                    (setup->bRequest != USB_REQ_SET_FEATURE))
                {                  
                    PDEBUG(": bRequestType = 0x%x\n", setup->bRequestType);
                    PDEBUG(": bReqest = 0x%x\n", setup->bRequest);
                    PDEBUG(": wValue = 0x%x : wIndex = 0x%x : wLength = 0x%x\n", wValue, wIndex, wLength);                  
                }
                // No data phase, IN status from gadget //
                dev->ep0_dir = USB_DIR_IN;
//                spin_unlock(&dev->lock);

                if (dev->driver->setup(&dev->gadget, setup) < 0)
                {
                    //ep0stall(dev);
                    PDEBUG(": EP0 -- Stall\n");
                }

//                spin_lock(&dev->lock);
                dev->ep0_state = WAIT_FOR_OUT_STATUS;
            }
            break;
    }   
}


/* JASON: need to rewrite this for uboot */
#if 0
//////////////////////////////////////
//
//  ViXS -- IRQ (Intrrupt Handler)
//
//////////////////////////////////////
static irqreturn_t vixs_irq (int irq, void *_dev)
{
    struct vixs		*dev = _dev;
    struct vixs_ep  *ep;
    struct vixs_request *req;
    struct list_head *next_element;

    unsigned int retval = FALSE;
    unsigned int loopcntl=1;
    unsigned int interruptvalue;
    unsigned int interruptstatus;
    unsigned int temp=0;
    //unsigned int i=0;
    unsigned int bytes_received=0;
    unsigned int num_of_entry=0;
		
//    spin_lock (&dev->lock);

    interruptvalue = MMR_READ (XC_MIPS3_INTERRUPT);
    interruptvalue &= MMR_READ (XC_MIPS3_INTERRUPT_MASK);

    if(interruptvalue == 0)
    {        
//        spin_unlock (&dev->lock);
        return IRQ_NONE;
    }

    retval = TRUE;
    //PDEBUG("\n");
    //PDEBUG("1: @@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@ Enter\n");

    while (loopcntl)
    {
        if (interruptvalue & XC_MIPS3_INTERRUPT_USBD_INT_MASK)
        {
            interruptstatus = MMR_READ(XC_USBD_INT_STATUS);
            interruptstatus &= MMR_READ(XC_USBD_MIPS_INT_MASK);

            // Bulk Out request completed
            if (interruptstatus & XC_USBD_INT_STATUS_BULKOUT0_INT_TRANS_DONE_MASK)
            {
                ep = &dev->ep [3];
                PDEBUG("@@@@@@@@@@@@@@@@@@@@ Vixs_irq: %s Transfer OUT Done -- Enter\n", ep->ep.name);
                // make sure any leftover request state is cleared //
                //PDEBUG("\nVixs_irq: %s - Clear OUT interrupt: num entry queue up =%d\n", ep->ep.name, dev->num_of_out_desc); 
                MMR_WRITE(XC_USBD_INT_STATUS, XC_USBD_INT_STATUS_BULKOUT0_INT_TRANS_DONE_MASK);

                temp = MMR_READ(XC_USBD_BULKOUT0_STATUS2);
                bytes_received = temp & XC_USBD_BULKOUT0_STATUS2_RCVD_BYTES_MASK;
                num_of_entry = (temp & XC_USBD_BULKOUT0_STATUS2_FINISHED_ENTRIES_MASK)>>(XC_USBD_BULKOUT0_STATUS2_FINISHED_ENTRIES_SHIFT);
                MMR_WRITE(XC_USBD_BULKOUT0_STATUS2, 0);
                //PDEBUG("Vixs_irq: %s - Finished Entries=0x%x, Bytes=0x%x\n", ep->ep.name, num_of_entry, bytes_received);

                while (num_of_entry)
                {
                    req = list_entry (ep->queue.next, struct vixs_request, queue);

                    if (req->req.actual == req->req.length)
                    {
                        req->req.actual = bytes_received;
                        //memcpy(req->req.buf, dev->inter_out_buffer[dev->recv_buffer_index], req->req.actual);
                        //dev->recv_buffer_index = (dev->recv_buffer_index+1)%MAX_TRANS_DESC;                       
                        done (ep, req, 0);
                        dev->num_of_out_desc-=1;
                        // Setup another transfer descriptor 
                        if(!list_empty (&ep->queue))
                        {
                            req = list_entry (ep->queue.next, struct vixs_request, queue);
                            next_element = ep->queue.next->next;
                            //PDEBUG("Vixs_irq: %s - Next req=0x%lx, insert=%d\n", ep->ep.name, req, req->insert);
                            while(req->insert==1)
                            {
                                if (!list_empty (&ep->queue))
                                {
                                    req = list_entry (next_element, struct vixs_request, queue);
                                    //PDEBUG("Vixs_irq: %s - Next req=0x%lx, insert=%d\n", ep->ep.name, req, req->insert);
                                    next_element = next_element->next;
                                }
                                else
                                {
                                    // Request List Empty
                                    //PDEBUG("Vixs_irq: - Request List Empty\n");
                                    break;
                                }
                            }
                            //PDEBUG("Vixs_irq: %s - We can setup descriptor for req=0x%lx\n", ep->ep.name, req);                            
                            req->insert = 1;
                            write_fifo(ep, &req->req);
                        }
                    }
                    else
                    {
                        while(1)                        
                            PDEBUG("Vixs_irq: - BULK OUT ERROR : actual 0x%x != length 0x%x\n", req->req.actual, req->req.length);
                    }

                    //PDEBUG("Vixs_irq: %s - Clear OUT interrupt: num entry queue up =%d\n", ep->ep.name, dev->num_of_out_desc); 
                    MMR_WRITE(XC_USBD_INT_STATUS, XC_USBD_INT_STATUS_BULKOUT0_INT_TRANS_DONE_MASK);
                    
                    temp = MMR_READ(XC_USBD_BULKOUT0_STATUS2);
                    bytes_received = temp & XC_USBD_BULKOUT0_STATUS2_RCVD_BYTES_MASK;
                    num_of_entry = (temp & XC_USBD_BULKOUT0_STATUS2_FINISHED_ENTRIES_MASK)>>(XC_USBD_BULKOUT0_STATUS2_FINISHED_ENTRIES_SHIFT);
                    MMR_WRITE(XC_USBD_BULKOUT0_STATUS2, 0);
                    //PDEBUG("Vixs_irq: %s - Finished Entries=0x%x, Bytes=0x%x\n", ep->ep.name, num_of_entry, bytes_received);             
                }
                //PDEBUG("@@@@@@@@@@@@@@@@@@@@ Vixs_irq: %s Transfer Done -- Leave\n", ep->ep.name);
            }

            // Interrupt In request completed
            if (interruptstatus & XC_USBD_INT_STATUS_INTIN0_INT_EMPTY_MASK)
            {
                ep = &dev->ep [2];
                PDEBUG("Vixs_irq: Interrupt In Transfer Done %s\n", ep->ep.name);
                
                // make sure any leftover request state is cleared //
                while (!list_empty (&ep->queue)) 
                {
                    req = list_entry (ep->queue.next, struct vixs_request, queue);
                    done (ep, req, (req->req.actual == req->req.length)? 0 : -EPROTO);
                }

                // Disable the interrupt mask
                temp = MMR_READ(XC_USBD_MIPS_INT_MASK);
                temp &= ~XC_USBD_MIPS_INT_MASK_MIPS_INTIN0_INT_EN_EMPTY_MASK;
                MMR_WRITE(XC_USBD_MIPS_INT_MASK, temp);
                
                MMR_WRITE(XC_USBD_INT_STATUS, XC_USBD_INT_STATUS_INTIN0_INT_EMPTY_MASK);
            }

            // Endpoint Control request completed
            if (interruptstatus & XC_USBD_INT_STATUS_EP0_INT_EP0_IN_EMPTY_MASK)
            {
                ep = &dev->ep [0];
                PDEBUG("Vixs_irq: Control Transfer Done(1): req=0x%lx, %s, USBD_INT_STATUS=0x%lx\n", req, ep->ep.name, MMR_READ(XC_USBD_INT_STATUS));                
                // make sure any leftover request state is cleared //
                while (!list_empty (&ep->queue)) 
                {
                    req = list_entry (ep->queue.next, struct vixs_request, queue);
                    done (ep, req, (req->req.actual == req->req.length)? 0 : -EPROTO);
                }
                //PDEBUG("Vixs_irq: Control Transfer Done(2): %s, USBD_INT_STATUS=0x%lx\n", ep->ep.name, MMR_READ(XC_USBD_INT_STATUS));
                // Disable the interrupt mask
                temp = MMR_READ(XC_USBD_MIPS_INT_MASK);
                temp &= ~XC_USBD_MIPS_INT_MASK_MIPS_EP0_INT_EN_EP0_IN_EMPTY_MASK;
                MMR_WRITE(XC_USBD_MIPS_INT_MASK, temp);
                
                //PDEBUG("Vixs_irq: Control Transfer Done(3): %s, USBD_INT_STATUS=0x%lx\n", ep->ep.name, MMR_READ(XC_USBD_INT_STATUS));                
                MMR_WRITE(XC_USBD_INT_STATUS, XC_USBD_INT_STATUS_EP0_INT_EP0_IN_EMPTY_MASK);
                //PDEBUG("Vixs_irq: Control Transfer Done(4): %s, USBD_INT_STATUS=0x%lx\n\n", ep->ep.name, MMR_READ(XC_USBD_INT_STATUS));                
            }
            
            // Setup packet received
            if (interruptstatus & XC_USBD_INT_STATUS_EP0_INT_SETUP_RCVD_MASK)
            {
                // Read the setup packet
                *dev->setup_packet = MMR_READ(XC_USBD_SETUP_CMD0);
                *(dev->setup_packet+1) = MMR_READ(XC_USBD_SETUP_CMD1);   
                //PDEBUG("\n");                                     
                PDEBUG("Vixs_irq: Setup Packet -- 0x%lx 0x%lx : USBD_INT_STATUS=0x%lx\n", *dev->setup_packet, *(dev->setup_packet+1), MMR_READ(XC_USBD_INT_STATUS));
                
                // Indicate firmware has get the setup packet 
                MMR_WRITE(XC_USBD_ACK, MMR_READ(XC_USBD_ACK)|XC_USBD_ACK_SETUP_CMD_EMPTY_ACK_MASK);

                USBD_setup_irq(dev, (struct usb_ctrlrequest *)dev->setup_packet);
#error af
                // Clear setup up packet interrupt
                MMR_WRITE(XC_USBD_INT_STATUS, XC_USBD_INT_STATUS_EP0_INT_SETUP_RCVD_MASK);
                //PDEBUG("Vixs_irq: Setup Packet Clear Int : USBD_INT_STATUS=0x%lx\n", MMR_READ(XC_USBD_INT_STATUS));
            }

            // Set Configuration
            if (interruptstatus & XC_USBD_INT_STATUS_INT_SETCONFIGURATION_MASK)
            {
#ifdef CONFIG_USB_ETH_RNDIS                    
                *dev->setup_packet = 0x00020900;
#else
                *dev->setup_packet = 0x00010900;
#endif
                *(dev->setup_packet+1) = 0x00000000;                                        
                PDEBUG("Vixs_irq: Set Configuration-- 0x%lx 0x%lx\n", *dev->setup_packet, *(dev->setup_packet+1));
                USBD_setup_irq(dev, (struct usb_ctrlrequest *)dev->setup_packet);

                // Clear setup up packet interrupt
                MMR_WRITE(XC_USBD_INT_STATUS, XC_USBD_INT_STATUS_INT_SETCONFIGURATION_MASK);
            }

            // Set Interface
            if (interruptstatus & XC_USBD_INT_STATUS_INT_SETINTERFACE_MASK)
            {
                *dev->setup_packet = 0x00010B01;
                *(dev->setup_packet+1) = 0x00000001;                                        
                PDEBUG("Vixs_irq: Set Interface-- 0x%lx 0x%lx\n", *dev->setup_packet, *(dev->setup_packet+1));
                USBD_setup_irq(dev, (struct usb_ctrlrequest *)dev->setup_packet);

                // Clear setup up packet interrupt
                MMR_WRITE(XC_USBD_INT_STATUS, XC_USBD_INT_STATUS_INT_SETINTERFACE_MASK);
            }

            // Reset
            if (interruptstatus & XC_USBD_INT_STATUS_INT_RESET_MASK)
            {
		    PDEBUG("Vixs_irq: get reset irq\n");
                USBD_reset_irq(dev);
                MMR_WRITE(XC_USBD_INT_STATUS, XC_USBD_INT_STATUS_INT_RESET_MASK);
            }

            // Suspend
            if (interruptstatus & XC_USBD_INT_STATUS_INT_SUSPEND_MASK)
            {
		    PDEBUG("Vixs_irq: get suspend irq\n");
		    USBD_suspend_irq(dev);
		    MMR_WRITE(XC_USBD_INT_STATUS, XC_USBD_INT_STATUS_INT_SUSPEND_MASK);           
            }

            // Wakeup
            if (interruptstatus & XC_USBD_INT_STATUS_INT_WAKEUP_MASK)
            {
		    PDEBUG("Vixs_irq: get wakeup irq\n");
                USBD_wakeup_irq(dev);
                MMR_WRITE(XC_USBD_INT_STATUS, XC_USBD_INT_STATUS_INT_WAKEUP_MASK);
            }   

            // Device Speed
            if (interruptstatus & XC_USBD_INT_STATUS_INT_SPEED_MASK)
            {
		    PDEBUG("Vixs_irq: get speed irq\n");
                USBD_speed_irq(dev);
                MMR_WRITE(XC_USBD_INT_STATUS, XC_USBD_INT_STATUS_INT_SPEED_MASK);
            }
        }

        interruptvalue = MMR_READ (XC_MIPS3_INTERRUPT);
        interruptvalue &= MMR_READ (XC_MIPS3_INTERRUPT_MASK);

        if(interruptvalue == 0)
		loopcntl = 0;
    }

//    spin_unlock (&dev->lock);
//    PDEBUG("Vixs_irq: @@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@ Leave\n");

    return IRQ_HANDLED;
}
#endif 

//-------------------------------------------------------------------------//

//////////////////////////////////////
//
//  ViXS -- Gadget Release
//
//////////////////////////////////////
static void gadget_release (struct device *_dev)
{
    struct vixs	*dev = dev_get_drvdata (_dev);
    PDEBUG("gadget_release: -- dev = 0x%lx\n", dev);
    kfree (dev);
}

int vixs_usbd_probe (void)
{
	struct vixs         *dev;
	int                 retval;
	int                 temp;
	int                 i=0;

	PDEBUG("%s: start\n", __func__);
	retval = vixs_usbd_init();

	do {
		int j = 0;
		for(j=0; i < 100; i++) {
			udelay(10000);
			printf(".");
		}
	} while(0);

	PRINT_PHY_CLK();

	if (retval) {
		printf("USBD init failed: %d\n", retval);
		goto done;
	}


    if (the_controller) 
	{
		PDEBUG("Vixs_Probe: -- Leave : ERROR 01\n");
		return -EBUSY;
	}

	// alloc, and start init
	dev = malloc (sizeof *dev);
	if (dev == NULL) {
		retval = -ENOMEM;
		goto done;
	} else {
		memset(dev, 0, sizeof(struct vixs));
		PDEBUG("Vixs_Probe: -- dev = 0x%lx\n", dev);
	}
	dev->resume_state = USB_STATE_NOTATTACHED;
	dev->usb_state = USB_STATE_POWERED;
	dev->ep0_dir = 0;
	dev->remote_wakeup = 0; // default to 0 on reset

	dev->dev = NULL;	/* no longer using Linux device model */
	dev->gadget.ops = &vixs_ops;
	dev->gadget.is_dualspeed = 1;
	dev->gadget.ep0 = &dev->ep[0].ep;

// Make the double link list point to itself
	INIT_LIST_HEAD(&dev->gadget.ep_list);

	dev->gadget.speed = USB_SPEED_UNKNOWN;
	dev->gadget.name = driver_name;

	dev->enabled = 1;

// Start Initialize XC3 USBD PCI device
	dev->physfb = 0x80000000; // Above 2G is all kernel memory space
	dev->fbsize = 0x4000000; // make it 64M

	/* We do not need FB anyway? */
//	dev->pmmfb = ioremap_nocache(dev->physfb, dev->fbsize);
//	PDEBUG("Vixs_Probe: Virtual Frame Buffer Base Addr = 0x%lx\n", dev->pmmfb);

	dev->physmmr = VIXS_MMR_BASE;
	dev->mmrsize = 0x10000; //64k
	dev->pmmr = (unsigned char *)VIXS_MMR_BASE;
	PDEBUG("Vixs_Probe: Virtual Register Base Addr = 0x%lx\n", dev->pmmr);

	the_controller = dev;

	usb_reinit (dev);


	// Endpoint Zero (Control) Descriptor define here
	// For all other endpoint, gadget layer called ep_enable with defined desc
	dev->ep[0].desc = &vixs_ep0_desc;
	dev->ep[0].ep.maxpacket = USB_MAX_CTRL_PAYLOAD;
    
	// Buffer to hold the setup packet    
	dev->setup_packet = (unsigned int *) malloc(sizeof(struct usb_ctrlrequest));
	PDEBUG("Vixs_Probe: Setup Packet Address = 0x%lx\n", dev->setup_packet);

	for (i=0; i<MAX_TRANS_DESC; i++)
	{
		dev->inter_out_buffer[i] = (unsigned int)malloc((2*1024));
		PDEBUG("Vixs_Probe: Inter Out Buffer Address[%d] = 0x%lx\n", i, dev->inter_out_buffer[i]);
	}

	for (i=0; i<MAX_TRANS_DESC; i++)
	{
		dev->inter_in_buffer[i] = (unsigned int)malloc((2*1024));
		PDEBUG("Vixs_Probe: Inter IN Buffer Address[%d] = 0x%lx\n", i, dev->inter_in_buffer[i]);
	}
        
	//Allocate status request
	//Initialize ep0 status request structure
	//FIXME: vixs_alloc_request() ignores ep argument
	dev->status_req = container_of(vixs_alloc_request(NULL), struct vixs_request, req);
	// allocate a small amount of memory to get valid address
	dev->status_req->req.buf = malloc(8);
	dev->status_req->req.dma = virt_to_phys(dev->status_req->req.buf);


	PRINT_PHY_CLK();
	
done:
	return (retval);

}


/* rewrite for uboot init */
#if 0
//////////////////////////////////////
//
//  ViXS -- Probe
//
//////////////////////////////////////
static int ___vixs_probe (struct platform_device *pdev)
{
    struct vixs         *dev;
    int                 retval;
    int                 temp;
    int                 i=0;

    PDEBUG("Vixs_Probe: -- Enter\n");
    if (the_controller) 
    {
        PDEBUG("Vixs_Probe: -- Leave : ERROR 01\n");
        return -EBUSY;
    }

    // alloc, and start init
//    dev = kzalloc (sizeof *dev, GFP_KERNEL);
    dev = kzalloc (sizeof *dev);
    if (dev == NULL)
    {
        retval = -ENOMEM;
        goto done;
    }
    else
    {
        memset(dev, 0, sizeof(struct vixs));
        PDEBUG("Vixs_Probe: -- dev = 0x%lx\n", dev);    
    }
    dev->resume_state = USB_STATE_NOTATTACHED;
    dev->usb_state = USB_STATE_POWERED;
    dev->ep0_dir = 0;
    dev->remote_wakeup = 0; // default to 0 on reset

//    spin_lock_init (&dev->lock);

    dev->dev = &pdev->dev;
    dev->gadget.ops = &vixs_ops;
    dev->gadget.is_dualspeed = 1;
    dev->gadget.ep0 = &dev->ep[0].ep;
    // Make the double link list point to itself
    INIT_LIST_HEAD(&dev->gadget.ep_list);

    // the "gadget" abstracts/virtualizes the controller
    strcpy (dev->gadget.dev.bus_id, "gadget");
    dev->gadget.dev.parent = &pdev->dev;
    dev->gadget.dev.dma_mask = pdev->dev.dma_mask;
    dev->gadget.dev.release = gadget_release;
    dev->gadget.speed = USB_SPEED_UNKNOWN;
    dev->gadget.name = driver_name;

    if (device_register (&dev->gadget.dev) < 0)
    {
        PDEBUG("Vixs_Probe: -- device_register Fail\n");    
        retval = -ENODEV;
        goto done;
    }
    else
    {
        PDEBUG("Vixs_Probe: -- device_register Success\n");    
    }
    dev->enabled = 1;

    // Start Initialize XC3 USBD PCI device
    dev->physfb = 0x80000000; // Above 2G is all kernel memory space
    dev->fbsize = 0x4000000; // make it 64M
    dev->pmmfb = ioremap_nocache(dev->physfb, dev->fbsize);
    PDEBUG("Vixs_Probe: Virtual Frame Buffer Base Addr = 0x%lx\n", dev->pmmfb);

    dev->physmmr = VIXS_MMR_BASE;
    dev->mmrsize = 0x10000; //64k
    dev->pmmr = (unsigned char *)VIXS_MMR_BASE;
    PDEBUG("Vixs_Probe: Virtual Register Base Addr = 0x%lx\n", dev->pmmr);

    the_controller = dev;

    //usb_reset (dev);
    usb_reinit (dev);

    // Endpoint Zero (Control) Descriptor define here
    // For all other endpoint, gadget layer called ep_enable with defined desc
    dev->ep[0].desc = &vixs_ep0_desc;
    dev->ep[0].ep.maxpacket = USB_MAX_CTRL_PAYLOAD;
    
    // Buffer to hold the setup packet    
    dev->setup_packet = (unsigned int *) kmalloc(sizeof(struct usb_ctrlrequest), GFP_KERNEL);
    PDEBUG("Vixs_Probe: Setup Packet Address = 0x%lx\n", dev->setup_packet);

    for (i=0; i<MAX_TRANS_DESC; i++)
    {
        dev->inter_out_buffer[i] = (unsigned int)kmalloc((2*1024), GFP_KERNEL);
        PDEBUG("Vixs_Probe: Inter Out Buffer Address[%d] = 0x%lx\n", i, dev->inter_out_buffer[i]);
    }

    for (i=0; i<MAX_TRANS_DESC; i++)
    {
        dev->inter_in_buffer[i] = (unsigned int)kmalloc((2*1024), GFP_KERNEL);
        PDEBUG("Vixs_Probe: Inter IN Buffer Address[%d] = 0x%lx\n", i, dev->inter_in_buffer[i]);
    }
        
    //Allocate status request
    //Initialize ep0 status request structure
    //FIXME: vixs_alloc_request() ignores ep argument
    dev->status_req = container_of(vixs_alloc_request(NULL, GFP_KERNEL), struct vixs_request, req);
    // allocate a small amount of memory to get valid address
    dev->status_req->req.buf = kmalloc(8, GFP_KERNEL);
    dev->status_req->req.dma = virt_to_phys(dev->status_req->req.buf);
    
    dev->Pci_Interrupt_Line = 6;
    if (request_irq (dev->Pci_Interrupt_Line, vixs_irq, IRQF_SHARED, driver_name, dev)!= 0) 
    {
        PDEBUG ("Vixs_Probe: Request interrupt %d failed\n", dev->Pci_Interrupt_Line);
        retval = -EBUSY;
        dev->Pci_Interrupt_Line = 0;
        goto done;
    }
    else
    {
        PDEBUG("Vixs_Probe: Hook IRQ Service Routine Successfully\n");
        dev->got_irq = 1;
    }
       
    return 0;

done:
    // Only come to here when there's error occures
    if (dev)
    {
        temp = vixs_remove (pdev);
    }
    return retval;
}
#endif

/* rewrite for uboot init */
#if 0
//////////////////////////////////////
//
//  ViXS -- Remove
//
//////////////////////////////////////
static int vixs_remove (struct platform_device *pdev)
{
    int i=0;
    PDEBUG("Vixs Remove - Enter\n");

    // Clean up the resources which allocated during probe()
    if (the_controller->setup_packet)
    {
        PDEBUG("Vixs Remove - free setup packet\n");
        kfree(the_controller->setup_packet);
    }

    for (i=0; i<MAX_TRANS_DESC; i++)
    {
        if (the_controller->inter_out_buffer[i])
        {
            PDEBUG("Vixs Remove - free inter out buffer[%d]\n", i);
            kfree((pvoid)the_controller->inter_out_buffer[i]);
        }
    }

    for (i=0; i<MAX_TRANS_DESC; i++)
    {
        if (the_controller->inter_in_buffer[i])
        {
            PDEBUG("Vixs Remove - free inter in buffer[%d]\n", i);
            kfree((pvoid)the_controller->inter_in_buffer[i]);
        }
    }
    
    if (the_controller->status_req->req.buf)
    {
        PDEBUG("Vixs Remove - free buffer inside status request\n");
        kfree(the_controller->status_req->req.buf);
    }
    
    if (the_controller->status_req)
    {
        PDEBUG("Vixs Remove - free status request\n");
        kfree(the_controller->status_req);
    }
       
    if (the_controller->got_irq)
    {
        PDEBUG("Vixs Remove - free IRQ\n");   
        free_irq (the_controller->Pci_Interrupt_Line , the_controller);
    }

    if (the_controller->pmmfb)
    {
        PDEBUG("Vixs Remove - Unmap frame buffer\n");    
        iounmap (the_controller->pmmfb);
    }

    if (the_controller->enabled)
    {
        PDEBUG("Vixs Remove - Disable PCI\n");    
    }

    device_unregister (&the_controller->gadget.dev);
    //device_remove_file (&pdev->dev, &dev_attr_registers);  

    the_controller = NULL;
    PDEBUG("Vixs Remove - Leaving\n");
    return 0;
}
#endif

/* do not need this in uboot? */
#if 0
//////////////////////////////////////
//
//  ViXS -- Shutdown
//
//////////////////////////////////////
static int vixs_shutdown(struct platform_device *pdev, pm_message_t state)
{
    unsigned int        temp;
    PDEBUG("Vixs Shutdown - Enter\n");
    // Disable  the USBD block//
    temp = MMR_READ(XC_CG_RESET_REG);
    temp |= XC_CG_RESET_REG_USBD_RESET_MASK;
    MMR_WRITE(XC_CG_RESET_REG, temp);
    PDEBUG("Vixs Shutdown - (XC_CG_RESET_REG=0x%lx) Leave\n", MMR_READ(XC_CG_RESET_REG));
    return 0;
}
#endif

//////////////////////////////////////
//
//  ViXS -- Resume
//
//////////////////////////////////////
//static int vixs_resume(struct platform_device *pdev)
//{
//	return 0;
//}


/* rewrite for uboot init */
#if 0
//-------------------------------------------------------------------------//
// pci driver glue; this is a "new style" PCI driver module //
static struct platform_driver vixs_pci_driver = {
	.probe   = vixs_probe,
	
       .shutdown = vixs_shutdown,
	.remove  = vixs_remove,
	.driver  = {
		.name = XC4_DRV_NAME,
	},
};
#endif

#if 0
static struct platform_device *xc3_usbd_device = NULL;
static struct resource xc3_usbd_resources[2];
#endif
#define WAIT_EP0_FREE { int busy=0;\
                    do \
                    {\
                        busy = readl(XC_USBD_EP0_STATUS+VIXS_MMR_BASE) & XC_USBD_EP0_STATUS_OP_BUSY_MASK;\
                    }while (busy);\ 
                    }\

//////////////////////////////////////
//
//  ViXS -- Init module
//
//////////////////////////////////////
int vixs_usbd_init (void)
{
    int error;
    u32 temp;
    
    PDEBUG("%s : Gadget Driver Init July 23, 2008:(%s)\n", driver_desc, DRIVER_VERSION);

     temp = readl(CG1_BLK_CLK_STOP0+VIXS_MMR_BASE);
     temp &= ~(1<<USBD_MCLK_STOP_SHIFT);
    PDEBUG("temp:%x\n",temp);

     writel(temp,VIXS_MMR_BASE+CG1_BLK_CLK_STOP0);

    temp = readl(XC_USBD_EP0_CTRL + VIXS_MMR_BASE);
    if((temp & XC_USBD_EP0_CTRL_OP_MODE_MASK) == 0x30000000)
    {
        vixs_usbd_skip_enum= 1;
    }
    if(!vixs_usbd_skip_enum)
    {
        // Reset USBD module
        temp = readl (XC_CG_RESET_REG + VIXS_MMR_BASE);
        temp |= XC_CG_RESET_REG_USBD_RESET_MASK;
        writel(temp, (XC_CG_RESET_REG+VIXS_MMR_BASE));
       WAIT_EP0_FREE;

        temp = readl (XC_CG_RESET_REG + VIXS_MMR_BASE);
        temp &= ~XC_CG_RESET_REG_USBD_RESET_MASK;
        writel(temp, (XC_CG_RESET_REG+VIXS_MMR_BASE));
        PDEBUG("Gadget Driver Init : CG_RESET_REG=(0x%lx)\n", temp);
        WAIT_EP0_FREE;

        // Put device into full speed    
        //temp = readl (XC_USBD_UDC20_CTRL+ VIXS_MMR_BASE);
        //temp |= (1<<XC_USBD_UDC20_CTRL_APP_EXP_SPEED_SHIFT);
        //writel(temp, (XC_USBD_UDC20_CTRL+VIXS_MMR_BASE));
        //PDEBUG("Gadget Driver Init : XC_USBD_UDC20_CTRL=(0x%lx)\n", temp);
        //udelay(1000);    

        // Put VBUS in Reset
        //temp = readl (XC_USBD_PHY_CTRL + VIXS_MMR_BASE);
        //temp &= ~(XC_USBD_PHY_CTRL_DETECT_VBUS_0_ENABLE_MASK|XC_USBD_PHY_CTRL_DETECT_VBUS_1_ENABLE_MASK);
        //writel(temp, (XC_USBD_PHY_CTRL+VIXS_MMR_BASE));
        //PDEBUG("Gadget Driver Init : XC_USBD_PHY_CTRL(VBUS)=(0x%lx)\n", temp);
        //udelay(1000);

        // Put the Phy in Suspend state
        //temp = readl (XC_USBD_UDC20_CTRL + VIXS_MMR_BASE);
        //temp &= ~(XC_USBD_UDC20_CTRL_PHY_SUSPEND_ENABLE_MASK);
        //writel(temp, (XC_USBD_UDC20_CTRL+VIXS_MMR_BASE));
        //PDEBUG("Gadget Driver Init : XC_USBD_UDC20_CTRL(Phys)=(0x%lx)\n", temp);
        //udelay(1000);


    // Configurate to Communication class
    temp = readl (XC_USBD_EP0_CTRL + VIXS_MMR_BASE);
    temp |= 3 << XC_USBD_EP0_CTRL_OP_MODE_SHIFT;
    writel(temp, (XC_USBD_EP0_CTRL+VIXS_MMR_BASE));
    PDEBUG("Gadget Driver Init : XC_USBD_EP0_CTRL=(0x%lx)\n", temp);
	udelay(1000);


	//default SDK USB PHY clock: 12MHZ Cry
    writel(0x00, (XC_USBD_PHY_CLK_CTRL+VIXS_MMR_BASE));
    temp =  readl (XC_USBD_PHY_CLK_CTRL + VIXS_MMR_BASE);
    PDEBUG("Gadget Driver Init : XC_USBD_PHY_CLK_CTRL=(0x%lx)\n", temp);
    udelay(1000);

    // Put the Phy in Suspend state
    temp = readl (XC_USBD_PHY_CTRL + VIXS_MMR_BASE);
    temp &= ~(XC_USBD_PHY_CTRL_PHY_PORT1_SUSPENDM_ENB_MASK);
    /* JASON: Ref. Dominic */
    temp &= ~(XC_USBD_PHY_CTRL_PHY_PORT1_VBUSVLDEXT_ENB_MASK);
    writel(temp, (XC_USBD_PHY_CTRL+VIXS_MMR_BASE));
    PDEBUG("Gadget Driver Init : XC_USBD_PHY_CTRL(Phys)=(0x%lx)\n", temp);
    udelay(1000);
    
    PRINT_PHY_CLK();


    // Clear all exiting USBD inturrput 
    writel (0xFFFFFFFF, XC_USBD_INT_STATUS + VIXS_MMR_BASE);
    PDEBUG("Gadget Driver Init : XC_USBD_INT_STATUS = 0x%lx\n", readl (XC_MIPS3_INTERRUPT + VIXS_MMR_BASE));

    // Clear all exiting USBD inturrput in MIP4
    writel (XC_MIPS3_INTERRUPT_USBD_INT_MASK, XC_MIPS3_INTERRUPT + VIXS_MMR_BASE);
    PDEBUG("Gadget Driver Init : XC_MIPS3_INTERRUPT = 0x%lx\n", readl (XC_MIPS3_INTERRUPT + VIXS_MMR_BASE));

    // improve signal quality suggested by Dominic 
    writel (XC_USBD_PHY_OVERRIDE1_PHY_TXPREEMPHASISTUNE_MASK|XC_USBD_PHY_OVERRIDE1_DEFAULT_VALUE \
            ,XC_USBD_PHY_OVERRIDE1 + VIXS_MMR_BASE);
    
    // Take the USBD out of soft reset
    writel(0xFFFFFFFF, (XC_USBD_SOFT_RST+VIXS_MMR_BASE));
    udelay(1000);
    writel(0, (XC_USBD_SOFT_RST+VIXS_MMR_BASE));
    udelay(1000);
    PDEBUG("Gadget Driver Init : XC_USBD_SOFT_RST=(0x%lx)\n", temp);
#ifdef CONFIG_USB_ETH_RNDIS

    PDEBUG("Gadget Driver Init : CORE Configure to RNIDS\n");
    // Configure the Core    
    // Bulk IN -- Config=2, Interface=1, Alt=0, EP Num=1
    // 3:0   - Endpoint Number
    // 4      - 0:OUT, 1:IN
    // 6:5   - 10:Bulk, 11:Interrupt
    // 10:7 - Configuration Number
    // 14:11 - Interface Number
    // 18:15 - Alternate setting
    // 29:19 - Maximum packet size
    // 31:30 - Reserved for non-isochronous IN endpoint
    // (Core Setting) -- 00|01000000000|0000|0001|0010|10|1|0001
    //writel(0x10000951, XC_USBD_EP0_WR_DATA + VIXS_MMR_BASE);
    // (Core Setting) -- 00|01000000000|0000|0001|0001|10|1|0001
    writel(0x100008D1, XC_USBD_EP0_WR_DATA + VIXS_MMR_BASE);
    temp = 0x8 << XC_USBD_EP0_CTRL_REG_ADDR_SHIFT;
    temp |= 1 << XC_USBD_EP0_CTRL_REG_WOP_SHIFT;
    temp |= 0 << XC_USBD_EP0_CTRL_REG_CSR_SEL_SHIFT;
    temp |= 3 << XC_USBD_EP0_CTRL_OP_MODE_SHIFT;
    writel(temp, XC_USBD_EP0_CTRL + VIXS_MMR_BASE);
    udelay(1000);

    // Bulk OUT -- Config=2, Interface=1, Alt=0, EP Num=3
    // (Core Setting) -- 00|01000000000|0000|0001|0010|10|0|0011
    //writel(0x10000943, XC_USBD_EP0_WR_DATA + VIXS_MMR_BASE);
    // (Core Setting) -- 00|01000000000|0000|0001|0001|10|0|0011
    writel(0x100008C3, XC_USBD_EP0_WR_DATA + VIXS_MMR_BASE);
    temp = 0xC << XC_USBD_EP0_CTRL_REG_ADDR_SHIFT;
    temp |= 1 << XC_USBD_EP0_CTRL_REG_WOP_SHIFT;
    temp |= 0 << XC_USBD_EP0_CTRL_REG_CSR_SEL_SHIFT;
    temp |= 3 << XC_USBD_EP0_CTRL_OP_MODE_SHIFT;
    writel(temp, XC_USBD_EP0_CTRL + VIXS_MMR_BASE);
    udelay(1000);

    // Interrupt IN -- Config=2, Interface=0, Alt=0, EP Num=2
    // (Core Setting) -- 00|00001000000|0000|0000|0010|11|1|0010
    //writel(0x2000172, XC_USBD_EP0_WR_DATA + VIXS_MMR_BASE);
    // (Core Setting) -- 00|00001000000|0000|0000|0001|11|1|0010
    writel(0x20000F2, XC_USBD_EP0_WR_DATA + VIXS_MMR_BASE);
    temp = 0x10 << XC_USBD_EP0_CTRL_REG_ADDR_SHIFT;
    temp |= 1 << XC_USBD_EP0_CTRL_REG_WOP_SHIFT;
    temp |= 0 << XC_USBD_EP0_CTRL_REG_CSR_SEL_SHIFT;
    temp |= 3 << XC_USBD_EP0_CTRL_OP_MODE_SHIFT;
    writel(temp, XC_USBD_EP0_CTRL + VIXS_MMR_BASE);
    udelay(1000);
#else
    PDEBUG("Gadget Driver Init : CORE Configure to CDC\n");
    // Configure the Core    
    // Bulk IN -- Config=1, Interface=1, Alt=1, EP Num=1
    // 3:0   - Endpoint Number
    // 4      - 0:OUT, 1:IN
    // 6:5   - 10:Bulk, 11:Interrupt
    // 10:7 - Configuration Number
    // 14:11 - Interface Number
    // 18:15 - Alternate setting
    // 29:19 - Maximum packet size
    // 31:30 - Reserved for non-isochronous IN endpoint
    // (Core Setting) -- 00|01000000000|0001|0001|0001|10|1|0001
    writel(0x100088D1, XC_USBD_EP0_WR_DATA + VIXS_MMR_BASE);
    temp = 0x8 << XC_USBD_EP0_CTRL_REG_ADDR_SHIFT;
    temp |= 1 << XC_USBD_EP0_CTRL_REG_WOP_SHIFT;
    temp |= 0 << XC_USBD_EP0_CTRL_REG_CSR_SEL_SHIFT;
    temp |= 3 << XC_USBD_EP0_CTRL_OP_MODE_SHIFT;
    writel(temp, XC_USBD_EP0_CTRL + VIXS_MMR_BASE);
    udelay(1000);

    // Bulk OUT -- Config=1, Interface=1, Alt=1, EP Num=3
    // (Core Setting) -- 00|01000000000|0001|0001|0001|10|0|0011
    writel(0x100088C3, XC_USBD_EP0_WR_DATA + VIXS_MMR_BASE);
    temp = 0xC << XC_USBD_EP0_CTRL_REG_ADDR_SHIFT;
    temp |= 1 << XC_USBD_EP0_CTRL_REG_WOP_SHIFT;
    temp |= 0 << XC_USBD_EP0_CTRL_REG_CSR_SEL_SHIFT;
    temp |= 3 << XC_USBD_EP0_CTRL_OP_MODE_SHIFT;
    writel(temp, XC_USBD_EP0_CTRL + VIXS_MMR_BASE);
    udelay(1000);

    // Interrupt IN -- Config=1, Interface=0, Alt=0, EP Num=2
    // (Core Setting) -- 00|00001000000|0000|0000|0001|11|1|0010
    writel(0x20000F2, XC_USBD_EP0_WR_DATA + VIXS_MMR_BASE);
    temp = 0x10 << XC_USBD_EP0_CTRL_REG_ADDR_SHIFT;
    temp |= 1 << XC_USBD_EP0_CTRL_REG_WOP_SHIFT;
    temp |= 0 << XC_USBD_EP0_CTRL_REG_CSR_SEL_SHIFT;
    temp |= 3 << XC_USBD_EP0_CTRL_OP_MODE_SHIFT;
    writel(temp, XC_USBD_EP0_CTRL + VIXS_MMR_BASE);
    udelay(1000);

#endif

        printf("After set CDC\n");
        PRINT_PHY_CLK();

        /* JASON: Ref. Dominic */
        /* Enable the PORT1 */
        temp = readl (XC_USBD_PHY_CTRL + VIXS_MMR_BASE);
        temp |= (XC_USBD_PHY_CTRL_PHY_PORT1_SUSPENDM_ENB_MASK);
        temp |= (XC_USBD_PHY_CTRL_PHY_PORT1_VBUSVLDEXT_ENB_MASK);
        writel(temp, (XC_USBD_PHY_CTRL+VIXS_MMR_BASE));
        PDEBUG("Gadget Driver Init : XC_USBD_PHY_CTRL(Phys)=(0x%lx)\n", temp);
        udelay(1000);
        PRINT_PHY_CLK();

    }
    /* linux specific. rewrite for uboot */
#if 0
    // allocate platform driver structure
    xc3_usbd_device = platform_device_alloc(XC4_DRV_NAME, -1);
    if (!xc3_usbd_device)
    {
        PDEBUG("Gadget Driver Init : platform_device_alloc failure\n");
        error = -ENOMEM;
        return error;
    }
    else
    {
        PDEBUG("Gadget Driver Init : platform_device_alloc success\n");
    }

    //memory resources
    xc3_usbd_resources[0].start = (VIXS_MMR_BASE + 0x2400);
    xc3_usbd_resources[0].end = (VIXS_MMR_BASE + 0x2400 + 0x200 -1);
//    xc3_usbd_resources[0].flags = IORESOURCE_MEM;
   //irq resources
    xc3_usbd_resources[1].start = 0x6;
    xc3_usbd_resources[1].end = 0;
//    xc3_usbd_resources[1].flags = IORESOURCE_IRQ;
    error = platform_device_add_resources(xc3_usbd_device, xc3_usbd_resources, 2);
    if (error)
    {
        PDEBUG("Gadget Driver Init : platform_device_add_resources failure\n");
        goto err_free_device;
    }
    else
    {
        PDEBUG("Gadget Driver Init : platform_device_add_resources success\n");
    }

    error = platform_device_add(xc3_usbd_device);
    if (error)
    {
        PDEBUG("Gadget Driver Init : platform_device_add failure\n");
        goto err_free_device;
    }
    else
    {
        PDEBUG("Gadget Driver Init : platform_device_add success\n");
    }

//    error = platform_driver_register (&vixs_pci_driver);
    if (error < 0)
    {
        PDEBUG("Gadget Driver Init : platform_driver_register failure\n");
        goto err_free_device;
    }
    else
    {
        PDEBUG("Gadget Driver Init : platform_driver_register success\n");
    }
#endif
   
    return 0; 

err_free_device:
//    platform_device_put(xc3_usbd_device);
    return error;
}
//module_init (init);

//////////////////////////////////////
//
//  ViXS -- Clean up module
//
//////////////////////////////////////
static void cleanup (void)
{
#if 0
    PDEBUG("%s : Gadget Driver cleanup: (%s)\n", driver_desc, DRIVER_VERSION);
   if(xc3_usbd_device)
    {
        platform_device_del(xc3_usbd_device);
        xc3_usbd_device = NULL;
        platform_driver_unregister (&vixs_pci_driver);
    }
#endif
}

int usb_gadget_handle_interrupts(void)
{
	volatile unsigned int interruptstatus;

	struct vixs *dev = the_controller;
	struct vixs_ep  *ep;
	struct vixs_request *req;
	struct list_head *next_element;

	unsigned int temp=0;
	unsigned int bytes_received=0;
	unsigned int num_of_entry=0;

	unsigned long flags = 0;
	
        while (1) {

			interruptstatus = MMR_READ(XC_USBD_INT_STATUS);
			interruptstatus &= MMR_READ(XC_USBD_MIPS_INT_MASK);

                     if(!interruptstatus)
                        break;

			// Device Speed
			// It must be handled first, otherwise all of the following request will be failed.
			if (interruptstatus & XC_USBD_INT_STATUS_INT_SPEED_MASK)
			{
				PDEBUG("Vixs_irq: get speed irq\n");
				USBD_speed_irq(dev);
				MMR_WRITE(XC_USBD_INT_STATUS, XC_USBD_INT_STATUS_INT_SPEED_MASK);
			}

			// Bulk Out request completed
			if (interruptstatus & 
			    XC_USBD_INT_STATUS_BULKOUT0_INT_TRANS_DONE_MASK) {
				ep = &dev->ep [3];
//				PDEBUG("@@@@@@@@@@@@@@@@@@@@ Vixs_irq: %s Transfer OUT Done -- Enter\n", ep->ep.name);
				// make sure any leftover request state is cleared //
				//PDEBUG("\nVixs_irq: %s - Clear OUT interrupt: num entry queue up =%d\n", ep->ep.name, dev->num_of_out_desc); 
				MMR_WRITE(XC_USBD_INT_STATUS, XC_USBD_INT_STATUS_BULKOUT0_INT_TRANS_DONE_MASK);

				temp = MMR_READ(XC_USBD_BULKOUT0_STATUS2);
				bytes_received = temp & XC_USBD_BULKOUT0_STATUS2_RCVD_BYTES_MASK;
				num_of_entry = (temp & XC_USBD_BULKOUT0_STATUS2_FINISHED_ENTRIES_MASK)>>(XC_USBD_BULKOUT0_STATUS2_FINISHED_ENTRIES_SHIFT);
				MMR_WRITE(XC_USBD_BULKOUT0_STATUS2, 0);
//				PDEBUG("Vixs_irq: %s - Finished Entries=0x%x, Bytes=0x%x\n", ep->ep.name, num_of_entry, bytes_received);

				while (num_of_entry) {
					req = list_entry (ep->queue.next,
							  struct vixs_request, queue);

					if (req->req.actual == req->req.length) {
						req->req.actual = bytes_received;
						//memcpy(req->req.buf, dev->inter_out_buffer[dev->recv_buffer_index], req->req.actual);
						//dev->recv_buffer_index = (dev->recv_buffer_index+1)%MAX_TRANS_DESC;                       

						/* JASON XXX */
//						inv_dcache_range((unsigned long)req->req.buf, (unsigned long)req->req.buf+req->req.length);
						done (ep, req, 0);
						dev->num_of_out_desc-=1;
						// Setup another transfer descriptor 
						if(!list_empty (&ep->queue)) {

							req = list_entry (ep->queue.next, struct vixs_request, queue);
							next_element = ep->queue.next->next;
							//PDEBUG("Vixs_irq: %s - Next req=0x%lx, insert=%d\n", ep->ep.name, req, req->insert);
							while(req->insert==1)
							{
								if (!list_empty (&ep->queue))
								{
									req = list_entry (next_element, struct vixs_request, queue);
									//PDEBUG("Vixs_irq: %s - Next req=0x%lx, insert=%d\n", ep->ep.name, req, req->insert);
									next_element = next_element->next;
								}
								else
								{
									// Request List Empty
									//PDEBUG("Vixs_irq: - Request List Empty\n");
									break;
								}
							}
							//PDEBUG("Vixs_irq: %s - We can setup descriptor for req=0x%lx\n", ep->ep.name, req);                            
							req->insert = 1;
							write_fifo(ep, &req->req);
						}
					}
					else {
						while(1)                        
							PDEBUG("Vixs_irq: - BULK OUT ERROR : actual 0x%x != length 0x%x\n", req->req.actual, req->req.length);
					}

					//PDEBUG("Vixs_irq: %s - Clear OUT interrupt: num entry queue up =%d\n", ep->ep.name, dev->num_of_out_desc); 
					MMR_WRITE(XC_USBD_INT_STATUS, XC_USBD_INT_STATUS_BULKOUT0_INT_TRANS_DONE_MASK);
                    
					temp = MMR_READ(XC_USBD_BULKOUT0_STATUS2);
					bytes_received = temp & XC_USBD_BULKOUT0_STATUS2_RCVD_BYTES_MASK;
					num_of_entry = (temp & XC_USBD_BULKOUT0_STATUS2_FINISHED_ENTRIES_MASK)>>(XC_USBD_BULKOUT0_STATUS2_FINISHED_ENTRIES_SHIFT);
					MMR_WRITE(XC_USBD_BULKOUT0_STATUS2, 0);
					//PDEBUG("Vixs_irq: %s - Finished Entries=0x%x, Bytes=0x%x\n", ep->ep.name, num_of_entry, bytes_received);             
				}
				//PDEBUG("@@@@@@@@@@@@@@@@@@@@ Vixs_irq: %s Transfer Done -- Leave\n", ep->ep.name);
			}

			// Interrupt In request completed
			if (interruptstatus & XC_USBD_INT_STATUS_INTIN0_INT_EMPTY_MASK)
			{
				ep = &dev->ep [2];
				PDEBUG("Vixs_irq: Interrupt In Transfer Done %s\n", ep->ep.name);
                
				// make sure any leftover request state is cleared //
				while (!list_empty (&ep->queue)) 
				{
					req = list_entry (ep->queue.next, struct vixs_request, queue);
					done (ep, req, (req->req.actual == req->req.length)? 0 : -EPROTO);
				}

				// Disable the interrupt mask
				temp = MMR_READ(XC_USBD_MIPS_INT_MASK);
				temp &= ~XC_USBD_MIPS_INT_MASK_MIPS_INTIN0_INT_EN_EMPTY_MASK;
				MMR_WRITE(XC_USBD_MIPS_INT_MASK, temp);
                
				MMR_WRITE(XC_USBD_INT_STATUS, XC_USBD_INT_STATUS_INTIN0_INT_EMPTY_MASK);
			}

			// Endpoint Control request completed
			if (interruptstatus & XC_USBD_INT_STATUS_EP0_INT_EP0_IN_EMPTY_MASK)
			{
				ep = &dev->ep [0];
				PDEBUG("Vixs_irq: Control Transfer Done(1): req=0x%lx, %s, USBD_INT_STATUS=0x%lx\n", req, ep->ep.name, MMR_READ(XC_USBD_INT_STATUS));                
				// make sure any leftover request state is cleared //
				while (!list_empty (&ep->queue)) 
				{
					req = list_entry (ep->queue.next, struct vixs_request, queue);
					done (ep, req, (req->req.actual == req->req.length)? 0 : -EPROTO);
					break; /* HACK! */
				}
				//PDEBUG("Vixs_irq: Control Transfer Done(2): %s, USBD_INT_STATUS=0x%lx\n", ep->ep.name, MMR_READ(XC_USBD_INT_STATUS));
				// Disable the interrupt mask
				temp = MMR_READ(XC_USBD_MIPS_INT_MASK);
				temp &= ~XC_USBD_MIPS_INT_MASK_MIPS_EP0_INT_EN_EP0_IN_EMPTY_MASK;
				MMR_WRITE(XC_USBD_MIPS_INT_MASK, temp);
                
				//PDEBUG("Vixs_irq: Control Transfer Done(3): %s, USBD_INT_STATUS=0x%lx\n", ep->ep.name, MMR_READ(XC_USBD_INT_STATUS));                
				MMR_WRITE(XC_USBD_INT_STATUS, XC_USBD_INT_STATUS_EP0_INT_EP0_IN_EMPTY_MASK);
				//PDEBUG("Vixs_irq: Control Transfer Done(4): %s, USBD_INT_STATUS=0x%lx\n\n", ep->ep.name, MMR_READ(XC_USBD_INT_STATUS));                
			}
            
			// Setup packet received
			if (interruptstatus & XC_USBD_INT_STATUS_EP0_INT_SETUP_RCVD_MASK)
			{
				// Read the setup packet
				*dev->setup_packet = MMR_READ(XC_USBD_SETUP_CMD0);
				*(dev->setup_packet+1) = MMR_READ(XC_USBD_SETUP_CMD1);   
				//PDEBUG("\n");                                     
				PDEBUG("Vixs_irq: Setup Packet -- 0x%lx 0x%lx : USBD_INT_STATUS=0x%lx\n", *dev->setup_packet, *(dev->setup_packet+1), MMR_READ(XC_USBD_INT_STATUS));
                
				// Indicate firmware has get the setup packet 
				MMR_WRITE(XC_USBD_ACK, MMR_READ(XC_USBD_ACK)|XC_USBD_ACK_SETUP_CMD_EMPTY_ACK_MASK);

				USBD_setup_irq(dev, (struct usb_ctrlrequest *)dev->setup_packet);

				// Clear setup up packet interrupt
				MMR_WRITE(XC_USBD_INT_STATUS, XC_USBD_INT_STATUS_EP0_INT_SETUP_RCVD_MASK);
				//PDEBUG("Vixs_irq: Setup Packet Clear Int : USBD_INT_STATUS=0x%lx\n", MMR_READ(XC_USBD_INT_STATUS));
			}

			// Set Configuration
			if (interruptstatus & XC_USBD_INT_STATUS_INT_SETCONFIGURATION_MASK)
			{
#ifdef CONFIG_USB_ETH_RNDIS						
				*dev->setup_packet = 0x00020900;
#else									
				*dev->setup_packet = 0x00010900;
#endif									
				*(dev->setup_packet+1) = 0x00000000;                                        
				PDEBUG("Vixs_irq: Set Configuration-- 0x%lx 0x%lx\n", *dev->setup_packet, *(dev->setup_packet+1));
				USBD_setup_irq(dev, (struct usb_ctrlrequest *)dev->setup_packet);

				// Clear setup up packet interrupt
				MMR_WRITE(XC_USBD_INT_STATUS, XC_USBD_INT_STATUS_INT_SETCONFIGURATION_MASK);
			}

			// Set Interface
			if (interruptstatus & XC_USBD_INT_STATUS_INT_SETINTERFACE_MASK)
			{
				*dev->setup_packet = 0x00010B01;
				*(dev->setup_packet+1) = 0x00000001;                                        
				PDEBUG("Vixs_irq: Set Interface-- 0x%lx 0x%lx\n", *dev->setup_packet, *(dev->setup_packet+1));
				USBD_setup_irq(dev, (struct usb_ctrlrequest *)dev->setup_packet);

				// Clear setup up packet interrupt
				MMR_WRITE(XC_USBD_INT_STATUS, XC_USBD_INT_STATUS_INT_SETINTERFACE_MASK);
			}

			// Reset
			if (interruptstatus & XC_USBD_INT_STATUS_INT_RESET_MASK)
			{
				PDEBUG("Vixs_irq: get reset irq\n");
				USBD_reset_irq(dev);
				MMR_WRITE(XC_USBD_INT_STATUS, XC_USBD_INT_STATUS_INT_RESET_MASK);
			}

			// Suspend
			if (interruptstatus & XC_USBD_INT_STATUS_INT_SUSPEND_MASK)
			{
				PDEBUG("Vixs_irq: get suspend irq\n");
				USBD_suspend_irq(dev);
				MMR_WRITE(XC_USBD_INT_STATUS, XC_USBD_INT_STATUS_INT_SUSPEND_MASK);           
			}

			// Wakeup
			if (interruptstatus & XC_USBD_INT_STATUS_INT_WAKEUP_MASK)
			{
				PDEBUG("Vixs_irq: get wakeup irq\n");
				USBD_wakeup_irq(dev);
				MMR_WRITE(XC_USBD_INT_STATUS, XC_USBD_INT_STATUS_INT_WAKEUP_MASK);
			}   

	}


	return (0);
}

void usbse_init(void)
{
    //asume usb 2.0
    the_controller->gadget.speed = USB_SPEED_HIGH;
	//clear interrupt (should be done in boot fw code, but it is ok)
    MMR_WRITE(XC_USBD_INT_STATUS, XC_USBD_INT_STATUS_INT_SETINTERFACE_MASK);
	MMR_WRITE(XC_USBD_INT_STATUS, XC_USBD_INT_STATUS_INT_SETCONFIGURATION_MASK);
	MMR_WRITE(XC_USBD_INT_STATUS, XC_USBD_INT_STATUS_INT_RESET_MASK);
	MMR_WRITE(XC_USBD_INT_STATUS, XC_USBD_INT_STATUS_INT_SUSPEND_MASK);   
	MMR_WRITE(XC_USBD_INT_STATUS, XC_USBD_INT_STATUS_INT_WAKEUP_MASK);
	MMR_WRITE(XC_USBD_INT_STATUS, XC_USBD_INT_STATUS_INT_SPEED_MASK);
}
/*
module_exit (cleanup);

MODULE_DESCRIPTION (DRIVER_DESC);
MODULE_AUTHOR ("DRIVER_AUTHOR");
MODULE_LICENSE ("GPL");
*/

